\ *****************************************************************************
\ ChooseFile.f  2016 Mar 05  Choose a file
\ *****************************************************************************

0  
   OFN_HIDEREADONLY OR
   OFN_LONGNAMES OR
\   OFN_ALLOWMULTISELECT OR   \ Note : if you select this option the return string will be the path then a list of filenames,
                              \ all space separated, then null terminated
CONSTANT DEFAULT-OPEN-FLAGS

0  
   OFN_HIDEREADONLY OR
   OFN_LONGNAMES OR
   OFN_READONLY OR
\   OFN_ALLOWMULTISELECT OR
CONSTANT DEFAULT-READONLY-FLAGS

0  
   OFN_HIDEREADONLY OR
   OFN_NOCHANGEDIR OR
   OFN_LONGNAMES OR
   OFN_OVERWRITEPROMPT OR
   OFN_EXTENSIONDIFFERENT OR
\   OFN_ALLOWMULTISELECT OR
CONSTANT DEFAULT-SAVE-FLAGS

CREATE ALL-FILES
   ,Z" All files (*.*)"       ,Z" *.*"  
   0 ,

CREATE SRC-FILES
   ,Z" SRC files (*.SRC)"     ,Z" *.SRC"  
   ,Z" All files (*.*)"       ,Z" *.*"  
   0 ,

CREATE FORTH-FILES
   ,Z" All                       *.*"  ,Z" *.*"
   ,Z" Forth f;fth;bld;ctl;4th;pi;fs"  ,Z" *.f;*.fth;*.bld;*.ctl;*.4th;*.pi;*.fs"  
   0 ,

CREATE TEXT-FILES
   ,Z" Text files (*.txt)"    ,Z" *.txt"
   ,Z" All files (*.*)"       ,Z" *.*"
   0 ,

CREATE LOG-FILES
   ,Z" log files (*.log)"     ,Z" *.log"
   ,Z" All files (*.*)"       ,Z" *.*"
   0 ,

CREATE PROGRAM-FILES
   ,Z" Programs (*.exe)"      ,Z" *.exe"
   ,Z" Library files (*.dll)" ,Z" *.dll"
   ,Z" All files (*.*)"       ,Z" *.*"
   0 ,

CREATE WEED-FILES
   ,Z" All                                 *.*"  ,Z" *.*"
   ,Z" Forth           f;fth;bld;ctl;4th;pi;fs"  ,Z" *.f;*.fth;*.bld;*.ctl;*.4th;*.pi;*.fs"  
   ,Z" Assembler            asm;nasm;s;inc;lst"  ,Z" *.asm;*.nasm;*.s;*.inc;*.lst"
   ,Z" C/C++/C#  c;cpp;cs;h;hpp;rc;cca;tol;hdl"  ,Z" *.cpp;*.c;*.cxx;*.h;*.hpp;*.rc;*.cca;*.tol;*.hdl"  
   ,Z" Text                                txt"  ,Z" *.txt"
   0 ,

CREATE HEADER-FILES
   ,Z" Header files (*.h*)"   ,Z" *.h*"  
   ,Z" All files (*.*)"       ,Z" *.*"  
   0 ,

CREATE BINARY-FILES
   ,Z" Binary image    bin;ramf;img"  ,Z" *.bin;*.ram;*.img"  
   ,Z" Text                     txt"  ,Z" *.txt"
   ,Z" All                      *.*"  ,Z" *.*"
   ,Z" Assembler asm;nasm;s;inc;lst"  ,Z" *.asm;*.nasm;*.s;*.inc;*.lst"
   ,Z" C/C++/C# c;cpp;cs;h;hpp;rc;cca;tol;hdl"  ,Z" *.cpp;*.c;*.cxx;*.h;*.hpp;*.rc;*.cca;*.tol;*.hdl"  
   0 ,

2048 constant |OFNfilenameBuffer| 
|OFNfilenameBuffer|  Buffer: OFNfilenameBuffer
|OFNfilenameBuffer|  Buffer: OFN_zInitialDirBuffer
Create s_OPENFILENAME  $58 allot

s_OPENFILENAME
   4 struct OFN_StructSize         \ size in CVARIABLEs of structure
   4 struct OFN_Owner              \ hwnd of the owner's window
   4 struct OFN_Instance           \ hinst, used for overlay templates
   4 struct OFN_zFilter            \ pairs of zstr, followed by null
   4 struct OFN_zCustomFilter      \ buffer to save user filters in
   4 struct OFN_MaxCustFilter      \ length of custfilter buffer
   4 struct OFN_FilterIndex        \ filter index for initial view
   4 struct OFN_zFile              \ buffer for filename
   4 struct OFN_MaxFile            \ length of zfile
   4 struct OFN_zFileTitle         \ buffer for file title
   4 struct OFN_MaxFileTitle       \ length of zfiletitle
   4 struct OFN_zInitialDir        \ where to run, NULL is current
   4 struct OFN_zTitle             \ title of dialog
   4 struct OFN_Flags              \ creation flags, see api
   2 struct OFN_FileOffset         \ offset in zfile past path info
   2 struct OFN_FileExtension      \ offset in zfile past path & name
   4 struct OFN_zDefExt            \ buffer with default extension
   4 struct OFN_CustData           \ custom data for hook
   4 struct OFN_Hook               \ callback address of custom routine
   4 struct OFN_zTemplateName      \ resource name of new template
   4 struct OFN_*pvReserved;
   4 struct OFN_dwReserved;
   4 struct OFN_FlagsEx;           \  1 OFN_EX_NOPLACESBAR  or 0
s_OPENFILENAME - constant |s_OPENFILENAME|

MAX_PATH Buffer: SavedStartDirectory 

: ChooseFile- ( a flags -- $name | 0 0 )
   
   MAX_PATH SavedStartDirectory GetCurrentDirectory drop \ save our current directory location

   s_OPENFILENAME |s_OPENFILENAME| ERASE
   |s_OPENFILENAME| OFN_StructSize !
   HINST OFN_Instance !  
   HWND OFN_Owner !  
   1 OFN_FilterIndex !
   OFNfilenameBuffer OFN_zFile !    \ returns a list of null-terminated strings here  
   |OFNfilenameBuffer| OFN_MaxFile !   
   OFN_zInitialDirBuffer OFN_zInitialDir !
   Z" Choose a file" OFN_zTitle ! 
   ( flags ) OFN_Flags !  
   ( a )     OFN_zFilter !   
   s_OPENFILENAME GetOpenFileName if  OFNfilenameBuffer zcount  else  0 0  then 

   SavedStartDirectory SetCurrentDirectory drop \ restore the original start directory
;

\ Choose a file using the given file specification at address a and the given file flags
: ChooseFile ( a flags -- $name | 0 0 )
   OFNfilenameBuffer |OFNfilenameBuffer| erase  
   OFN_zInitialDirBuffer |OFNfilenameBuffer| erase
   ChooseFile-
;

\ Choose a file starting in the given directory
: ChooseFile+ ( a flags $StartDir -- $name | 0 0 )
   OFNfilenameBuffer |OFNfilenameBuffer| erase
   OFN_zInitialDirBuffer zplace
   ChooseFile-
;

\ Usage :
: ttchoose ( -- )   WEED-FILES DEFAULT-OPEN-FLAGS ChooseFile  dup if  2 spaces type   else  2drop  ."  No file selected!"  then ;


: SaveAsFile- ( a flags -- $name | 0 0 )
   
   MAX_PATH SavedStartDirectory GetCurrentDirectory drop \ save our current directory location

   s_OPENFILENAME |s_OPENFILENAME| ERASE
   |s_OPENFILENAME| OFN_StructSize !
   HINST OFN_Instance !  
   HWND OFN_Owner !  
   1 OFN_FilterIndex !
   OFNfilenameBuffer OFN_zFile !    \ returns a list of null-terminated strings here  
   |OFNfilenameBuffer| OFN_MaxFile !   
   OFN_zInitialDirBuffer OFN_zInitialDir !
   Z" Save the file" OFN_zTitle ! 
   ( flags ) OFN_Flags !  
   ( a )     OFN_zFilter !   
   s_OPENFILENAME GetSaveFileName if  OFNfilenameBuffer zcount  else  0 0  then 

   SavedStartDirectory SetCurrentDirectory drop \ restore the original start directory
;

\ Choose a file using the given file specification at address a and the given file flags
: SaveAsFile ( a flags -- $name | 0 0 )
   OFNfilenameBuffer |OFNfilenameBuffer| erase  
   OFN_zInitialDirBuffer |OFNfilenameBuffer| erase
   SaveAsFile-
;

\ Choose a file starting in the given directory
: SaveAsFile+ ( a flags $StartDir -- $name | 0 0 )
   OFNfilenameBuffer |OFNfilenameBuffer| erase
   OFN_zInitialDirBuffer zplace
   SaveAsFile-
;

