\ GIF.f  GIF file parser  Howerd Oakford  www.inventio.co.uk 
\ Based on the MSDOS GIF code by myself circa 2003, which was based on
\ something even earlier, maybe from Wil Baden - sorry I can't find the original...
\ This is work in progress. 
\ ToDo: fix the horizontal hold on the ASCII art display!
\ 
\ The code size, without text strings is roughly 5 Kbytes in SwiftForth ( 32 bit Windows PC )
\ and roughly 2 Kbytes in polyForth ( 16 bit MSDOS PC ).
\ For a 240 x 128 x 1 bit colour image the image size is reduced 
\ from 240 x 128 / 8 = 3840 bytes to 2314 bytes.
\ The GIF file created by GIMP from Inventio.bmp has a 768 byte colour table, 
\ which is not required for monochrome, and should be removed.
\ So there is a net gain of 3840 - ( 2314 + 2048 - 768 ) = 246 bytes.
\ This could be even better with some optimisations for program size.
\ For a typical 1024 x 768 x 24 bit colour image the compression ratio is 7.3 times,
\ and the image size dwarfs the size of the GIF decompressor.
\ Usage : 
\ GIF                            \ dispays the pre-compiled file Inventio.gif as ASCII art
\ s" Inventio.gif" ShowGIFfile   \ displays the given file Inventio.gif, provided the width 
                                 \ is not greater than 256 pixels, otherwise fill a buffer.
\ ttAll                          \ show all files
                                 
decimal

: GIFversion   ." Gif V1.0 2016 Feb 09 " ;  

\ ****************************************************************************
\ ****************************************************************************

here constant StartHere 

\ *****************************************************************************
\ Variables
\ *****************************************************************************

variable CURRENT-CODE-BITS
variable CURRENT-CODE-MASK
variable vImageWidth
variable vImageHeight
variable vImageColourDepth

\ *****************************************************************************
\ GIF structure access in the Buffer 
\ *****************************************************************************
\ These words store the offset into the Buffer for each element in the signature structure
#12 cells constant |Signature| 
|Signature|  Buffer: Signature[]  

Signature[] 
         dup constant SIGNATURE
   cell+ dup constant COLOUR-MAP
   cell+ dup constant SCREEN
   cell+ dup constant IMAGE
   cell+ dup constant RASTER
   cell+ dup constant EXTENSION
   cell+ dup constant EXTENSION-GRAPHIC
   cell+ dup constant EXTENSION-COMMENT
   cell+ dup constant EXTENSION-TEXT
   cell+ dup constant EXTENSION-APPLICATION
   cell+ dup constant EXTENSION-NETSCAPE
   cell+ dup constant MYSTRUCT
   cell+
Signature[] - |Signature| = not [if]  cr .( ooops!!! )  abort  [then]

\ *****************************************************************************
\ Bitmap raster format output buffer, 32 bit colours 
\ *****************************************************************************

: GetColourFromMap ( c -- u )   
   3 * COLOUR-MAP @ + @ $00FFFFFF and 
;

variable BitmapBufferOriginPtr
variable BitmapBufferPtr

variable BitmapBufferOriginPtr2

\ : BitmapBuffer[] ( -- a )   BitmapBufferOriginPtr @ ;

#10 2** dup * ( 1024 x 1024 pixels ) constant |BitmapBufferNumberOfItems|
\ |BitmapBufferNumberOfItems| cells Buffer: BitmapBuffer[]

: AllocateBitmapBuffer[] ( -- )
   BitmapBufferOriginPtr @ if  exit  then
   |BitmapBufferNumberOfItems| cells allocate if  cr ." allocate failed!!!"  #9008 throw  then    
   BitmapBufferOriginPtr !
;

: FreeBitmapBuffer[] ( -- )
   BitmapBufferOriginPtr @ free if  cr ." free failed!!!"  #9009 throw  then 
   0 BitmapBufferOriginPtr @ !
;   

: AllocateBitmapBuffer2[] ( -- )
   BitmapBufferOriginPtr2 @ if  exit  then
   |BitmapBufferNumberOfItems| cells allocate if  cr ." allocate failed!!!"  #9010 throw  then    
   BitmapBufferOriginPtr2 !
;

: FreeBitmapBuffer2[] ( -- )
   BitmapBufferOriginPtr2 @ free if  cr ." free failed!!!"  #9011 throw  then 
   0 BitmapBufferOriginPtr2 @ !
;   

DRAWING_SIZE_X DRAWING_SIZE_Y * cells constant FRAME_PTR_NUMBER_OF_BYTES

: CopyGIFtoFramePtrAll ( -- )
   BitmapBufferOriginPtr @ FramePtr @ FRAME_PTR_NUMBER_OF_BYTES cmove
;

: CopyGIFtoFramePtrPart ( percentage -- )
   >r 
   BitmapBufferOriginPtr @ FramePtr @ FRAME_PTR_NUMBER_OF_BYTES r> #100 */ cmove
;


DRAWING_SIZE_X 3 4 */ constant START_COLUMN

: xxxCopyGIFlineToFramePtrColumn3 ( line# -- )
   START_COLUMN cells dup >r BitmapBufferOriginPtr @ +
   r> FramePtr @ +
   DRAWING_SIZE_X START_COLUMN - cmove
;

: xxxCopyGIFtoFramePtr ( -- )
   DRAWING_SIZE_Y 0 do
      i xxxCopyGIFlineToFramePtrColumn3
   loop
;

\ *****************************************************************************
\ *****************************************************************************

: BitmapBufferInit ( -- )   
   0 BitmapBufferPtr ! 
;

: >BitmapBuffer ( -- a)   BitmapBufferPtr @ |BitmapBufferNumberOfItems| 1- and cells BitmapBufferOriginPtr  @ + ;

defer PutPixel ( u -- )

: (PutPixel) ( u -- )
   >BitmapBuffer !  1 BitmapBufferPtr +! 
;

: PixelToBuffer  ['] (PutPixel) is PutPixel ;

variable PixelsOnOneLine

\ write pixel c's colour to the bitmap buffer, and display the image as ASCII art if possible...
: (PutPixel+) ( u -- )   
   (PutPixel)
   vImageWidth @ $100 > if    \ do not try to display GIF images that are too wide, 
      exit                    \ just put them in the BitmapBuffer    
   then 
   GREEN.  
   dup $FCFFFE >= if
      ."  "  drop
   else
      ." X"  drop \ 1 .hex  
   then
\   space  
   RESTORE.

   1 PixelsOnOneLine +!
   PixelsOnOneLine @ vImageWidth @ > if  cr  0 PixelsOnOneLine !  then    
;

: PixelToBuffer+  ['] (PutPixel+) is PutPixel ;

\ *****************************************************************************
\ Send pixels straight to the FramePtr Drawing area
\ *****************************************************************************

variable FrameBufferPtr 

: FrameBufferInit ( -- )   
   0 FrameBufferPtr ! 
   BitmapBufferInit
;

\ : >FrameBuffer ( -- a )   FrameBufferPtr @ cells FramePtr @ + ;
: >FrameBuffer ( -- a )   FrameBufferPtr @ |BitmapBufferNumberOfItems| 1- and cells BitmapBufferOriginPtr @ + ;

: RGBconvert ( u -- u' )
   >r
   r@ $10 rshift $FF and 
   r@ $FF and $10 lshift
   r@ $0000FF00 and or or
   r> drop
;

: (PutPixelToFrameBuffer) ( u -- ) 
\   FramePtr @ 0= drop -1 if  drop  exit  then
   RGBconvert 
   >FrameBuffer !  1 FrameBufferPtr +! 
   1 BitmapBufferPtr +! 
;

: PixelToFrameBuffer ( -- ) 
   FramePtr @ if
      ['] (PutPixelToFrameBuffer) is PutPixel 
   else
      ['] (PutPixel) is PutPixel
   then
;


\ write pixel c's colour to the bitmap buffer
: PixelEmit ( c -- )   GetColourFromMap ['] PutPixel catch if  drop  then ;
\ : PixelEmit ( c -- )   GetColourFromMap ['] (PutPixel) catch if  drop  then ;

\ *****************************************************************************
\ Source GIF file, one byte at a time
\ *****************************************************************************

variable SourceFile[]Ptr

: SourceFile[] SourceFile[]Ptr @ ;

variable SourceFilePtr  
variable SourceFileSize 

: SourceFileInit ( a n -- )   SourceFileSize !  SourceFile[]Ptr !  0 SourceFilePtr ! ;

\ Get the next octet value from the file buffer
: GetSourceFile8 ( -- c ) 
   SourceFilePtr @ SourceFile[] + c@
\   1 SourceFilePtr +!              \ without end limit check
   SourceFilePtr @ 1+ dup SourceFileSize @ > if  #10002 throw  then  SourceFilePtr !   \ throw at End Of File
; 

\ Get the next LittleBigEndian 2 octet value from the file buffer
: GetSourceFile16LE ( -- w )   GetSourceFile8  GetSourceFile8 #08 lshift  or ;

\ Get the next LittleEndian 4 octet value from the file buffer
: GetSourceFile32LE ( -- u )   GetSourceFile16LE  GetSourceFile16LE #16 lshift  or ;

\ put the curent buffer pointer into the variable at a
: IS-HERE ( a -- )   SourceFile[] SourceFilePtr @ +  swap ! ;

\ *****************************************************************************
\ Get a Byte from the list of counted strings of data in the file buffer
\ *****************************************************************************

variable #BYTES

\ gets the next byte from a list of counted strings.
\ get the length if we have no bytes in the current string, then get the next byte in the string
\ return 0 if none left, else the byte and -1
\ The list looks like :
\ 03 XX XX XX 04 YY YY YY YY .... 02 ZZ ZZ 00
\  The list is terminated by a zero length string ( a single 0 ).
: GET-BYTE ( -- c -1 | 0 )
   #BYTES @ 0= if
      GetSourceFile8 #BYTES ! 
   then
   #BYTES @ if  -1 #BYTES +!  GetSourceFile8  -1  else  0  then ;

: StringListSkip
   begin  GET-BYTE dup if  swap drop ( !BUF )  then  0= until 
;

: StringListShowHex ( -- ) 
   begin  GET-BYTE dup if  swap 2 .hex space   then  0= until 
;

: StringListShowASCII ( -- ) 
   begin  GET-BYTE dup if  swap emit  then  0= until 
;

\ *****************************************************************************
\ Bit buffer - feed in whole octets, read out the required number of bits
\ *****************************************************************************

#32 constant |BitBuffer#Bits|
\ |BitBuffer#Bits| 8 / Buffer: BitBuffer[]
variable BitBuffer[]       \ a 32 bit bit buffer

variable BitBufferPtrIn    \ pointer the next bit position to write into the BitBuffer[]  
variable BitBufferPtrOut   \ pointer the next bit position to read from the BitBuffer[] 

\ initialise the BitBuffer[]
: BitBufferInit   0 BitBuffer[] !  0 BitBufferPtrIn !  0 BitBufferPtrOut ! ;

\ return the number of bits in the BitBuffer[]
: GetNumberOfBitsInButBuffer ( -- n )   
   BitBufferPtrIn @ BitBufferPtrOut @ - |BitBuffer#Bits| 1- and 
\   dup 2 .hex space  Get-XY drop 80 > if  cr  then 
;

\ put 8 bits ( one octet ) into the bit-buffer
: BitBufferC! ( c -- )   
   BitBuffer[] BitBufferPtrIn @ 3 rshift $03 and + c! 
   BitBufferPtrIn @ 8 + |BitBuffer#Bits| 1- and  BitBufferPtrIn ! 
;


\ GetCodeBits  gets CURRENT-CODE-BITS bits out of the bit-buffer
\ Note that the bytes are received ( and put into the buffer ) in Little Endian order, 
\ and that up to three bytes may have to be fetched for a 10 bit or greater code size : 
\ 1 bit, 8 bits, 1 bit may be stored in 3 consecutive bytes
: GetCodeBits ( -- n )  
   BitBufferPtrOut @ 8 /mod  swap  >r  >r
   r@     3 and  BitBuffer[] + c@  
   r@ 1 + 3 and  BitBuffer[] + c@ $08 lshift or
   r> 2 + 3 and  BitBuffer[] + c@ $10 lshift or  
   r> rshift 
   CURRENT-CODE-MASK @ 1- and
   BitBufferPtrOut @  CURRENT-CODE-BITS @ + |BitBuffer#Bits| 1- and  BitBufferPtrOut ! 
;

\ *****************************************************************************
\ Get a Code
\ *****************************************************************************

: @ROOT-CODE-BITS ( -- c)  RASTER @ c@ ;

: @ROOT-CODE-MASK ( -- u)   @ROOT-CODE-BITS 2** ;
\ NEXT-FREE-CODE  is the next available location in the code table
\ As non-root codes are received, we build up the table
variable NEXT-FREE-CODE

\ THIS-CODE  holds the current code
variable THIS-CODE

\ PREVIOUS-CODE holds the last code received

\ FIRST-ROOT  holds the starting point for a walk through the code table
variable FIRST-ROOT

variable CURRENT-PREFIX
variable CODE-TO-EXPAND

\ FIRST-AFTER-CLEAR  is true for the first code received after a
\  clear code. This special code is used to prime the LZW
\  decompressor, and does not create a table entry.
variable FIRST-AFTER-CLEAR

\ resets all variables to the state specified for the current image block
: RESET-LZW ( -- )   
   @ROOT-CODE-BITS  1+ CURRENT-CODE-BITS !
   @ROOT-CODE-MASK  2* CURRENT-CODE-MASK !
   @ROOT-CODE-MASK  2+ NEXT-FREE-CODE !      \ the 2+ is because free codes start after the Clear and EOI codes 
;

\ gets the next Code from the input stream
\ It fetches bytes and puts them in the bit-buffer until there are enough bits to remove the next code
: GET-CODE-RAW ( -- n -1 | 0  )
   begin  
      GetNumberOfBitsInButBuffer CURRENT-CODE-BITS @ <  while
         GET-BYTE if  BitBufferC!  else  0  exit  then  
   repeat
   GetCodeBits  -1 
\   CURRENT-CODE-BITS @ #12 >= if 
\      cr ." GET-CODE-RAW = "  over 2 .hex   ."   CURRENT-CODE-BITS = "  CURRENT-CODE-BITS @ 2 .hex
\   then
;

\ gets the next Code from the input stream, acting on clearCode and EOI codes
: GET-CODE ( -- n -1 | 0  )   
   0 FIRST-AFTER-CLEAR !
   begin  
      GET-CODE-RAW if  
         THIS-CODE !  
         THIS-CODE @ @ROOT-CODE-MASK 1+ = if
            ."  EOI"  StringListSkip  
            0  exit  
         then
         THIS-CODE @ @ROOT-CODE-MASK = if
\            ."  CLR"  
            RESET-LZW   
            -1 FIRST-AFTER-CLEAR !
         else
            THIS-CODE @ -1  exit  
         then
      else  
         0  exit  
      then
   again 
;

#12 constant MAXIMUM_CODE_BITS

\ tests to see of the code size needs to be increased
: INCREASE-CODE-SIZE? ( -- )
   NEXT-FREE-CODE @ CURRENT-CODE-MASK @ = 
   CURRENT-CODE-BITS @ MAXIMUM_CODE_BITS < and \ do not increase the code size past the maximum
   if
      1 CURRENT-CODE-BITS +!
      CURRENT-CODE-MASK @ 1 lshift CURRENT-CODE-MASK !
\      ."  BITS="  CURRENT-CODE-BITS @ 1- 2 .hex
   then 
;

\ *****************************************************************************
\ GIF table
\ *****************************************************************************

\ the table of #4096 Codes, each three byte section contains a one byte extension and two byte code for each entry
#12 2** constant |GIF_TableNumberOfEntries|
|GIF_TableNumberOfEntries| 3 * Buffer: GIF_Table[]
: GIF_TableInit ( -- )   GIF_Table[] |GIF_TableNumberOfEntries| 3 * erase ;

\ enters a code plus extension pair into table entry n
: GIF_Table! ( code extension n -- )   |GIF_TableNumberOfEntries| 1- and  3 * GIF_Table[] + >r  r@ c!  r> 1+ w! ;

\ returns the code plus extension pair from table entry n 
: GIF_Table@ ( n -- extension code )   |GIF_TableNumberOfEntries| 1- and  3 * GIF_Table[] + >r  r@ c@  r> 1+ w@ ;

\ *****************************************************************************
\ GIF code stack
\ *****************************************************************************

\ CodeStackPtr  counts the number of pixels pushed onto the CodeStack or data Stack
variable CodeStackPtr
variable Max_CodeStackPtr  

\ create our own stack rather than using the system data stack
$1000 constant |CodeStack|
|CodeStack| Buffer: CodeStack[] 

: CodeStackInit ( -- )
   0 CodeStackPtr !   
;

: >CodeStackItem ( n -- a )   |CodeStack| 1- and  2*  CodeStack[] + ;

: CodeStackPush ( u -- )
   CodeStackPtr @ >CodeStackItem w!
   1 CodeStackPtr +!
   CodeStackPtr @ |CodeStack| >= if   cr ."  CodeStackPtr too big!!!"  ekey #27 = if  cr ." ok " cr  quit  then  then 
   CodeStackPtr @ Max_CodeStackPtr @ max  Max_CodeStackPtr ! 
;

: CodeStackPop ( -- u )
   -1 CodeStackPtr +!
   CodeStackPtr @ >CodeStackItem w@
;

{   
\ use the system data stack - simpler code but relies of a reasonably deep stack begin available
0 constant |CodeStack|

: CodeStackInit ( -- )
   0 CodeStackPtr !   
;

: CodeStackPush ( u -- )   
   1 CodeStackPtr +!
;

: CodeStackPop ( -- u )  ;
}

: UNWIND ( -- )
\   cr ." unwind depth = "  depth .   ."    Max_CodeStackPtr = " Max_CodeStackPtr @ .
   CodeStackPtr @ 0 ?do  CodeStackPop PixelEmit  loop 
;

\ *****************************************************************************
\ LZW decompression
\ *****************************************************************************

: LZW ( -- p1 p2 ... px )    
   CodeStackInit
   THIS-CODE @ @ROOT-CODE-MASK < if  
      \ its a root code
      THIS-CODE @ FIRST-ROOT !  
      exit  
   then
   THIS-CODE @ NEXT-FREE-CODE @ = if  
      \ its a new code 
      FIRST-ROOT @ CodeStackPush
      CURRENT-PREFIX @ CODE-TO-EXPAND !
   else 
      \ its an existing code
      THIS-CODE @  CODE-TO-EXPAND !
   then
   begin
      CODE-TO-EXPAND @ @ROOT-CODE-MASK > while
      CODE-TO-EXPAND @ GIF_Table@  CODE-TO-EXPAND !  CodeStackPush
   repeat
   CODE-TO-EXPAND @ FIRST-ROOT ! 
;

\ creates the next entry in the table of Codes,  unless this is the first code after a clear code, 
\ in which case the LZW decompressor is primed by putting the ( root ) code into LAST-ROOT-CODE .
: NEXT-TABLE! ( code -- ) 
   THIS-CODE !
   LZW
   FIRST-ROOT @ CodeStackPush
   UNWIND
   FIRST-AFTER-CLEAR @ if
      THIS-CODE @ FIRST-ROOT !
   else
      CURRENT-PREFIX @ FIRST-ROOT @ NEXT-FREE-CODE @ GIF_Table!
      1 NEXT-FREE-CODE +!
      INCREASE-CODE-SIZE? 
   then
   THIS-CODE @ CURRENT-PREFIX !
;

\ gets codes until a zero length counted string or an End Of Information ( EOI ) code is seen
: GET-CODES ( -- ) 
   BitBufferInit  
   RESET-LZW   
   begin  
      GET-CODE dup if  
         swap ( code -- ) NEXT-TABLE!
      then
   0= until
\   cr ." GET-CODES done "  ekey #27 = if  cr ."  ok " cr  quit  then      
;

\ Colour map
: ShowColourMapEntry ( -- )
   GetSourceFile8 
   GetSourceFile8 $08 lshift or 
   GetSourceFile8 $10 lshift or  
   6.hex space
;

( Colour map)
: ShowColourMap ( n -- )   
\   0 max #196 min 
   BLUE.
   0 ?do  
      i $0F and 0= if cr 6 spaces  then 
      ShowColourMapEntry
   loop
   RESTORE. 
;

\ processes a GIF extension block
: GIF-EXTENSION ( -- )
   GetSourceFile8    \ get the extension type
   cr  ."    Extension block, code = "  dup BLUE. 2 .hex RESTORE.  2 spaces 
   case
      $F9 of  EXTENSION-GRAPHIC  IS-HERE  StringListShowHex  endof
      $FE of  EXTENSION-COMMENT  IS-HERE  cr ."       Comment | "  MAGENTA. StringListShowASCII RESTORE.  endof
      $01 of  EXTENSION-TEXT     IS-HERE  StringListShowHex  endof
      $FF of  EXTENSION-APPLICATION  IS-HERE  StringListSkip
         s" NETSCAPE2.0"  EXTENSION-APPLICATION @ over compare 0= if   
            cr ."       NETSCAPE2.0"
            EXTENSION-APPLICATION @ #11 + EXTENSION-NETSCAPE ! 
         then
      endof
      ." GIF-EXTENSION Unknown Extension code ? "  drop   endcase
;

\ processes a GIF image block
: GIF-IMAGE
   cr  ."    Image block "
\   GetSourceFile8 $2C = not abort" Start of Image block is not $2C !!!" 
   IMAGE IS-HERE  
   cr ."       Image left position = " GetSourceFile16LE BLUE. 2 .hex RESTORE.
   cr ."       Image top  position = " GetSourceFile16LE BLUE. 2 .hex RESTORE.
   cr ."       Image width         = " GetSourceFile16LE BLUE. 2 .hex RESTORE.
   cr ."       Image height        = " GetSourceFile16LE BLUE. 2 .hex RESTORE.
   cr ."       Bits                = " GetSourceFile8 >r 
   BLUE. r@ 2 .hex RESTORE.   
   r@ $80 and if 
      COLOUR-MAP IS-HERE  
      r@ 7 and 1+ 2**  
      cr ."       Image colour map size = "  dup BLUE. 2 .hex RESTORE. 
      ShowColourMap
   then
   r> drop
   
   RASTER IS-HERE  
   
   SourceFilePtr @ >r    \ saved for checking the data size at the end
   
   GetSourceFile8 
   cr ."    LZW minimum code size = " dup BLUE. 2 .hex RESTORE.
   CURRENT-CODE-BITS !  \ initial size of LZW strings

   vImageWidth @ 2/ #05 - PixelsOnOneLine !  \ for ASCII art ToDo: fix the horizontal hold

   cr ."    Image data : "  cr 
   GET-CODES         \ decompress the image data in LZW compressed form
\   StringListSkip    \ skip the image data in LZW compressed form - use this to check that we can get the compressed data OK
   cr ."    Image data size = " SourceFilePtr @ r> - BLUE. 5 .hex RESTORE.
; 

\ select and action for each type of Block
: GIF-PARSE
\   cr BLUE. $10 0 do  GetSourceFile8 2 .hex space  loop  RESTORE.  exit
   begin
      GetSourceFile8 case
         $2C of  GIF-IMAGE  0  endof
         $21 of  GIF-EXTENSION  0  endof
         $3B of  cr GREEN. ." Normal Trailer byte" cr  RESTORE.  -1  endof
         cr  ." Unknown block type" 3 .hex  -1  cr  
         endcase
   until
;

here constant EndHere 

: TotalSize ( -- u )
    EndHere StartHere - 
   |Signature| |BitmapBufferNumberOfItems| cells + |GIF_TableNumberOfEntries| 3 * + |CodeStack| +  -
;

\ shows the GIF file at address a and of length n
: ShowGIF ( a n -- )
 
   ( a n ) SourceFileInit 

   0 Max_CodeStackPtr !
   0 #BYTES !  
  
   BitmapBufferInit 

   GIF_TableInit
   SIGNATURE |Signature| erase   \ all pointers to sections are now zero

   SIGNATURE IS-HERE  
   s" GIF89a" SIGNATURE @ over compare 0= not ABORT" Not a GIF89a signature!"  
   cr ."    Signature = "  BLUE.  6 0 do  GetSourceFile8 emit  loop  RESTORE.
   SCREEN    IS-HERE  
   cr ."    Width                   = "  GetSourceFile16LE vImageWidth !   BLUE. vImageWidth @ 4 .hex  RESTORE. 
   cr ."    Height                  = "  GetSourceFile16LE vImageHeight !  BLUE. vImageHeight @ 4 .hex  RESTORE.

   GetSourceFile8 >r 
   cr ."    Bits                    = " r@ BLUE. 2 .hex RESTORE.    r@ $7F and vImageColourDepth ! 
   cr ."    Background colour index = "  GetSourceFile8 BLUE. 2 .hex  RESTORE.
   cr ."    Pixel aspect ratio      = "  GetSourceFile8 BLUE. 2 .hex  RESTORE.
   r@ $80 and if  
      COLOUR-MAP IS-HERE  
      r@ 7 and 1+ 2**  
      cr ."       Screen colour map size = "  dup BLUE. 2 .hex RESTORE. 
      ShowColourMap
   then
   r> drop
   counter >r

   GIF-PARSE
   
   counter r> - cr ." Elapsed time = " decimal BLUE. .P3 RESTORE. ."  seconds" 
   
   cr ." SourceFileSize before decompression = "  SourceFileSize @  BLUE. 5 .hex RESTORE.  3 spaces SourceFileSize @   #10 #1024 */ BLUE. .P1 RESTORE. ."  kBytes" 
   cr ." Bitmap size after decompression     = "  BitmapBufferPtr @ BLUE. 5 .hex RESTORE.  3 spaces BitmapBufferPtr @  #10 #1024 */ BLUE. .P1 RESTORE. ."  kBytes" 
   cr ." Percentage reduction (decimal)      = "  1000 SourceFileSize @ BitmapBufferPtr @ 1 max */  BLUE. .P1 ."  %" RESTORE.
   cr ." Compression factor   (decimal)      = "  10 BitmapBufferPtr @ SourceFileSize @   */  BLUE. .P1 ."  times" RESTORE.
   \ check that the decompressed bitmap image is the right size
   BitmapBufferPtr @ vImageWidth @ vImageHeight @ * = not if  
      cr  RED. ." Bitmap size does not match image size = "  RESTORE.  vImageWidth @ vImageHeight @ * BLUE. 5 .hex RESTORE.
   then

   cr ." Max_CodeStackPtr                    = "  Max_CodeStackPtr @ BLUE. 5 .hex RESTORE.
   cr
   cr ." TotalSize (roughly)                 = "  TotalSize BLUE. dup 5 .hex RESTORE.  3 spaces 10 1024 */ BLUE. .P1 RESTORE. ."  kBytes" 

   CopyGIFtoFramePtrAll
   
\   FreeBitmapBuffer[]
;

\ : $8.hex, ( u -- )   base @ >r  hex  0 <#  BL HOLD [char] , HOLD BL HOLD  # # # # # # # #  [char] $ HOLD #> type  r> base ! ;

: CompileFileTo, ( $filename -- )
   page
   ." create "  2dup type 2 spaces 
   ReadMyFile 
   MyFileBuffer[] MyFileSize @ 
   dup $8.hex, 
   cell 1- +  cell /  0 ?do  
      i $0F and 0= if  cr 3 spaces  then 
      dup @ $8.hex,  cell+
   loop  drop
;

: CompileFile ( $filename -- )   \ read a file into memory at here, starts with a cell containing the file size

   2dup MyFileNameSaved[] place
   R/O OPEN-FILE if  cr RED. ." Could not open file "  RESTORE.  BLUE.  MyFileNameSaved[] count type  RESTORE. space  #10001 throw  then  MyFileHandle !
   
   MyFileHandle @ FILE-SIZE drop
   drop     \ double to single number ( the file size is always < 4Gbytes )
   >r       \ save the file size
   r@  ,    \ compile the file size 
   here r@  allot  r@ MyFileHandle @  READ-FILE  if  
      MyFileHandle @ CLOSE-FILE drop  cr RED. ." Could not read file "  RESTORE.  #10002 throw  
   then   
   MyFileSize !  
   MyFileHandle @ CLOSE-FILE drop
   r> drop
;

: ShowCompiledGIFinfo ( a -- ) 
   FrameBufferInit
   PixelToFrameBuffer
\   -Log NullConsole
   dup @ 
   swap cell+ swap ['] ShowGIF catch if  2drop  then 
 \  RestoreConsole
;

variable vCompiledGIFbuffer

: ShowGIFbuffer ( -- )
\   page
   0 FrameBufferPtr ! 
   PixelToFrameBuffer
   NullConsole
   vCompiledGIFbuffer @ dup @ 
   swap cell+ swap ['] ShowGIF catch if  2drop  then 
   CLOSE-PERSONALITY
;

: ShowCompiledGIF ( a -- ) 
   FrameBufferInit
   PixelToFrameBuffer
   vCompiledGIFbuffer !
   ShowGIFBuffer
;


create Picture1.gif
   PictureFilename1 count CompileFile
create Picture2.gif
   PictureFilename2 count CompileFile

: ShowStartupPicture ( -- )   
   \ Unzip in ShowCompiledGIF always saves to BitmapBufferOriginPtr , so we copy it to BitmapBufferOriginPtr2
\   Picture2.gif ShowCompiledGIF 
\   BitmapBufferOriginPtr @ BitmapBufferOriginPtr2 @  |BitmapBufferNumberOfItems| cells cmove

   Picture1.gif ShowCompiledGIF 
;  

: InitGIFsystem ( -- )
   0 BitmapBufferOriginPtr !
   0 BitmapBufferOriginPtr2 !
   AllocateBitmapBuffer[]
   AllocateBitmapBuffer2[]   
;

: DeinitGIF ( -- )   FreeBitmapBuffer[] ;

\ *****************************************************************************
\ Tests
\ *****************************************************************************

\ parses the given filename  as a GIF file
: ShowGIFfile ( $-- )  
   page  GIFversion  
   cr ." Reading file "  BLUE. 2dup type RESTORE.
   ReadMyFile MyFileBuffer[] MyFileSize @ ShowGIF
;

\\

: ttGIF1 ( -- )   s" Inventio.gif"          ShowGIFfile ;
: ttGIF2 ( -- )   s" SinarAPpicture.gif"    ShowGIFfile ;
: ttGIF3 ( -- )   s" Test1_lzw.gif"         ShowGIFfile ;
: ttGIF4 ( -- )   s" TestPhotoFromJPEG.gif" ShowGIFfile ;

: ttAll ( -- )
   ttGIF1  ekey #27 = if  exit  then 
   ttGIF2  ekey #27 = if  exit  then 
   ttGIF3  ekey #27 = if  exit  then 
   ttGIF4  ekey #27 = if  exit  then 
;

\\

\ *****************************************************************************
\ More tests
\ *****************************************************************************

create MyTestFile ( -- a ) 
\ File length #  string0 #   A B C D E  string1 #   1  2  3  4  5 string2 #   a  b  c   null terminator
    $11 nc,  10         04  41 42 43 45        05  31 32 33 34 35        03  61 62 63   00

: ttGetByte ( -- ) 
   MyTestFile count SourceFileInit
   cr ." Getting all bytes from the list of counted strings : " 
\      begin  GET-BYTE dup if  swap 2 .hex  space  then  0= until 
      begin  GET-BYTE dup if  swap emit  then  0= until 
;

: ttGifTable ( -- )   
   GIF_Table[] |GIF_TableNumberOfEntries| dump 
;

: ttCBITS ( -- )
   0 #GetCodeBits ! 
   counter
   #1000000 0 do  GetCodeBits drop  loop 
   counter swap - ." GetCodeBits time decimal = "  4 .dml  ."  ms "
   ."  for decimal " #GetCodeBits @ 4 .dml  ."  times" 
;

: ttGetCode ( -- )   GET-CODE if  5 .hex  else  ." 0 !!! "  then ;

: ttCodeStack ( -- )
   cr  #10 0 do  i dup CodeStackPush .  loop
   cr  CodeStackPtr @ 0 ?do CodeStackPop .  loop
;

: GetScreenWidth ( -- u )   SCREEN @ + w@ ;
: GetScreenHeight ( -- u )   SCREEN @ + 2+ w@ ;

\ *****************************************************************************
\ Bit buffer tests
\ *****************************************************************************

: .binary ( u c -- )  base @ >r  2 base ! u.r  r> base ! ;

: .binaryLSB ( u c -- )   0 ?do  dup $00000001 and if  ." 1" else  ." 0"  then  1 rshift  loop  drop ;

: .binaryLSB+ ( u c -- )   0 ?do  
   i CURRENT-CODE-BITS @ mod 0= if  space  then 
   dup $00000001 and if  ." 1" else  ." 0"  then  1 rshift  loop  drop 
;

variable SavedBitBuffer[]

: (ttGetCodeBits) ( c -- )  
   BitBufferInit  SavedBitBuffer[] @ BitBuffer[] !
   dup CURRENT-CODE-BITS !  
   2** CURRENT-CODE-MASK !
   cr ." Binary LSB first          :"  BitBuffer[] @ $20 .binaryLSB+ 
\   cr ." Getting "  CURRENT-CODE-BITS @ . ."  bits at a time : "  4 0 do  GetCodeBits 4 .hex  space  loop
   cr ." Getting "  CURRENT-CODE-BITS @ 2 u.r ."  bits at a time : "  
   $20 CURRENT-CODE-BITS @ 1- + CURRENT-CODE-BITS @ / 1 max  $20 min 
\   1+    \ show one more - this will wrap round to the start of the BitBuffer[]
    0 ?do  GetCodeBits CURRENT-CODE-BITS @ .binaryLSB  space  loop  
;

: ttGetCodeBits ( -- )  
   BitBufferInit  $88 BitBufferC! $77 BitBufferC!  $33 BitBufferC!  $55 BitBufferC!  
   BitBuffer[] @ SavedBitBuffer[] !
   cr ." BitBuffer[] has been loaded with " BitBuffer[] 4 over + swap do  i c@ 2 .hex  space  loop 
   cr ." Binary MSB first          : "  BitBuffer[] @ $20 .binary 
   $21 $01 do  i (ttGetCodeBits)  loop 
;

