\ GraphicsPrimitives.f from polyForth32 (~1994) 2016 Mar 04

: >pixel ( x y -- a )   DRAWING_SIZE_X * + cells  FramePtr @ + ;   

: HorizontalLine ( x y length -- )    
   >r  >pixel  r>
   DRAWING_SIZE_X min 0 ?do  
      vDrawingForegoundColour @ over !
      cell+
   loop  
   drop
;

: VerticalLine ( x y length -- )    
   >r  >pixel  r>
   DRAWING_SIZE_X min 0 ?do  
      vDrawingForegoundColour @ over !
      DRAWING_SIZE_X cells +
   loop  
   drop
;

: HorizontalLineOR ( x y n -- )    \ r g b
   >r  >pixel  r>
   DRAWING_SIZE_X min 0 ?do  
      vDrawingForegoundColour @ over +OR 
      cell+
   loop  
   drop
;

: VerticalLineOR ( x y n -- )    
   >r  >pixel  r>
   DRAWING_SIZE_X min 0 ?do  
      vDrawingForegoundColour @ over +OR
      DRAWING_SIZE_X cells +
   loop  
   drop
;

2variable BoxOrigin
BoxOrigin constant BoxOriginY
BoxOrigin cell+ constant BoxOriginX 

2variable BoxSize
BoxSize constant BoxSizeY
BoxSize cell+ constant BoxSizeX

\ Note : the 1- in the code below is to set the origin for the line to the correct point
\ Imagine a single pixel '*' at (1,1) and a 3 x 3 pixel box ( + + + ... )  starting at (0,0)
\ The horizontal length = BoxSizeY = 3 , the start of the lower horizontal line is at (2,0) 
\   0 1 2 
\ 0 + + +
\ 1 + * +
\ 2 + + +
: Box ( OrgX OrgY SizeX SizeY rgb -- )
   vDrawingForegoundColour !
   BoxSize 2!
   BoxOrigin 2!

   BoxOrigin 2@                                    BoxSizeX @ HorizontalLine  
   BoxOrigin 2@ BoxSizeY @ + 1- 0 max              BoxSizeX @ HorizontalLine  
   BoxOrigin 2@                                    BoxSizeY @ VerticalLine    
   BoxOriginX @ BoxSizeX @ + 1- 0 max BoxOriginY @ BoxSizeY @ VerticalLine    
;

: Rectangle ( OrgX OrgY SizeX SizeY rgb -- )
   vDrawingForegoundColour !
   BoxSize 2!
   BoxOrigin 2!
   BoxSizeY @ 0 ?do 
      BoxOrigin 2@ i + BoxSizeX @ HorizontalLine  
   loop
;


\ *****************************************************************************
\ Line drawing algorithm                        1/26/01 Brad Eckert
\ *****************************************************************************

{ 
\ This is probably a variation of the classic Bresenham line drawing algorithm.
\ Tested in Win32forth.

anew teststuff
-1 1 rshift invert 0 2constant wround   \ 0x0000.8000 for 16-bit Forth

variable ch
: plot          ( x y -- ) at-xy ch @ emit ;

\ ---------------------------------------------------------------------------
\ First, let's define the inner loop since this is where the work gets done.
\ We'll step one point at a time in one direction and a fraction of a point
\ in the other direction. The variable slope part uses a double-wide
\ accumulator and we use the upper part of the accumulator to plot the point.
\ In a real application, these would be coded in assembly.

: RoundPoint    ( d -- d int )                 \ round to nearest point
                2dup wround d+ nip ;

: ShallowLine   ( slope 0 Y0 X1+1 X0 -- )      \ slope is between -1 and 1
        do      RoundPoint i swap plot 2over d+
        loop    4drop ;

: SteepLine     ( slope 0 X0 Y1+1 Y0 -- )      \ 1/slope is between -1 and 1
        do      RoundPoint i plot  2over d+
        loop    4drop ;

\ Now for the main word, where we look at the slope and decide which type of line to draw.

variable Point0X  
variable Point0Y
variable Point1X  
variable Point1Y

: Prescale      ( dy*2^cellsize dx -- slope 0 0 )
\ Compute the slope of lines used by the line drawers.
                2dup xor >r
                over 0< if >r dnegate r> then
                dup  0< if negate then
                um/mod nip 0
                r> 0< if dnegate then 0 ;

: Line { x0 y0 x1 y1 \ dx dy -- ]
\ Locals make our job much easier here. You could substitute VALUEs 
\ if your Forth doesn't support this popular local variable syntax.
\ i.e. 0 value x0 0 value y0 ... to y1 to x1 to y0 to x0 ...
                x0 x1 - dup to dx  abs
                y0 y1 - dup to dy  abs -
                dup 0< 0=
        if      drop
                0 dy dx Prescale
                dx 0< if y0 else y1 then
                x0 x1 dx 0< if swap then
                swap 1+ swap			
                ShallowLine exit
        then
        if      0 dx dy Prescale
        else    0 dx dy = if 1 else -1 then 0 \ exact diagonal
        then    dy 0< if x0 else y0 then
                y0 y1 dy 0< if swap then
                swap 1+ swap
                SteepLine
                ;

}

\ *****************************************************************************
\ Bresenham straight line 
\ Uses WinX and WinY for current pixel coordinate
\ *****************************************************************************

variable direction1 
variable direction2 
variable delta1   
variable delta2   

: plot ( x y -- )    \ plot a single pixel in the current colour, at position (x,y)
\    >h 2dup cr . . 
   >pixel  \ dup 8 .hex  
\    >t
   vDrawingForegoundColour @ swap !
;

: ttplot2 ( -- )
   >t $FF0000FF vDrawingForegoundColour !  page >h
;
: ttplot ( -- )
   174 708  plot key drop  \ 5FAEB10
   174 709  plot key drop  \ 5FAEB14
   173 710  plot key drop  \ 5FADB18
   172 711  plot key drop  \ 5FACB1C
   171 712  plot key drop  \ 5FABB20
   171 713  plot key drop  \ 5FABB24
;

: asd ( -- )
   begin
      
      $FF00FF00 $61A72B0 !
      100 ms 
      $FFFF0000 $61A72B0 !
      100 ms
   ekey? until
;

: ttplot3 ( colour -- )
   >r
   r@ $5FAEB10 !
   r@ $5FAEB14 !
   r@ $5FADB18 !
   r@ $5FACB1C !
   r@ $5FABB20 !
   r@ $5FABB24 !
   r> drop
;

: lineRelative ( dx dy -- )   \ draw a line from the current graphics cursor of length (dx,dy)
   over abs 2*   
   over abs 2*   
   2dup > dup >r  
   if  delta1 ! delta2 !  else  delta2 ! delta1 !  swap  then  \ set the deltas
   0< 2* 1+ direction1 ! dup 0< 2* 1+ direction2 !             \ set the directions
   dup abs swap                                                \ count                           
   dup 0< not - abs                                            \ error                          
   begin  
      WinX @ WinY @ plot  
      over                               
   while 
      swap 1-  swap delta1 @ +   dup delta2 @ > if  delta2 @ -  direction1 @   else  0  then   
      direction2 @      
      r@ if  swap  then  WinY +!  WinX +!                              
   repeat 
   r> drop 
   2drop 
;  
   
: lineBresenham ( x y -- )   \ draw a line from the current graphics cursor to (x,y)   
   swap WinX @ -  swap WinY @ - lineRelative ; 

\ *****************************************************************************
\ Line tests
\ ---------------------------------------------------------------------------

create ttPoints  
   0   , 0   ,
   0   , 100 ,
   0   , 200 ,
   100 , 200 ,
   200 , 200 ,
   200 , 100 ,
   200 , 0   ,
   100 , 0   ,

: ttPth   7 and 2* cells ttPoints + 2@ 50 + swap 100 + ;

: ttStar ( -- )   \ draw a star
   magenta vDrawingForegoundColour !
   150 150 WinY ! WinX ! 
   8 0 do  i ttPth  WinY ! WinX !  i 3 + ttPth lineBresenham  loop
;

\\ 

\ *****************************************************************************
\ *****************************************************************************
\ Below this line is old polyForth code... not used anymore...
\ *****************************************************************************
\ *****************************************************************************

decimal

\ ******************************************************************************
\  601  Graphics Common Support
\ ******************************************************************************

\ ******************************************************************************
\  606  Common support
\ ******************************************************************************

\ FramePtr  holds the address of the display refresh buffer.

\ PHYSICAL  stores the maximum y and x pixel numbers.  The pixel
\    count for wash is maintained at  PHYSICAL 2 CELLS + . 
2VARIABLE PHYSICAL   CELL ALLOT 

\ #LASTROW  starting offset of the last EGA/VGA row of pels.
0 CONSTANT #LASTROW 

#USER    

\ CP  holds the graphics cursor's current position.
2 CELLS +USER CP     

\ #COLOR  holds the value of the current color to be written to
\    the graphics display when drawing or writing characters. 
\ CELL +USER #COLOR 

\ 'DRAW  points to the current line drawing behavior.
CELL +USER 'DRAW  

\ ?PenOff  is set by  PenOff  and tested by  DRAW .  Allows vectored movement of  CP  without drawing anything
CELL +USER ?PenOff      

\ ******************************************************************************
\  607  Graphics personality
\ ******************************************************************************

\ RECTANGLE  draws a solid rectangle of  delta y  height and
\    delta x  length with the upper left corner at the current
\    position. 
cell +USER 'RECTANGLE  
: xxxRECTANGLE     'RECTANGLE @EXECUTE ; 

\ BOX  draws a box with the upper left corner at  CP . 
cell +USER 'BOX        
: xxxBOX ( dv)           'BOX @EXECUTE ; 


\ ******************************************************************************
\  608  Graphics personality continued
\ ******************************************************************************

\ WRITE  writes a character string at current position using
\    current font.  WRITE  is the graphics equivalent of  TYPE . 
cell +USER 'WRITE      : WRITE ( a n)      'WRITE @EXECUTE ; 

\ OVERSIZE  selects the oversize character mode. 
cell +USER 'OVERSIZE   : OVERSIZE ( v)  'OVERSIZE @EXECUTE ; 

\ STANDARD  selects the faster, normal size character output. 
cell +USER 'STANDARD   : STANDARD       'STANDARD @EXECUTE ; 

\ ACROSS  and  DOWN  select the direction that characters are
\    drawn when using  WRITE . 
cell +USER 'ACROSS     : ACROSS           'ACROSS @EXECUTE ; 
cell +USER 'DOWN       : DOWN               'DOWN @EXECUTE ; 

TO #USER


\ AT  relocates the Current graphics Position. 
: AT ( v -- )   CP 2! ; 


\ +DRAW  provides  y,x  delta drawing; y,x are offsets from  CP .
: DRAW ( v -- )   ?PenOff @ IF  AT  0 ?PenOff !  EXIT  THEN
   'DRAW @EXECUTE ; 

\ PenOff  lifts the drawing "pen".  The next subsequent  DRAW  will
\    move the  CP  but not actually draw the line.
: PenOff   -1 ?PenOff ! ; 

\ +DRAW  provides  y,x  delta drawing; y,x are offsets from  CP .
: +DRAW ( dv)   CP 2@ V+ DRAW ; 

\ ******************************************************************************
\  610  Line Algorithm Initialization
\ ******************************************************************************

\ SLOPE  is a  2VARIABLE  which is set by  ?SLOPE  and contains
\    the ratio of x to y movement between two sets of screen
\    coordinates.  The first cell always contains the larger
\    value, ie. the direction of greater movement.
2VARIABLE SLOPE   

\ ?NEG  is also set by  ?SLOPE  and is actually two one-byte
\    flags.  The first byte determines positive or negative motion
\    in the direction of greater movement, the second in the
\    direction of lesser movement. 
variable ?NEG   

\ X/Y  is a flag set by  ?SLOPE  which determines whether greater
\    motion is in the x or y direction. 
variable X/Y   

\ Edges   holds clipping values for N S E and West pixels.
4 cells Buffer: Edges 

{ 
0 EAX
1 ECX
2 EDX
}

\ ?SLOPE  determines the slope of a line from the current position
\ to the vector coordinate on the stack.  It sets  SLOPE ,  ?NEG
\ and  X/Y .
CODE ?SLOPE ( v -- ) 

   EAX EAX SUB   
   
   0 [EBP] ECX MOV
   CP cell+ [U]  ECX SUB   
   0< IF   ECX NEG  1 # EAX MOV   THEN
   ECX 0 [EBP] MOV
   CP [U]  EBX SUB   0< IF   EBX NEG   1 # AH MOV   THEN
   ECX ECX SUB
   EBX 0 [EBP] CMP   0< IF   2 CELLS # CL MOV    EBX 0 [EBP] XCHG   AH AL XCHG   THEN   
   EAX SHL   EAX SHL   EAX EDX MOV  
   ?NEG EAX ADDR        EDX  0 [EAX] MOV
   X/Y EAX ADDR         CL   0 [EAX] MOV  
   0 [EBP] ECX MOV
   SLOPE EAX ADDR       ECX  0 [EAX] MOV   
   SLOPE CELL+ EAX ADDR EBX  0 [EAX] MOV   
\ 2DROP
   4 [EBP] EBX MOV                  \ refill tos
   8 # EBP ADD                      \ and discard two cells
   RET END-CODE
 
\ ?*/  skips a ratio calculation that would divide by zero
: ?*/ ( n n n -- n )   ?dup IF  */  EXIT  THEN  DROP ; 

\ ******************************************************************************
\  926  14-bit Fixed-point fraction arithmetic
\ ******************************************************************************

\ plus1  returns the value 1.00000000000000 (binary) in 14-bit
\    fraction notation.
#16384 CONSTANT plus1 

\ *.  multiplies two numbers, one of which is a fraction, to
\    return a whole number.
\ : *.  plus1 */ ;
CODE *. ( n f -- n ) 
   0 [EBP] EAX MOV                  \ get n
   EBX IMUL                         \ n1*n2 => edx:eax
   4 # EBP ADD                      \ clean up stack
   plus1 # ECX MOV                  \ 
   ECX IDIV                         \ n1*n2/n3
   EAX EBX MOV                      \ quotient in tos
   RET END-CODE

\ /.  divides two whole numbers to give a result expressed as
\    a fraction.  The divisor must be positive.  To convert
\    an integer to a fraction, use:  plus1 /. .
\ : /.  plus1 swap */ ;
CODE /. ( n n -- f ) 
   0 [EBP] ECX MOV                  \ get n2
   plus1 # EAX MOV                  \ divisor
   ECX IMUL                         \ n1*n2 => edx:eax
   4 # EBP ADD                      \ clean up stack
   EBX IDIV                         \ n1*n2/n3
   EAX EBX MOV                      \ quotient in tos
   RET END-CODE


\ .F  types out a fraction with four decimal places.
: .F ( f)   10000 *.  dup ABS 0  <# # # # #  [CHAR] . HOLD
   # SIGN  #> TYPE SPACE ; 

\ ******************************************************************************
\  932  Trigonometry for 14-bit fractions
\ ******************************************************************************

\ Add 16 bytes to the structure, using the 16 hex ASCII words following, then ignore the rest of the line
\ This allows captured packet data to be copied and pasted...
: xxx16c,   base @ >r  hex 
   #16 0 do   bl word  count number  c,  loop 
   r> base ! 
   postpone \ 
;  

create triangle[]
\ ' TRIANGLE HEX 10 - 30 DUMP
\ 16nc, 08 54 52 49 41 4E 47 4C 45 20 20 20 A4 34 06 00  ^TRIANGLE   4^^
 16c, 58 66 D1 E0 66 D1 E0 98 79 02 F7 D8 2D 00 40 00  Xffy^-^@^
 16c, 00 F7 D8 50 AD 97 FF 27 42 C1 80 44 74 F8 05 00  ^P'B-Dt^^ ok
{ 
triangle[] dasm
47EFC4   EAX POP                        58
47EFC5   AX SHL                         66D1E0
47EFC8   AX SHL                         66D1E0
47EFCB   CBW                            98
47EFCC   47EFD0 JNS                     7902
47EFCE   EAX NEG                        F7D8
47EFD0   4000 # EAX SUB                 2D00400000
47EFD5   EAX NEG                        F7D8
47EFD7   EAX PUSH                       50
47EFD8   LODSD                          AD
}

CODE TRIANGLE ( a -- f )   
   EBX SHL   EBX SHL   
   CBW
   0< IF   EBX NEG   THEN   plus1 # EBX SUB   
   EBX NEG   
   RET END-CODE

\ COS  computes a polynomial approximation of cosine, using constants from Hart
\ The polynomial is factored to minimize multiplications
: COS ( a -- f)   
   TRIANGLE  dup dup *.  dup #-71 *.
      #1301 + OVER *.  #-10582 + *.  #9352 + OVER *.  + ; 

: SIN ( a -- f)      #4096 - COS ; 

: TAN ( a -- f f)   dup SIN  SWAP COS ; 

: COT ( a -- f f)   TAN SWAP ; 

: CSC ( a -- f f)   SIN plus1 SWAP ; 

: SEC ( a -- f f)   COS plus1 SWAP ; 

: ttsin ( -- )   $400 dup SIN . dup cos . ;


\ ******************************************************************************
\  612  EGA/VGA color support
\ ******************************************************************************
{ 
HEX   $A0000 FramePtr !                  

\ !REG  writes  b2  to port  a , then  b1  to port  a+1 .  Used to
\    control the EGA port-mapped registers.
: !REG ( b1 b2 a -- )   
\   2 POP   0 POP   v OUT   2 INC   0 POP   v OUT   RET  END-CODE
   2DROP DROP 
;
 
\ !GPX  GPX register control defining word:
: !GPX ( b)   CONSTANT  DOES> ( n)   @ $3CE !REG ; 

\ +SET  sets color and enables color display.
1 !GPX +SET  

\ *COLOR  sets color match register 
2 !GPX *COLOR   

\ LOGIC  sets logic register. 
3 !GPX LOGIC   

\ RMODE  sets read mode for  @PEL  and  PAINT . 
5 !GPX RMODE   
}

\ [M]  defines draw logic modes.
: [M] ( b)   CONSTANT  DOES>   @ drop ; 

\ NORM ,  AND'D ,  OR'D  and  XOR'D  set the logic mode of the EGA
\    card when performing writes to the screen memory. 
0 [M] NORM   8 [M] AND'D   10 [M] OR'D   18 [M] XOR'D

\ mask  selects pixel to be altered.                              
\ alpha  resets the EGA/VGA card to enable BIOS character output. 

{ 
ASSEMBLER DEFINITIONS                                           
: mask   3CE # 2 MOV   8 #b 0 MOV   h v OUT ;                   
: rest   3CE # 2 MOV   FF08 #h 0 MOV   h v OUT ;                
FORTH DEFINITIONS                                               
}
\ mask  selects pixel to be altered.
: mask ;
: rest ;

\ alpha  resets the EGA/VGA card to enable BIOS character output.
CODE alpha   rest   
   RET  END-CODE


( Color Selection) 
\ SetDrawingColours  sets a new foreground and background colours and saves it in  vDrawingForegoundColour 

: SetDrawingColours ( rgbForeground rgbBackground -- )   
   vDrawingForegoundColour 2! 
;                                        

0 >RGB  SetDrawingColours                                                                
     
\ WASH  fills the entire display with the current background colour
: xxxWASH ( rgb -- )
   BGR>RGB
   FramePtr @
   DRAWING_SIZE_X DRAWING_SIZE_Y * cells 0 do
      2dup i + !
   cell +loop
   2drop
;

\ WASH2  fills the entire display with the current background colour
: xxxWASH2 ( -- )
\   vDrawingForegoundColour 2@ drop    \ Background colour
   FramePtr @
   DRAWING_SIZE_X DRAWING_SIZE_Y * cells 0 do
\       i $FF and >RGB drop 8 lshift over i + !
       $FFFFFF Random over i + !
   cell +loop
   drop
;

\ ******************************************************************************
\  613  Color Selection
\ ******************************************************************************

\ COLOR:  color defining word whose execution behavior is to call
\    SetDrawingColours  with the value placed on the stack during compile
\    time.
: COLOR: ( c -- )   CONSTANT  DOES>  @ SetDrawingColours ; 

\ DARK  changes all pixels to the DOS standard background color.
: DARK    vDrawingForegoundColour 2@  15 >RGB SetDrawingColours WASH  SetDrawingColours ; 

\ LIGHT  changes all pixels to the DOS standard foreground color.
: LIGHT   vDrawingForegoundColour 2@   0 >RGB SetDrawingColours WASH  SetDrawingColours ; 

\ ******************************************************************************
\  614  Pel plotting
\ ******************************************************************************
{ 
\ (PEL)  is a subroutine that takes the y coordinate in the W
\    register and the x coordinate in the 0 register and returns
\    the bit mask in the 0 register, the bit number in the 1
\    register and the screen address (relative to the DOS segment)
\    in the W register.
LABEL (PEL)   
   0 0 OR   0< IF   0 0 SUB   ELSE   0 PHYSICAL CMP 
   0< IF   PHYSICAL 0 MOV   THEN THEN   W W OR   0< IF   W W SUB
   THEN   W PHYSICAL CELL+ CMP   0< IF   PHYSICAL CELL+ W MOV   
   THEN   W SHL   W 1 MOV   W SHL   W SHL   1 W ADD   W 3 # SHL 
   b 0 1 MOV   0 3 # SHR   0 W ADD   7 #b 1 AND   $8080 # 0 MOV  
   h v 0 ROR   FramePtr W ADD   RET   
   
\ !PELS  stores  n  color values starting at the given vector,
\    retrieving them from the given address.  The pels are placed
\    vertically up from the starting vector.  This reverses the
\    process of retrieving them in  @PELS .
CODE !PELS ( v a n)   2 POP   W POP   0 POP            
   S ) I XCHG   U PUSH   I W XCHG   (PEL) CALL   2 U MOV   mask 
   1 #b 0 MOV   v OUT   2 INC   BEGIN   b LODS   v OUT          
   0 hi W ) FS sg OR   50 # W SUB   U DEC   0= UNTIL            
   U POP   I POP   ' PAUSE 2+ JMP                               

\ @PELS  retrieves  n  color values starting at the given vector,
\    placing them at the given address. 
CODE @PELS ( v a n)   2 POP   W POP   0 POP            
   S ) I XCHG   U PUSH   R PUSH   2 R MOV   I W XCHG            
   (PEL) CALL   I W XCHG   3CE # 2 MOV   5 #h 0 MOV   h v OUT   
   0 DEC   v OUT   2 INC   1 INC   1 PUSH   W PUSH              
   0 0 XOR   R 1 MOV   1NZ IF   REP b STOS   THEN               
   3 # U MOV   W POP   1 POP   BEGIN   R PUSH   I PUSH   W PUSH 
      b U 0 MOV   v OUT   BEGIN   b I ) FS sg 0 MOV   b v 0 RCL 
         b W ) 0 MOV   0 RCL   b STOS   $50 # I SUB   R DEC      
      0= UNTIL   W POP   I POP   R POP   U DEC   0< UNTIL       
   BEGIN   0F #b W ) XOR   W INC   R DEC   0= UNTIL             
   R POP   U POP   I POP   ' PAUSE 2+ JMP   
}

\ PEL  sets a pel at v = (x,y) in the selected color and mode
: PEL ( v -- )   >pixel  vDrawingForegoundColour @  swap ! ; 

\ @PEL  given a display vector returns the current value of
\    that pel from 0 to 15.
: @PEL ( v -- u )   >pixel @ ; 

\ ******************************************************************************
\  616  Line drawing
\ ******************************************************************************

\ ?Out  counts Edges down & clips line when an edge goes negative.
: ?Out  
   \ h Edges 3) DEC   0< IF   1 # 1 MOV   THEN 
;            
                                                                
\ The (E) (W) (N) (S) "South" subroutines are used by  FASTLINE   
\   to locate the adjacent pels when incrementing the line.      
\   They can only be used in a code definition, preceding a CALL 
\   or PIP instruction.                                          
                                                                
( Line drawing)   
\ ASSEMBLER                                     
BEGIN ( W)   \ 0 hi ROL   0 # W SBB   ?Out   RET                  
BEGIN ( E)   \ 0 hi ROR   0 # W ADC   ?Out   RET                  
BEGIN ( N)   \ $80 # W SUB   ?Out   RET                            
BEGIN ( S)   \ $80 # W ADD   ?Out   RET                                                                                          
\ FORTH                                                           

\ 'DIR  contains the addresses of these direction words.
CREATE 'DIR  , , , , 

\ ******************************************************************************
\  617  Fast line drawing
\ ******************************************************************************

\ (LINE)  is the basic word for fast line drawing.  Variations
\    of  DRAW  are more versatile for varying effects with line
\    drawing, while  FASTLINE  is much faster for simple straight
\    line drawing.  It uses a simple incremental algorithm
\    which subtracts the smaller of the two deltas from the larger
\    while incrementing in the direction of greatest movement.
\    When the result of the subtraction goes negative,
\    the larger delta is added back in to the result and
\    the direction of less movement is incremented.
\ CODE (LINE)   
\   CP [U] EAX MOV   CP CELL+ [U] EDI MOV   
\   (PEL) CALL
{ 
   I PUSH   U PUSH   R PUSH   
   SLOPE 1 MOV   1 I MOV   1 INC
   I SHR   SLOPE CELL+ R MOV   ?NEG 2 hi MOV   b X/Y 2 MOV
   BEGIN   
      2 PUSH   mask   2 POP   
      EAX hi W ) FS sg OR   
      3 3 SUB
      b 2 3 MOV   2 hi 3 ADD   
      'DIR 3) PIP   R I SUB
      0< IF   SLOPE I ADD   b 2 3 MOV   2 CELLS #b 3 XOR
         b ?NEG 1+ 3 ADD   'DIR 3) PIP   THEN   LOOP
   R POP   U POP   I POP   
}
\   RET  END-CODE

\ ******************************************************************************
\  651  Textured Line Support
\ ******************************************************************************

\ SPACING  specifies which points along a line will be drawn.
variable SPACING   

\ 'BRUSH  is a vector for drawing the points along a line. 
variable 'BRUSH   

\ Lim  is used for clipping the line within the world window.
2VARIABLE Lim 

\ SLOPE2 ,  X/Y2  and  ?NEG2  are the saved copy of  SLOPE ,  X/Y
\    and  ?NEG  so that they won't be affected by  drawLine .
2VARIABLE SLOPE2    
variable ?NEG2 
variable X/Y2   
 
\ CODE (/Lim) ( -- )  
\   h ?NEG 2 MOV   b 2 0 MOV   b X/Y 0 ADD
\   Edges 0) 1 MOV   1 Lim MOV   2 hi 2 XOR   b 2 0 XOR
\   2 CELLS #b 0 XOR    Edges 0) 0 MOV   0 Lim CELL+ MOV
\   -1 #b 2 hi TEST   0= NOT IF   SLOPE CELL+ NEG   THEN    
\   RET  END-CODE

variable reg0
variable reg1
variable reg2

: (/Lim) ( -- )  
   ?NEG w@  reg2 ! 
   0 reg0 !
   reg2 c@ reg0 c!  
   X/Y c@ reg0 +!
   Edges reg0 @ + @ reg1 !   reg1 @ Lim !   
   
   reg2 w@ dup >< $FF and swap  $FF and xor  reg0 @ xor reg0 !
   reg0 @  2 CELLS  XOR reg0 !   
   Edges reg0 @ + @ reg0 !   reg0 @ Lim CELL+ !
   reg2 w@ >< $FF and  $FF = if  SLOPE CELL+  @ NEGATE  SLOPE CELL+ !  then
;

: ShowVars ( -- )  
   hex
   cr   SLOPE @ 8 u.r space 
   SLOPE @ 8 u.r space    
   SLOPE cell + @ 8 u.r space    
   ?Neg @ 8 u.r space    
   X/Y @ 8 u.r space    
   Edges @ 8 u.r space    
   Edges cell + @ 8 u.r space    
   Edges 2 cells + @ 8 u.r space    
   Edges 3 cells + @ 8 u.r space    
   Lim @ 8 u.r space          \ $20
   Lim cell+ @ 8 u.r space    \ $42000000 
;

: ttlim ( -- )  decimal
   1 ?Neg !  $0203 X/Y !  $10 Edges ! $20 Edges cell + ! $30 Edges 2 cells + ! $40 Edges 3 cells + !
   $50 SLOPE ! $60 SLOPE cell+ !
   0 Lim ! 0 Lim cell+ !
   ShowVars
   (/Lim)
   ShowVars
;

{ 
   ?NEG w@   
   dup $FF and  dup X/Y +!
   Edges + @ dup Lim !   
   ?NEG w@ dup ><  xor  dup >r   b 2 0 XOR
   2 CELLS #b 0 XOR    Edges 0) 0 MOV   0 Lim CELL+ MOV
   r> >< $FF and $FF = IF   
      SLOPE CELL+ @ NEGATE  SLOPE CELL+ !
   THEN    
;
}

\ /LIM  initializes  Lim  from  Edges  and saves a copy of
\    SLOPE ,  X/Y  and  ?NEG  to  SLOPE2 ,  X/Y2  and  ?NEG2
\    respectively. 
: /LIM ( -- )    
   (/Lim)  
   SLOPE 2@ SLOPE2 2!  X/Y C@ X/Y2 C!  ?NEG @ ?NEG2 ! ; 

\ ?Lim  exits the calling definition if the line goes outside
\    the defined world window.
\ CODE ?Lim ( n n -- )   
\   0 POP   1 POP   1 Lim 0) CMP   0< IF
\      7 CELLS # S ADD   3 CELLS # R ADD   ' EXIT JMP   
\   THEN   
\   RET  END-CODE
: ?Lim ( value limOffset -- )
   Lim + @ < if
       7 0 do drop  loop  r> drop  exit  
       \ #8001 throw
   then
;

\ Linear  returns the nth point on a line whose slope has been set
\    by  ?SLOPE .
: Linear ( v n -- v v)   
   dup SLOPE2 2@ ?*/  
   2dup ABS CELL ?Lim
   dup         0 ?Lim 
   DROP  ?NEG2 C@ IF  YNEGATE  THEN
   X/Y2 C@ IF  SWAP  THEN ; 


\ Brush  executes what's in  'BRUSH  for every  SPACINGth  point
\    on the line from  CP  to  v .
: drawBrushLine ( v -- )   
   ?SLOPE  /LIM  CP 2@  
   SLOPE2 @ 0 DO  
      2dup I Linear
      V+  'BRUSH @EXECUTE  
   SPACING @ +LOOP  
   2DROP 
; 

\ FASTLINE  draws a line from the current graphics position. 
: (FASTLINE) ( v -- )   
   /LIM  CP 2@  
   SLOPE2 @ 0 DO 
      cr 2dup swap . . 
      2dup I Linear 
      2dup swap . . 
      V+  2dup swap . . PEL  
   LOOP  
   2DROP 
;

\ FASTLINE  draws a line from the current graphics position. 
: FASTLINE ( v -- )   2dup ?SLOPE  (FASTLINE)  AT ;                  

\ drawLine  same as  FASTLINE  but does not clip to the world window.
: drawLine ( v -- )   Edges 4 CELLS 127 FILL  FASTLINE ; 

\ ******************************************************************************
\  631  Line clipping
\ ******************************************************************************

\ CUT  sets  Edges  to clip the given line during fast draw Line.
CODE CUT ( v -- )   
{
   1 POP   2 POP   (X) 'ORIGIN CELL+ U) 0 MOV
   Edges # W MOV   2 0 SUB   STOS   FRAME CELL+ U) 0 ADD   0 NEG
   STOS   'ORIGIN U) 0 MOV   FRAME U) 0 ADD   1 0 SUB   STOS
   FRAME U) 0 SUB   0 NEG   STOS   
}
   RET  END-CODE
 
\ -IN  determines whether a line falls into a clipping region.
CODE -IN ( -- t)   
{
   Edges # W MOV   0 0 SUB   4 # 1 MOV
   BEGIN   SCAS   0> NOT WHILE   LOOP   0 DEC   THEN
   0 COM   0 PUSH   
}
   RET  END-CODE
 
\ ?CLIP  EXITs the calling definition if neither  CP  nor the
\    vector on the stack lie within the world window.  Otherwise
\    returns the coordinates of the point to draw to and the
\    coordinates for the updated  CP .
: ?CLIP ( v -- v v )   
   CP 2@ CUT  -IN IF  
      2dup CUT -IN IF  AT  R> DROP EXIT  THEN  
      CP 2@  2OVER AT  EXIT  
   THEN  
   2dup 
; 
                                                             
\ -CLIP  initializes  Edges  to ensure no line will reach the clipping limit.
: -CLIP   'DRAW ASSIGN ( v -- )   drawLine ; 

\ +CLIP  sets  'DRAW  to clipped line drawing.
: +CLIP   'DRAW ASSIGN  ?CLIP FASTLINE  AT ; 

\ ******************************************************************************
\  618  Figures
\ ******************************************************************************

2VARIABLE ROWV  CELL ALLOT 

\ DARKEN  darkens a horizontal row given the starting address
\    and mask in  ROWV  and  ROWV CELL+  and the ending address
\    and mask on the stack with the address on top.  DARKEN  is
\    also used in solid figure drawing.
CODE DARKEN ( m a)   
   { 
   1 POP   2 POP   ROWV W MOV
   b ROWV CELL+ 0 MOV   0 hi 0 hi SUB
   BEGIN   0 0 hi OR   b 0 SHR   CY UNTIL   2 hi 2 hi SUB
   BEGIN   2 2 hi OR   b 2 SHL   CY UNTIL   ES PUSH   FS PUSH
   ES POP   W 1 SUB   0< NOT IF   0= IF   2 hi 0 hi AND
      ELSE   2 PUSH   mask   0 hi W ) FS sg OR   W INC   1 DEC
         0> IF   -1 #b 0 hi MOV   mask   -1 #b 0 MOV
            REP b STOS   THEN   0 POP   THEN   mask
      0 hi W ) FS sg OR   THEN   ES POP   
      }
   RET  END-CODE
 
\ ?MASK  calculates offset and pixel mask of given coordinates.
CODE ?MASK ( v -- m a)   
{ 
   0 POP   W POP   (PEL) CALL
   0 PUSH   W PUSH   
}
   RET  END-CODE
 
\ ******************************************************************************
\  620  Character Output
\ ******************************************************************************

\ &CHAR  writes to the V/EGA screen with the upper left corner at
\    cursor location  CP , the character  c  from the character
\    font located at address  a .
CODE &CHAR ( a c -- )   
{ 
   0 POP   W POP   I PUSH   U PUSH   W I XCHG
   CP CELL+ U) W MOV   W SHL   W 1 MOV   W SHL   W SHL   1 W ADD
   W 3 # SHL   CP U) 2 MOV   b 2 1 MOV   2 3 # SHR   2 W ADD
   FramePtr W ADD   7 # 1 AND   (Y) vPixelsPerGlyph U) 3 MOV   b 3 MUL
   0 I ADD   BEGIN   0 0 SUB   b LODS   v h 0 ROR   0 0 hi XCHG
      0 3 hi MOV   mask   0 hi W ) FS sg OR   3 hi 0 hi MOV
      mask   0 hi 1 W) FS sg OR   50 # W ADD   1 #b 3 SUB
   0= UNTIL   U POP   I POP   
}
   RET  END-CODE
 
\ ******************************************************************************
\  623 {+ PAINT-COLORS ( Microsoft Windows Paintbrush colors
\ ******************************************************************************
{ 
 0 COLOR: BLACK       1 COLOR: YELLOW    2 COLOR: MAGENTA    3 COLOR: RED 
 
 4 COLOR: CYAN        5 COLOR: GREEN     6 COLOR: BLUE    7 COLOR: DARK-GRAY 
 
 8 COLOR: WHITE       9 COLOR: DARK-YELLOW   10 COLOR: DARK-MAGENTA 
 
11 COLOR: DARK-RED   12 COLOR: DARK-CYAN     13 COLOR: DARK-GREEN 
 
14 COLOR: DARK-BLUE  15 COLOR: GRAY 
}
 
\ ******************************************************************************
\  624  FILL left
\ ******************************************************************************

\ ?LEFT  scans left along the current line until it finds a pel
\    that is other than the determined inside color.  It then
\    places the address and mask of the pel on the stack.
CODE ?LEFT ( m a --  m a )   
{ 
   W POP   0 POP   0 hi 0 hi SUB
   b W ) FS sg 2 MOV   2 COM   BEGIN   b 0 SHL   CY NOT WHILE
      b 2 0 TEST   0= NOT UNTIL   NEVER IF   SWAP THEN   W DEC
      -1 # 0 MOV   STD   ES PUSH   FS PUSH   ES POP
      BEGIN   b SCAS   0= NOT UNTIL   ES POP   CLD   W INC
      b W ) FS sg 0 MOV   0 COM   1 1 SUB   h 0 1 BSF
   1 # 0 MOV   v b 0 ROL   THEN   b 0 ROR   CY IF   W INC
   THEN   FramePtr W CMP   U< IF   FramePtr W MOV   80 #b 0 MOV
   THEN   0 PUSH   W PUSH   
}
   RET  END-CODE
 
\ ******************************************************************************
\  625  Check right
\ ******************************************************************************

\ ?RIGHT  scans right along the current line until it finds a pel
\    that is other than the determined inside color.  It then
\    places the address and mask of the pel on the stack.
CODE ?RIGHT ( m a -- m a )   
{ 
   W POP   0 POP   0 hi 0 hi SUB
   b W ) FS sg 2 MOV   2 COM   BEGIN   b 0 SHR   CY NOT WHILE
      b 2 0 TEST   0= NOT UNTIL   NEVER IF   SWAP THEN   W INC
      -1 # 0 MOV   ES PUSH   FS PUSH   ES POP
      BEGIN   b SCAS   0= NOT UNTIL   ES POP   W DEC
      b W ) FS sg 0 MOV   0 COM   1 1 SUB   h 0 1 BSR
   1 # 0 MOV   b v 0 ROL   THEN   b 0 ROL   CY IF   W DEC
   THEN   FramePtr 1 MOV   0FFFF # 1 ADD   W 1 CMP   U< IF
      1 W MOV   1 #b 0 MOV   THEN   0 PUSH   W PUSH   
}
   RET  END-CODE

 
\ ******************************************************************************
\  626  Check up
\ ******************************************************************************

LABEL UP/DOWN
{ 
   HERE USE    CELL W) 0 MOV
   ROWV W MOV   1 POP   1 ROWV 2 CELLS + MOV   0 W ADD   0 1 ADD
   2 POP   2 2 hi MOV   b ROWV CELL+ 2 MOV   W 1 SUB   BEGIN
      b W ) FS sg 0 MOV   BEGIN   b 2 0 TEST   0= IF   BEGIN 2>R
               1 1 OR   0> NOT IF   2 2 hi CMP   0= NOT IF
               SWAP THEN   b 2 ROR   CY IF   W INC   1 DEC
      >R ROT JMP   R> THEN   SWAP AGAIN   BEGIN   b 2 0 TEST
      2R> 0= NOT UNTIL   ELSE   2 PUSH   W PUSH   THEN
      1 1 OR   0> NOT IF   2 2 hi CMP   0= NOT WHILE   THEN
   b 2 ROR   CY IF   W INC   1 DEC   0> WHILE   ES PUSH
      FS PUSH   ES POP   REP b SCAS   ES POP   0= NOT WHILE
   W DEC   1 INC   THEN   b W ) FS sg 0 MOV   THEN   ROT JMP
   THEN   THEN   THEN   2 hi 2 MOV   2 PUSH
   ROWV 2 CELLS + PUSH   
}
   RET  END-CODE
\ ?UP  checks the row 1 above the current row looking for changes
\    from  -COL  to not  -COL .  When it finds a transition it
\    pushes that address and mask on the stack.  If it finds no
\    "un-painted" pels it leaves nothing on the stack.  This
\    allows a stack-oriented recursive procedure.
CODE ?UP ( m a -- [m' a'...] )   
   -50 ,  UP/DOWN JMP
   
\ ?DOWN  is like  ?UP  but checks 1 below the current row. 
CODE ?DOWN ( m a - [m' a'...])   50 ,  UP/DOWN JMP

\ ******************************************************************************
\  627  Painting closed shapes
\ ******************************************************************************

: #PS ( -- max-parameter-stack )   S0 @ R0 @ - ;

: ?RUNAWAY   DEPTH  #PS CELL/ 1- > ABORT" Shape too complex" ;  

: *COLOR  drop ;
: RMODE   drop ;

\ PAINT  entirely fill an enclosed shape with the current color. 
: PAINT ( v -- )   
   2dup @PEL  dup 15 XOR *COLOR  vDrawingForegoundColour @ =        
   ABORT" already painted"  8 RMODE  -1  ROT ROT  ?MASK         
   BEGIN  
      dup 1+ 
   WHILE  
      2dup  ?LEFT  ROWV 2!  ?RIGHT            
      dup FramePtr @ #LASTROW BOUNDS WITHIN ABORT" not closed"      
      ?UP  ?DOWN  DARKEN  
      ?RUNAWAY  
   REPEAT  
   DROP ;      

\ ******************************************************************************
\  630  Vectored DRAW
\ ******************************************************************************

\ +SLOPE  increments by one pel on the line which is described
\    by the contents of  SLOPE ,  ?NEG  and  X/Y .  It leaves three
\    items on the stack, the lowest of which must be returned.
CODE +SLOPE ( s -- s v )  
{ 
   W POP   1 # 1 MOV   2 2 SUB
   ?NEG 0 MOV   b 0 0 OR   0= NOT IF   1 NEG   THEN
   SLOPE CELL+ W SUB   0< IF   
      SLOPE W ADD   2 INC
      0 hi 0 hi OR   0= NOT IF   2 NEG   THEN   
   THEN
   4 #b X/Y TEST   0= IF   1 2 XCHG   THEN
   W PUSH   2 PUSH   1 PUSH   
}
   RET  END-CODE
 

\ ******************************************************************************
\  633  Character Output
\ ******************************************************************************

\ WRITE  is the graphics equivalent of  (TYPE) .
: xxxWRITE ( a n -- )   
   BOUNDS ?do  
      'LOCAL-FONT @ I C@ &CHAR
\      PROP 2@  CP 2@ V+ CP 2!  
   loop 
;

\\ 

variable EmitThis
\ PLACE    is the graphics equivalent of  EMIT .
: EMIT-G ( c -- )   EmitThis c!  EmitThis 1 WRITE DROP ; 

\ SPCS     is the graphics equivalent of  SPACES .
: SPCS ( u)   0 MAX  ?dup IF  0 DO  BL EMIT-G  LOOP  THEN ; 

\ SCRUB  is used to erase text fields.  CP  is preserved.
\    Eg:  : .quote   1 ?R@ COUNT  dup BLACK SCRUB  WHITE WRITE ;
\    The version below the  \\  is much faster but,
\    specific for  ACROSS  text only.
: SCRUB ( n)   CP 2@  ROT 0 MAX  ?dup IF  0 DO  219 EMIT-G  LOOP
   THEN  AT ; 

\ GMSG     is the graphics equivalent of  MSG .  Its string is
\    terminated with  \ .
: GMSG   CREATE  [CHAR] \ STRING  DOES>  COUNT WRITE ; 

\ ******************************************************************************
\  634  Oversize Fonts
\ ******************************************************************************
{ 
\ .SCAN  displays one scan row of the character as described in
\    the mask byte at  a  for  n  bits.  The row is drawn with a
\    rectangle of size *FONT for each character bit.  Bits are
\    displayed high=left.
: .SCAN ( a n -- a' n)   
   >R COUNT  R@ 0 DO  
      dup 128 AND IF  *FONT 2@ RECTANGLE  THEN  
      CP 2@ *FONT @ + AT  
      2*  
   LOOP
   DROP R> ; 

\ .CHAR  displays ASCII character  c  from the currently selected
\    font table with the upper left corner of the character at
\    CP .  The size of the character is the size of the selected
\    font times the x,y values in  *FONT .
: .CHAR ( c -- )   
   >R  
   CP 2@  2dup  vPixelsPerGlyphX () @ - 2@  OVER R> *
   'LOCAL-FONT @ + SWAP  ROT 0 DO  
      .SCAN 2>R  *FONT CELL+ @ 0 V+
      2dup AT  2R>  
   LOOP  
   2DROP 2DROP  
   AT 
; 
}

\ ******************************************************************************
\  636  World Coordinate Scaling
\ ******************************************************************************

\ SetOrigin  sets the lower-left corner of the world window to  v  in
\    screen coordinates.
: SetOrigin ( v -- )   (X) 'ORIGIN 2! ; 

\ SetSize  sets the size of the world window to  delta v  in screen
\    coordinates.
: SetSize ( dv -- )   YNEGATE  (X) FRAME 2! ; 

\ RANGE  sets the selected axis' world coordinate scale.
: RANGE ( l h)   OVER - REACH !  LOW ! ; 

\ PLACES  sets the number of decimal places in the selected axis.
: PLACES ( n)   #PLACES ! ; 

: *dSCALE ( n - n)   FRAME @  REACH @ ?*/ ; 

\ *SCALE  and  *dSCALE  translate world and world delta scalar
\    into screen and screen delta scalar, respectively for the
\    selected axis.
: *SCALE ( n - n)   LOW @ -  *dSCALE  'ORIGIN @ + ; 

\ dSCALE  and  SCALE  use  *dSCALE  and  *SCALE  to translate
\    coordinate pairs.
: dSCALE ( v - v')   SWAP (Y) *dSCALE  SWAP (X) *dSCALE ; 

\ *SCALE  and  *dSCALE  translate world and world delta scalar
\    into screen and screen delta scalar, respectively for the
\    selected axis.

\ dSCALE  and  SCALE  use  *dSCALE  and  *SCALE  to translate
\    coordinate pairs.
: SCALE ( v - v')   SWAP (Y) *SCALE   SWAP (X) *SCALE ; 

\ *DRAW  and  *+DRAW  draw in the world window using world
\    coordinates.
: *DRAW ( v)   SCALE DRAW ; 

: *+DRAW ( v)   dSCALE +DRAW ; 



\ ******************************************************************************
\  639  Plot Formats
\ ******************************************************************************

#USER   CELL +USER 'MARK-DATA   CELL +USER 'FETCH-DATA 
 
\ TIP  holds the thickness of certain brushed lines.
        CELL +USER TIP          TO #USER 

\ MARK-DATA  is used to plot one point on a graph.
: MARK-DATA ( n)   'MARK-DATA @EXECUTE ; 

\ FETCH-DATA  is used to put one data point onto the stack.
: FETCH-DATA ( i - n)   'FETCH-DATA @EXECUTE ; 

\ LINES   vectors  MARK-DATA  to plot a line graph.
: LINES   'MARK-DATA ASSIGN ( v)  *DRAW ; 

\ HISTOGRAM   vectors  MARK-DATA  to plot a histogram with the
\    leading line corresponding to the data point.

\ +HISTOGRAM  vectors  MARK-DATA  to plot a histogram with the
\    center corresponding to the data point.
: HISTOGRAM   'MARK-DATA ASSIGN ( v)  SCALE  CP 2@ DROP
   OVER DRAW DRAW ; 

\ +HISTOGRAM  vectors  MARK-DATA  to plot a histogram with the
\    center corresponding to the data point.
: +HISTOGRAM   'MARK-DATA ASSIGN ( v)  SCALE  2dup dup
   CP 2@ SWAP  >R -  2/ -  R> OVER DRAW DRAW DRAW ; 

\ ******************************************************************************
\  640  More Plot Formats
\ ******************************************************************************

\ BARS  vectors  MARK-DATA  to plot a bar graph.  The value in
\    TIP  is the width of the bars.  The left side of the bar is
\    the actual x data position.
: BARS   'MARK-DATA ASSIGN ( v)   SCALE 2dup AT  1+ CUT
   -IN NOT IF  (Y) 'ORIGIN @  0 *SCALE MIN  CP 2@ DROP -
      TIP @ RECTANGLE  THEN ; 

\ RULE  works like  DRAW  but takes two vector coordinates and
\    draws a line between them leaving  CP  at the coordinate
\    that was on top of the stack.
: RULE ( v v2 -- )   2SWAP AT  drawLine ; 

\ .Plus  outputs a + to the graphics screen at the current position.
: .Plus  ( -- )
   CP 2@ 2dup  TIP @ 2/ 2*  dup >R  2/ -  2dup  R@ + RULE  
   R@ 2/ 0 V-  2dup R> 0 V+ RULE 
; 

\ MARKS  vectors  MARK-DATA  to plot +'s on a graph.
: MARKS   2 TIP !  'MARK-DATA ASSIGN ( v)   PenOff *DRAW .Plus ; 

\ ******************************************************************************
\  642  Grids And Ticks
\ ******************************************************************************

\ ORIENT  finds the other coordinate (y if  ()  is x,
\    x if  ()  is y) from the vector variable on the stack.
\    Swaps the arguments if  ()  is set to x.
: ORIENT ( n a - v)   () @ IF  CELL- @  ELSE
      CELL+ @  SWAP  THEN ; 

\ <AXIS>  draws an axis (grid) line at world position  n  across
\    the selected axis.
: <AXIS> ( n)   *SCALE 'ORIGIN ORIENT AT  0 FRAME ORIENT
   +DRAW ; 

\ GRIDS  draws  n  + 1 grid lines within the defined drawing area
\    ( SetSize ) at the defined origin ( SetOrigin ).  Grid lines
\    will be spaced evenly when the number is an even modulo of
\    the pixel frame size.
: GRIDS ( n)   'DRAW @  +CLIP  PenOff  SWAP  dup 1+ -1 DO
      dup REACH @  I ROT */  LOW @ +  <AXIS>  LOOP
   DROP  'DRAW ! ; 

\ NICK  draws a "nick" mark at world position  n  across the
\    selected axis.

\ NICKS  draws  n  + 1 nicks along the selected axis.
: NICK ( n)   'DRAW @  -CLIP  SWAP *SCALE 'ORIGIN ORIENT
   () @  2/ dup  2 XOR  SWAP V+ AT  () @  dup 4 XOR
   SWAP VNEGATE +DRAW  'DRAW ! ; 

\ NICKS  draws  n  + 1 nicks along the selected axis.
: NICKS ( n)   dup 1+ 0 DO  dup REACH @  I ROT */  LOW @ +
      NICK  LOOP  DROP ; 

\ ******************************************************************************
\  643  Graph Labels
\ ******************************************************************************

\ (ADJUST)  centers characters on the x axis and right justifies
\    them on the y axis.  Justification is based on the x and y
\    values of the vector variable,  vPixelsPerGlyph  (pixels per character).
: (ADJUST) ( n -- n )   
   dup  () @ IF  vPixelsPerGlyphY @  2/  SWAP 1+
      vPixelsPerGlyphX @ *  VNEGATE  (Y)  
   ELSE  
      vPixelsPerGlyphX 2@ >R
      SWAP R> -2 */  
   THEN  
   PenOff +DRAW 
; 

\ (LABEL)  does output number conversion inserting a decimal
\    point  #PLACES  left of the least significant digit when
\    #PLACES  is non-zero.
: (LABEL) ( n - a n)   dup ABS  0  <#  #PLACES @ ?dup IF
      0 DO  #  LOOP  [CHAR] . HOLD  THEN  #S  SIGN  #> ; 

\ DrawLabel  draws one label, value  n , at  n .
: DrawLabel ( n)   dup *SCALE  'ORIGIN ORIENT AT  'LABEL @EXECUTE
   (ADJUST) WRITE ; 

\ LABELS  draws  n+1  labels evenly spaced along the selected
\    axis.
: LABELS ( n)   dup 1+ 0 DO  dup  REACH @  I ROT */  LOW @ +
      DrawLabel  LOOP  DROP ; 

\ DECIMAL-LABELS  sets labels to show with decimal places.
: DECIMAL-LABELS   ['] (LABEL) 'LABEL ! ; 

\ ******************************************************************************
\  644  Auto-sized Grids & Axes
\ ******************************************************************************

\ STEP  generates the values for spacing and placing the grid
\    and axes frame on the screen.

\ STEPS  uses  STEP  and returns values to  (GRID)  and  AXIS
\    to draw grids and axes with the proper spacing in
\    world coordinate space.
: STEP   REACH @  10 / 2 SWAP  dup dup ABS /  SWAP 1 DO
      OVER 3 MOD  0= 5 AND  20 +  10 */  1 M+  dup I -  +LOOP
   NEGATE  NIP ; 

\ STEPS  uses  STEP  and returns values to  (GRID)  and  AXIS
\    to draw grids and axes with the proper spacing in
\    world coordinate space.
: STEPS ( - n l h)   STEP >R  LOW @  R@ ABS  OVER REACH @ +
   OVER / *  R@ 0> IF  SWAP  R@ + SWAP  THEN  R>  ROT ROT ; 

\ LIMITS  finds max and min coordinates of graph.
: LIMITS ( - h l)   REACH @  LOW @  2dup < IF  SWAP  THEN ; 

: (GRID)   STEPS DO  I <AXIS>  dup +LOOP  DROP ; 

: AXIS   STEPS DO  I NICK  I DrawLabel  dup +LOOP  DROP ; 

: AXES   (X) AXIS  (Y) AXIS ; 

\ GRID and AXES  attempt to make as nearly as 10 evenly spaced
\    divisions along the axes.  When nothing close is possible,
\    rather than approximate the divisions, they leave the ends
\    open.
: GRID   'DRAW @  +CLIP  (X) (GRID)  (Y) (GRID)  'DRAW ! ; 

\ PAPER  draws an auto-divided grid with x and y axes.
: PAPER   GRID AXES ; 

\ ******************************************************************************
\  939  Normalization
\ ******************************************************************************
{ 
   0 EAX
   1 ECX
   2 EDX
   W EDI
}
\ NORMALIZE  returns a pair of single precision values which
\    maintain the ratio of the passed doubles.
CODE NORMALIZE ( d d -- n n )   
   #63 # ECX MOV   
   BEGIN   
      0 [EBP] SHL   EBX RCL
      OV NOT IF   
         EDI SHL   4 [EBP] RCL   
         OV NOT IF   ROT  LOOP   THEN
      4 [EBP] RCR   EBX SAR   
      ELSE   
         EBX RCR   
      THEN   
   8 # EBP ADD                      \ and clean up stack
   RET  END-CODE
 
\ ******************************************************************************
\  645  2-D Figures
\ ******************************************************************************

\ *AT  is a version of  AT  scaled to world coordinates.
: *AT ( v)   SCALE AT ; 

\ |PEL|  is a form of  PEL  that is clipped to the visible world window.
: |PEL| ( v)   2dup CUT -IN IF  2DROP EXIT  THEN  PEL ; 

\ *PEL  is a version of  PEL  scaled to world coordinates.
: *PEL ( v)   SCALE |PEL| ; 

\ UDEC  moves the given number one toward zero, negative numbers count up, positive numbers count down.
CODE UDEC ( n -- n )   
   EBX EAX MOV    EAX EAX OR   0< IF   EAX INC
   ELSE   1 # EAX SUB   0 # EAX ADC   THEN   EAX EBX MOV
   RET  END-CODE
 
\ *BOX  is a version of BOX scaled to world coordinates.
: *BOX ( dv)   dSCALE BOX ; 

\ *RECTANGLE  draws a solid rectangle of  delta y  height and
\    delta x  length (world coordinate) with the upper left corner
\    at the current position.
: *RECTANGLE ( dv)   dSCALE RECTANGLE ; 

\ ******************************************************************************
\  646  Ellipse
\ ******************************************************************************

variable x    variable y     
variable dX   variable dY 
variable XX   variable 2XX   variable YY   variable 2YY 
variable 'SET4    : SET4   'SET4 @EXECUTE ; 
 
\ /ELL  initializes the slope variables and returns the decision value.
: /ELL ( v -- n)   
   0 x !  0 dX !  
   dup *  dup XX !  
   dup + 2XX !
   dup >R  y !  
   R@ dup *  dup YY !  
   dup + 2YY !  
   2XX @  R@ *  dY !  
   YY @  XX @  
   R> * -  
   XX @ 4 /  + 
; 

\ QUADS  accepts a delta vector and returns the screen coordinates
\    for each quadrant.

CODE QUADS ( v -- v1 v2 v3 v4 )  
   0 [EBP] EDX MOV         \ y delta , x is in EBX
   3 2 * 1+ cells # EBP SUB  \ make room for 3.5 new vectors on the EBP stack, 
   \ This gives four output vectors at offsets (x,y) 0,4 8,12 16,20 24,28
   \ 3.5 is 4 less the y input value already on the [EBP] stack
   CP [U] EAX MOV          \ x origin
   EAX #24 [EBP] MOV       \ v1.x origin 
   EBX #24 [EBP] ADD       \ + v1.x delta 
   EAX #16 [EBP] MOV       \ v2.x origin 
   EBX #16 [EBP] SUB       \ + v2.x delta 
   EAX #08 [EBP] MOV       \ v3.x origin 
   EBX #08 [EBP] SUB       \ + v3.x delta 
   EAX #00 [EBP] MOV       \ v4.x origin 
   EBX #00 [EBP] ADD       \ + v4.x delta   
   CP cell+ [U] EAX MOV    \ y origin
   EAX #28 [EBP] MOV       \ v1.y origin 
   EDX #28 [EBP] ADD       \ + v1.y delta 
   EAX #20 [EBP] MOV       \ v2.y origin 
   EDX #20 [EBP] ADD       \ + v2.y delta 
   EAX #12 [EBP] MOV       \ v3.y origin 
   EDX #12 [EBP] SUB       \ + v3.y delta 
   EAX #04 [EBP] MOV       \ v4.y origin 
   EDX #04 [EBP] SUB       \ + v4.y delta   
   POP(EBX)                \ move the top [EBP] item into EBX = Top of Stack register
   RET  END-CODE
{ 
: QUADS-HL ( v -- v1 v2 v3 v4 )
   2>R 
   CP 2@ 2R@ V+
   CP 2@ 2R@ NEGATE V+
   CP 2@ 2R@ V-
   CP 2@ 2R> SWAP NEGATE SWAP V+
;

: ttq ( -- )   decimal
   cr ." High Level : "  #20 #30 CP 2! #15 #20 QUADS-HL 8 0 do  4 u.r  loop 
   cr ." Low Level  : "  #20 #30 CP 2! #15 #20 QUADS 8 0 do  4 u.r  loop 
;
}

\ -SOLID  sets  ELLIPSE  to write 4 pels for the mirror image.
: -SOLID ( -- )   'SET4 ASSIGN  y @ x @ QUADS  |PEL|  |PEL|  |PEL|  |PEL| ; 

\ SOLID  sets  ELLIPSE  to write 2 lines for the mirror image. 
: SOLID ( -- )   'SET4 ASSIGN  CP 2@  y @ x @ QUADS  CP 2!  DRAW  CP 2! DRAW  CP 2! ; 

\ ******************************************************************************
\  647  Ellipse/Circle
\ ******************************************************************************

{ 
   0 EAX
   1 ECX
   2 EDX
   W EDI
}

\ xADJ  and  yADJ  maintain the amount of angular motion along the
\    plotted path.
: xADJ ( n -- n )   
   1 x +!
   dx @  2YY @ + dup dx !
   +
;

: yADJ ( n -- n )   
   -1 y +!
   dy @ 2XX @ - dup dy ! 
   -
;

{
: ttadj ( -- )
   decimal
   #20 2XX !  #10 2YY !  #100 dX ! #120 dY ! 
   #1000 xADJ . \ 1110
   #2000 yADJ . \ 1900
;
}

\ ELLIPSE  expects the  y  and  x  semi-axis parameters as screen coordinates.                                                 
\ The ellipse drawn will be closed and may be  PAINTed .          
\ The  SET4  function does overlap 1 pel at the quadrant       
\ junction (only a problem if  X'ORing).       
: ELLIPSE ( y x -- )   
   /ELL  
   BEGIN  
      dX @ dY @  < 
   WHILE  
      SET4     
      0 OVER < IF  yADJ  THEN  xADJ  YY @ +  
   REPEAT             
   XX @ YY @ -  3 2 */  dX @ dY @ + -  2 /  +                   
   BEGIN  
      y @ 0< NOT 
   WHILE  
      SET4  dup 0< IF  xADJ  THEN         
      yADJ  XX @ +  
   REPEAT  
   DROP 
;                                                                                           

: *ELLIPSE ( y x)   dSCALE ELLIPSE ; 

\ CIRCLE  is passed the radius as world coordinates.  This allows
\    the adjustment of the aspect ratio for screen dimensions.
: CIRCLE ( r)   dup dSCALE  YNEGATE ELLIPSE ; 

\ ******************************************************************************
\  648  Arcs using sine and cosine
\ ******************************************************************************

\ ORG-ARC  is a variable used in  ARC  to maintain the current
\    center position of the arc.
#USER   

2 CELLS +USER ORG-ARC   
CELL +USER RAD-ARC   

TO #USER 

\ RAD-ARC  holds the radius of a circle. 

\ RADIUS  sets  RAD-ARC  to radius n in world coordinates.
: RADIUS ( n -- )   RAD-ARC ! ; 

\ DEGREE  converts an angle in degrees to a 31 bit fractional
\    angle such as is expected by  ARC .
: DEGREE ( a --  f)   360 /. ; 

\ Radial  expects a 31 bit fractional angle and a length and
\    leaves the offset from the current position for a radial
\    of that length and direction.
: Radial ( f n -- v )   
   >R  dup SIN  SWAP COS  R> plus1  V*/ 
; 

\ ARC  draws an arc from the low angle to the high angle on
\    the stack with radius  RAD-ARC  in world coordinates.
\    Circles and ellipses can be drawn with  0 360 DEGREE ARC .
\    Different elliptical ratios can be achieved by adjusting the
\    x or y axis of the world coordinate window.
: ARC ( l h)   2048 RAD-ARC @  (Y) *dSCALE ABS /  CP 2@ 2dup
   ORG-ARC 2!  PenOff  2>R SWAP ROT DO  I RAD-ARC @ Radial dSCALE
      ORG-ARC 2@ V+ DRAW  dup +LOOP  DROP  2R> AT ; 

\ ******************************************************************************
\  652  Textured Line Types
\ ******************************************************************************

\ BRUSH  ASSIGNs  Brush  to  'DRAW .
: BRUSH   'DRAW ASSIGN ( v)   ?CLIP drawBrushLine AT ; 

\ THICK  sets  'BRUSH  to draw lines of thickness  n .
: THICK ( n)   TIP !  BRUSH  'BRUSH ASSIGN ( v)
   TIP @ 2/  dup V+ AT  TIP @ dup RECTANGLE ; 

\ O'S  sets  'BRUSH  to draw lines with circles of radius  n .
: O'S ( n)   TIP !  BRUSH  'BRUSH ASSIGN ( v)   AT
   TIP @ CIRCLE ; 

\ PLUSES  sets  'BRUSH  to draw lines with the "+" character.
: PLUSES ( n)   TIP !  BRUSH  'BRUSH ASSIGN ( v)   AT .Plus ; 

\ ******************************************************************************
\  653  Graph Plotting
\ ******************************************************************************

: 1-fill ( v)   dup >R  AT  'ORIGIN CELL+ @  R> drawLine ; 

\ FILLED  is used in graph drawing.  It fills from a point on a
\    line to either 0 in world coordinates or  (Y) LOW , depending
\    on whether or not there is a zero-crossing in  (Y) RANGE .
: FILLED ( n)   SPACING !  BRUSH  'BRUSH ASSIGN ( v)
   dup  CP @ SPACING @  2dup + >R  OVER + WITHIN IF
      R@ SWAP - SLOPE2 2@  X/Y2 C@ 0= IF  SWAP  THEN  ?*/
      ?NEG2 C@ IF  NEGATE  THEN  +  R> 1-fill
   ELSE  R> DROP 2DROP  THEN ; 

\ GRAPH  is used to output a table of data to the screen as a
\    graph of any of the supported types.  It will plot one point
\    of data for each table entry, continuing to  (X) RANGE
\    point.  No data in the table should exceed the upper or lower
\    limits of  (Y) RANGE .  GRAPH  plots the data values
\    vertically.
: GRAPH ( ...)   (X) 'ORIGIN 2@ AT  LIMITS DO  I FETCH-DATA
      I MARK-DATA  LOOP ; 

\ CELL-DATA  assumes that an address is on the stack that contains
\    CELL  sized data.  That address is removed when the last data
\    element is retrieved.
: CELL-DATA   'FETCH-DATA ASSIGN ( a i - a n | n)   dup CELLS
   (X) REACH @ 1- ROT = IF  SWAP  ELSE  OVER  THEN  + @ ; 

 
\ ******************************************************************************
\  654  Graphics personality support
\ ******************************************************************************

: +SET   drop ;

\ -ATTR  preserves the current  ATTRIBUTE  across a BIOS call
\    setting the background color to 0 while it executes the
\    given token.
: -ATTR ( a)   ATTRIBUTE @ dup >R  15 AND ATTRIBUTE !
   EXECUTE  R> ATTRIBUTE !  vDrawingForegoundColour @ +SET ; 

\ ******************************************************************************
\  655  Native display output
\ ******************************************************************************


CODE (FASTTYPE) ( a n -- )   
{ 
   L# U) 1 MOV   b CORNER 1+ U) 1 ADD
   50 # 0 MOV   1 MUL   10 # 1 MOV   1 MUL   0 W MOV
   C# U) W ADD   b CORNER U) W ADD   S ) W SUB   FramePtr W ADD
   rest   2 POP   S ) 2 XCHG   I PUSH
   BEGIN   
      (Y) vPixelsPerGlyph U) 0 MOV   0 1 MOV   b 2 ) MUL   2 INC
      'LOCAL-FONT U) 0 ADD   0 I MOV   W PUSH
      BEGIN   b LODS   0F #b ATTRIBUTE U) TEST   0= IF
            F0 #b ATTRIBUTE U) TEST   0= IF   0 0 SUB
            ELSE   0 COM   THEN   THEN   b 0 W ) FS sg MOV
         50 W) W LEA   LOOP   W POP   W INC   CELL S) DEC
   0= UNTIL   I POP   CELL S) S LEA   ' PAUSE 2+ JMP   
   (X)
}
   RET  END-CODE
 

\ FASTTYPE  is the graphics version of  (TYPE)  that places the
\    character pixels from the current font onto the graphics
\    screen.  This routine assumes that the font is 8 by 16 pixels
\    on the screen, but the font can be any size.  The VGA/EGA
\    hardware changes all 0 pixels to color index 0, but all 1
\    pixels are changed to the current graphics color.  Thus, both
\    the background and foreground pixels within the character are
\    changed by this routine.  The text cursor location defined by
\    C#  and  L#  is used to position the character and  ATTRIBUTE
\    is used to determine how the character is displayed.  0 in
\    the low nibble invertes the pixels and 0 in both nibbles
\    makes the character invisible.
: FASTTYPE ( a n -- )   
\   BEGIN  
\      WRAP? 
      (FASTTYPE)  
\   WHILE
\      +WRAP  
\   REPEAT 
; 

\ ******************************************************************************
\  657  Single Screen Development
\ ******************************************************************************
#USER 
CELL +USER TOP
CELL +USER CORNER
CELL +USER BOTTOM
TO #USER

\ #LINES  sets the size of the text window to n lines.
: #LINES ( n -- )   
   >R  
   C# 2@  CORNER 2C@ V-  25 R@ -  0 MAX
   0 CORNER 2C!  R> 1- BOTTOM 1+ C!  SWAP AT-XY  
   0 0 TOP 2C! 
; 


\ SC  gives a 3 line text window at the bottom of the screen.
: SC   3 #LINES ; 

\ -SC  gives the whole screen to text.
: -SC   25 #LINES ; 

\ ******************************************************************************
\  660  EGA/VGA Configuration
\ ******************************************************************************

\ CONFIG  defines the E/VGA display initialization words.
: CONFIG   CREATE  DOES>   
   >BODY dup 2@
   2dup 1 1 V- PHYSICAL 2!  2dup 20 */ PHYSICAL 2 CELLS + !
   0 ROT 1- 8 M*/ DROP ['] #LASTROW !
   2 CELLS + @ 20 0 >pixel  
   \ SC-GRAPHICS  
   ['] FASTTYPE 'TYPE !
;

\ ******************************************************************************
\  603  Initialization
\ ******************************************************************************

( Initialization)   
\ : SMALL  Font5x8x1  FontSet ;
: SMALL  Font5x7x1  FontSet ;
: xxxBIG    Font8x16x1 FontSet ;

CONFIG /VGA   480 640 , ,  18 , 

: INITIAL   BIG  STANDARD  LINES  1 THICK  1 SPACING !  +CLIP
   -SOLID  (Y) DECIMAL-LABELS  (X) DECIMAL-LABELS  CELL-DATA ; 

: /GRAPHICS  
   0 >RGB SetDrawingColours 
   479 0 SetOrigin  479 639 SetSize 
   0 750 (Y) RANGE  0 1000 (X) RANGE  /VGA 
   INITIAL 
; 

\\

\ ******************************************************************************
\  663  Kbd Cursor movement
\ ******************************************************************************

\ ?mark  is the drawing flag.  0= yes, draw.  Also selects cursor.
variable ?mark   variable Tlast   variable REPS 

\ Tlast  is the saved time of the last operation. 

\ REPS  is incremented up to a step limit or reset based upon the
\    delay since the last keypress. 

 
HERE  $80 C, $78 C, $64 C, $52 C, $48 C, $24 C,            
      $F8 C, $C0 C, $A0 C, $90 C, $08 C, $04 C,  

\ POINTER  displays the cursor.  Two types are supported.
\    Move the  HERE  to use the other type.
: POINTER   XOR'D  vPixelsPerGlyphY @  6 vPixelsPerGlyphX !
   LITERAL  ?mark  @ &CHAR  vPixelsPerGlyphX !  NORM ; 

\ |draw|  clips to the screen size and draws to the delta vector.
: |draw| ( dv)   CP 2@ V+  0 MAX >R  0 MAX
   PHYSICAL 2@ >R  MIN  2R> MIN  DRAW ; 

\ Travel  moves the cursor about the screen.
: Travel ( dv)   ?mark @ ?PenOff !  |draw| ; 

\ FASTER  returns a flag representative of the repeat key window.
: FASTER ( - t)   COUNTER  Tlast @  OVER Tlast !  - 6 < ; 

\ 2GO  returns a step count based upon the elapsed time since last
\    called.  This step provides acceleration and single pel
\    moves.
: 2GO ( -- n )   REPS @  FASTER IF  1+  BL MIN  ELSE  DROP 2
   THEN  dup REPS !  2/ ; 

\ ******************************************************************************
\  664  Motion control
\ ******************************************************************************

\ action  is the linked list entry pointer.
variable action 

\ KEY-DOES  adds an entry to the  action  list and compiles the
\    execution.
: KEY-DOES ( a)   action >LINK C,  ] ; 

\ ACT  traverses the linked list, executing the matched function
\    or  BELL  as default.
: ACT ( k)   action  BEGIN  @ ?dup WHILE  2dup CELL+ C@ = UNTIL
   CELL+ 1+ CALL  ELSE  BELL  THEN  DROP ; 

\ DENOTE  displays the current cursor position.
: DENOTE   alpha  C# 2@  CP 2@  SWAP 5 U.R  5 U.R  SWAP AT-XY ; 

\ FNKEY  returns a value representative of the key pressed.  This
\    value is not unique for ASCII keys, but is sufficient here.
: FNKEY ( - k)   KEY  ?dup 0= IF  'KEY 1+ C@  THEN ; 

\ CURSOR  is the entry execution for this utility.  The cursor and
\    pen will be the currently selected color.  Therefore the
\    phrase  GREEN CURSOR  will provide the desired contrast.
: CURSOR   -CLIP  1 ?mark !  0 REPS !  BEGIN  POINTER
      FNKEY  POINTER  ACT  DENOTE  AGAIN ;   

\ ******************************************************************************
\  666  Screen buffer support
\ ******************************************************************************

\ SCRBUF  holds the address of the current screen buffer.
#USER    CELL +USER SCRBUF   TO #USER 

\ BUFSIZE  size of each screen buffer.  There is room for one
\    screen's pixels plus one cell to indicate who it belongs to.
640 480 * CELL+ CONSTANT BUFSIZE 

\ SCREEN-BUFFERS  holds all the screen buffers.
CREATE SCREEN-BUFFERS  BUFSIZE /ALLOT 

\ GET-SCREEN  finds a free screen buffer.  Will abort if one
\    is not available.
: GET-SCREEN   0  1 0 DO  SCREEN-BUFFERS I
      BUFSIZE * +  dup @ 0= IF  NIP  LEAVE  ELSE  DROP  THEN
   LOOP  dup 0= ABORT" Insufficient screen allocation!"
   dup CELL+ SCRBUF !  STATUS SWAP ! ; 

\ FREE-SCREEN  releases a screen buffer.
: FREE-SCREEN   0 SCRBUF @ CELL- ! ; 

\ ******************************************************************************
\  671  Color chart
\ ******************************************************************************

BIG  vPixelsPerGlyphX 2@  3 * 2CONSTANT cRECT 

: .BOX ( n)   vDrawingForegoundColour @  SWAP >RGB SetDrawingColours  cRECT RECTANGLE
   WHITE cRECT BOX  PenOff cRECT 0 SWAP +DRAW  DROP  >RGB SetDrawingColours ; 

: COLORS   cRECT 0 -2 V*  PHYSICAL @ + AT
   16 0 DO  CP 2@  I .BOX  I (.) 3 OVER - SPCS  WRITE
      cRECT 0= V+ AT  LOOP ; 


\ Inventio Software Forth conversion program
\ V1.0 11 Oct 2015 06:57  www.inventio.co.uk

\ ******************************************************************************
\  673  Graphic Demos
\ ******************************************************************************

\ ******************************************************************************
\  674  Miscellaneous
\ ******************************************************************************


\ DATA  an array used for graphics demos.
-? CREATE DATA   32 CELLS ALLOT 


\ RANDOMIZE  fills a table with random numbers.  It will fill
\    (X) REACH @  number of table elements with values ranging
\    from  (Y) LOW @  to  (Y) LOW @  +  (Y) REACH @ .  It is used
\    to supply dummy data for demo graphs.
: RANDOMIZE ( a)   (Y) REACH @  LOW @  ROT  (X) REACH @ CELLS
   OVER + 1+ SWAP DO  OVER RANDOM  OVER + I !  CELL +LOOP
   2DROP ; 


\ MIX  picks a random color.
: MIX   16 RANDOM >RGB SetDrawingColours ; 


\ 1SEC , 2SEC , and 4SEC  provide time delays.
: 1SEC   1000 MS ; 


: 2SEC   1SEC 1SEC ; 


: 4SEC   2SEC 2SEC ; 


: *X ( n - n)   PHYSICAL @  1000 */ ; 


: *Y ( n - n)   PHYSICAL CELL+ @  1000 */ ; 


: /KPELS ( v - v)   *X >R  *Y R> ; 


\ ******************************************************************************
\  675  VGA Setup Parameters
\ ******************************************************************************

( VGA Setup Parameters)                                         
HERE ( *)                                                       
( Origin          Size         y range     x range     Decimals 
 x      y       x      y       l     h     l     h      x   y ) 
125 ,  851 ,   814 ,  724 ,   -4 ,   8 ,   0 ,   24 ,   0 , 0 , 
117 ,  730 ,   845 ,  622 ,    0 , 200 ,   0 ,   30 ,   0 , 1 , 
 90 ,  750 ,   890 ,  145 ,    0 ,  10 ,   0 ,   30 ,   2 , 1 , 
164 ,  626 ,   672 ,  250 ,    0 ,  10 ,   0 ,   10 ,   0 , 0 , 
156 ,  620 ,   625 ,  600 ,    0 ,  50 ,   0 ,   30 ,   1 , 2 , 
  0 , 1000 ,  1000 , 1000 ,    0 , 479 ,   0 ,  639 ,   0 , 0 , 
  0 ,  835 ,  1000 ,  835 ,    0 , 400 ,   0 ,  639 ,   0 , 0 , 
                                                                
\ SETUP  allows setup of graph drawing parameters for a number of
\    different graphs from a pre-defined table.  Each entry
\    contains 40 bytes of parameters for physical origin, size and
\    numerical ranges.  Typing  n SETUP  chooses the desired line
\    of the table and sets the parameters to the tabled values,
\    thus simplifying the graph drawing process.
: SETUP ( n -- )   
   10 CELLS *  ( *) LITERAL +  10 CELLS BOUNDS DO
      I 2@  2 CELLS +LOOP  (X) #PLACES 2!  SWAP RANGE
   SWAP (Y) RANGE  /KPELS SetSize  /KPELS SetOrigin ; 


\ ******************************************************************************
\  678  Graph demo with random contents
\ ******************************************************************************


\ 3-GRAPHS  draws three small graphs with various modes of line
\    drawing used for plotting the values.  The first graph is
\    drawn with  SETUP  number 2, and axes labeled with  SMALL
\    characters.  DATA  is filled with  RANDOMIZEd  data ;  GRAPH
\    takes the data from  DATA  and plots it.  The second and
\    third graphs are similar but with different  ORIGINs ,
\    colors, and plot modes.
: 3-GRAPHS   2 SETUP DARK  Font8X16x1 FontSet ACROSS  DATA RANDOMIZE
   WHITE  10 dup (X) GRIDS LABELS  5 dup (Y) GRIDS LABELS
   YELLOW LINES  DATA GRAPH  553 110 *AT
   (X) 'ORIGIN 2@  2dup 100 0 V- SetOrigin  DARK-GREEN
   10 dup (X) GRIDS LABELS  5 dup (Y) GRIDS LABELS
   CYAN HISTOGRAM  DATA GRAPH  232 110 *AT  200 0 V- SetOrigin
   RED  10 dup (X) GRIDS LABELS  5 dup (Y) GRIDS LABELS
   WHITE MARKS  DATA GRAPH  BIG  4SEC ; 


\ GRAPHS  is the overall graph drawing demo.  It selects colors,
\    randomizes  DATA , and plots graphs of varying types.
\    Time delays are included between graphs.
: GRAPHS   1 SETUP DARK  +CLIP  1 SPACING !  Font8x16x1 FontSet ACROSS
   DATA RANDOMIZE  WHITE  GRID AXES  RED LINES  DATA GRAPH
   4SEC  DATA RANDOMIZE  GREEN  PenOff DATA GRAPH  4SEC  DARK
   BLUE  GRID AXES  10 TIP !  MAGENTA BARS  DATA GRAPH
   2SEC  3-GRAPHS  1 FILLED  3-GRAPHS ; 


\ ******************************************************************************
\  679  Sample graphs
\ ******************************************************************************


\ PROFITS  is the array that contains the desired data.
CREATE PROFITS   -3 ,  -1 ,  -4 ,  -2 ,  -1 ,  1 ,  3 ,  4 , 


\ (p&l)  plots a graph from the data stored at  PROFITS .
: (p&l)   0 SETUP  DARK-GRAY WASH  Font8x16x1 FontSet ACROSS  DARK-GREEN
   12 dup dup (X) GRIDS LABELS NICKS
   6 dup dup (Y) GRIDS LABELS NICKS
   BIG ACROSS  2 2 OVERSIZE  10 2 *AT
   YELLOW ." PROJECTED PROFIT & LOSS"  STANDARD  -5 8 *AT
   YELLOW ." Months since start-up"  5 -3 *AT  DOWN
   ." Profits $M"  20 TIP !  BARS  1 SPACING !
   DARK-RED PROFITS GRAPH ; 


: P&L   (p&l)  4SEC ; 

\ ******************************************************************************
\  680  Picture demos
\ ******************************************************************************


\ SCATTER  chooses two random numbers and plots a pel at that
\    location.  The high resolution mode is assumed.
: SCATTER ( n)   0 DO  FRAME 2@ 1+ RANDOM
      SWAP NEGATE 1+ RANDOM  SWAP PEL  LOOP ; 


\ drawSTARS  draws  n  random pels in a square of side  RAD-ARC  about
\    CP .
: drawSTARS ( n -- )   
   0 DO  
      RAD-ARC @  dup RANDOM  SWAP 2/ -
      RAD-ARC @ 2*  dup RANDOM  SWAP 2/ -  CP 2@ V+  PEL  
   LOOP 
; 


\ (galaxy)  draws multiple randomized squares of 30  drawSTARS  in a
\    spiral of increasing  RAD-ARC , then  SCATTERs  300 "stars"
\    over the entire picture.
: (galaxy)   1 RADIUS  750 0 DO  CP 2@  I 6 * 360 MOD DEGREE
      RAD-ARC @ Radial  PenOff +DRAW  30 drawSTARS  I I' I - 30 /
      1+ MOD  0= ABS RAD-ARC +!  AT  LOOP  300 SCATTER ; 


\ GALAXY  sets the color and mode, calls  (galaxy) , and delays
\    4 seconds.
: GALAXY   5 SETUP BLACK WASH  250 320 *AT WHITE (galaxy)  4SEC ; 


\ ******************************************************************************
\  681  More graphics demos
\ ******************************************************************************


\ BUBBLES  draws circles of random sizes, locations and colors.
: BUBBLES ( n)   0 DO  DARK  100 0 DO  MIX  50 RANDOM 30 +
      500 RANDOM  639 RANDOM  *AT  CIRCLE  LOOP  LOOP ; 


\ MONDRIAN  is similar to  BUBBLES  but draws thin rectangles.
: MONDRIAN ( n)   0 DO  MIX  479 RANDOM  639 RANDOM *AT
      100 RANDOM  100 RANDOM  *RECTANGLE  LOOP ; 


\ SCRIBBLE  draws random lines covering the entire screen.
: SCRIBBLE ( n)   PHYSICAL 2@  250 320 *AT  ROT 0 DO  MIX
      OVER RANDOM  OVER RANDOM  drawLine  LOOP 2DROP ; 


\ MOIRES  draws circles in high resolution that overlap, thus
\    producing moire patterns.
: MOIRES   300 130 *AT  6 1 DO  100  dup 0 DO  I CIRCLE  2 +LOOP
      4 - 0 SWAP  PenOff *+DRAW  LOOP  90 250 *AT
   BIG ACROSS  ." HIGH RESOLUTION" ; 


\ PATTERNS  runs all of the above routines, setting the proper
\    color, resolution, and delay between each display.
: PATTERNS   5 SETUP  5 BUBBLES  1SEC  DARK  1 THICK
   2000 SCRIBBLE  1SEC DARK  2000 MONDRIAN  1SEC DARK
   YELLOW MOIRES  1SEC ; 


\ ******************************************************************************
\  682  Circular maze drawing
\ ******************************************************************************


\ (MAZE)  draws a circular maze with 20 rings each containing a
\    random missing half-octant.
: (MAZE)   8  dup 20 * OVER + OVER DO  I RADIUS  360 RANDOM
      dup 310 + DEGREE  SWAP DEGREE  SWAP ARC  dup +LOOP
   RAD-ARC +!  0 360 DEGREE ARC ; 


\ MAZE  draws a circular  maze  then fills and unfills it,
\    demonstrating area filling of a complicated closed shape.
: MAZE   5 SETUP DARK  +CLIP  60 200 *AT  WHITE  BIG ACROSS
   ." AREA FILLING OF CLOSED SHAPES"  275 320 *AT  RED (MAZE)
   1SEC  CP 2@ DARK-GREEN PAINT  1SEC  CP 2@ DARK-GRAY PAINT
   2SEC ; 


\ ******************************************************************************
\  683  FORTH, Inc. logo
\ ******************************************************************************

#USER   VECTOR GAP   TO #USER 


\ TAILS  draws a triangle for the ends of the lines of size  n .
: TAILS   GAP @ 0 DO  CP 2@  I I' - 1+ 0 +DRAW  1- AT  LOOP ; 


\ HATCHES  draws the rectangles for the logo.
: HATCHES   GAP @ 2* 0  GAP () @ - 2@ NEGATE  26 1  () @ IF
      SWAP 9 + 2>R VNEGATE 2>R  SWAP NEGATE  2R> 2R>  THEN  V*
   GAP @ 2* dup YNEGATE  5 0 DO  CP 2@ TAILS AT  2OVER RECTANGLE
   2dup PenOff +DRAW  2>R 2OVER V+ 2R>  LOOP  2DROP 2DROP 2DROP ; 


\ 'LOGO  is a table containing the "bits" in the logo.
CREATE 'LOGO  HEX   2 C,  0A C,  9 C,  14 C,  0A C,  DECIMAL 

\ MIDDLE  draws the "bits" in the middle of the logo.
-? : MIDDLE   5 0 DO  (X) CP 2@  I 'LOGO + C@  5 0 DO  dup 16 AND
      IF  GAP 2@ VNEGATE RECTANGLE  THEN  0 GAP @ 2* PenOff +DRAW
      2*  LOOP DROP  (Y) GAP @ 2* 0  V- AT  LOOP ; 

\ LOGO  draws the entire FORTH, Inc. logo of size  n .
: LOGO ( n)   dup GAP 2!  CP 2@  +CLIP HATCHES  GAP 2@  16 -16 V*
   2OVER V+ AT  (Y) HATCHES  (X) GAP 2@ 16 0 V*  V+ AT
   XOR'D MIDDLE NORM ; 


\ ******************************************************************************
\  684  Area filling demo
\ ******************************************************************************


\ CURTAIN  paints behind the logos.
: CURTAIN   CP 2@ PAINT  1SEC ; 


\ 1BORDER  draws a border using O'S .
: 1BORDER   6 O'S  4 SPACING !  5 5 *AT
   (X) REACH 2@ 10 10 V-  2dup RED *BOX  +CLIP WHITE *BOX ; 


\ 2BORDER  draws a border with larger O'S .
: 2BORDER   15 O'S  7 SPACING !  35 25 *AT  430 585 *BOX ; 


\ 3LOGOS  draws three small FORTH, Inc. logos with a size of 5
\    across the screen.
: 3LOGOS ( n)   130 *AT  3 0 DO  CP 2@  4 LOGO  AT  0 184 PenOff
      *+DRAW  LOOP ; 


\ 3LOGOS  draws three small FORTH, Inc. logos with a size of 5
\    across the screen.

\ LOGOS  draws a border and two pairs of  3LOGOS , then uses the
\    area filling routine to erase the figures progressively.  The
\    demo finishes with one large logo.
: LOGOS   6 SETUP DARK  1BORDER DARK-RED  340 3LOGOS
   150 3LOGOS  1SEC  15 15 *AT  DARK-GRAY CURTAIN  DARK-RED
   CURTAIN  DARK-GRAY CURTAIN  GRAY CURTAIN  DARK-RED CURTAIN
   GRAY CURTAIN  110 280 AT  DARK-RED 8 LOGO  2SEC ; 

\ ******************************************************************************
\  687  Line demos
\ ******************************************************************************

VARIABLE 'LTRS 


\ LTRS  and  SQ'S  assigns  'BRUSH  letter and box drawing
\    to the line drawing algorithm.
: LTRS   BRUSH  'BRUSH ASSIGN ( v)   AT  'LTRS C@ PLACE ; 


-? : +LINE ( dv)   Edges 4 CELLS 127 FILL  CP 2@ V+ FASTLINE ; 


: SQ'S ( n)   TIP !  BRUSH  'BRUSH ASSIGN ( v)   CP 2@ 2>R  AT
   TIP @ dup  2dup  0 SWAP +LINE  0 +LINE  VNEGATE
   0 SWAP +LINE  0 +LINE  2R> AT ; 


\ 3BORDER  and  4BORDER  each draw a border using circles but
\    their use of spacing is different resulting in quite
\    different forms.
: 3BORDER   10 O'S  9 SPACING !  120 57 *AT  245 521 *BOX
   6 SPACING !  245 521 *BOX ; 


: 4BORDER   5 O'S  7 SPACING !  145 83 *AT  196 465 *BOX ; 


\ ******************************************************************************
\  688  Line demos
\ ******************************************************************************


: "LINE"   250 210 *AT  ." LINES OF VARIOUS TEXTURES" ; 


: PLOTS   DATA RANDOMIZE PenOff  DATA GRAPH  2SEC  GRAY PenOff
   DATA GRAPH ; 


\ LINE-DEMO  draws a border with various modes of line drawing
\    using circles, rectangles and characters.  It then draws
\    graphs using different plotting modes for the data.
: LINE-DEMO   5 SETUP DARK  BIG ACROSS  YELLOW LINES
   2BORDER  CYAN "LINE" 2SEC  3BORDER  DARK-GREEN 4BORDER
   GRAY  "LINE" 1SEC   3 SETUP  10 SPACING !  5 THICK
   DARK-MAGENTA PLOTS  Font8x16x1 FontSet LTRS  14 'LTRS C!  WHITE PLOTS
   8 SQ'S  RED PLOTS  4 PLUSES  BLUE PLOTS  2SEC ; 


\ ******************************************************************************
\  690  3-D  demos
\ ******************************************************************************


\ TUNNEL  draws concentric rectangles with slightly different
\    placement of the center of each form.
: TUNNEL   DARK-GREEN  300 230 *AT  1 54 DO  I I  3 dup V*
   *BOX  I 3 MOD  dup  PenOff +DRAW  -1 +LOOP ; 


\ CONE  is similar to  TUNNEL  but uses circles.
: CONE   DARK-RED  180 440 *AT  50 2 DO  I 2* CIRCLE
   I 3 MOD  CP +!  LOOP ; 


\ TUBE  draws a series of arcs each with a different
\    aspect ratio, radius and placement.
: TUBE   WHITE  300 30 *AT  52 0 DO  I CIRCLE  PenOff 3 3 +DRAW
   LOOP ; 


\ SHAPES  shows the above forms with the chosen color and delay.
: SHAPES   5 SETUP DARK  +CLIP  TUBE  TUNNEL  CONE  2SEC ; 


\ CIRCLES  draws concentric circles of various colors to show the
\    ability to mix colors and produce various textures.
: CIRCLES   5 SETUP DARK  240 320 *AT  vDrawingForegoundColour @
   DARK-RED  0 120 DO  I CIRCLE  -2 +LOOP
   DARK-GREEN  170 0 DO  I CIRCLE  2 +LOOP
   DARK-GRAY  0 120 DO  I CIRCLE  -2 +LOOP
   170 0 DO  I CIRCLE  4 +LOOP
   170 0 DO  I CIRCLE  2 +LOOP  1SEC  >RGB SetDrawingColours ; 


\ ******************************************************************************
\  691  3-D  Cube
\ ******************************************************************************


#USER   VECTOR SIDE  CELL+   TO #USER 


: (Z)   2 CELLS () ! ; 


: ZSIDE ( v)   CP 2@  2SWAP  PenOff *+DRAW  (Z) SIDE @ 2/ NEGATE
   dup *+DRAW  AT ; 


\ CUBE  draws a three dimensional cube with stack parameters
\    in the following order: height, length and depth.
: CUBE ( y x z)   (Z) SIDE !  (X) SIDE 2!  CP 2@
   SIDE 2@ 1 1 V+  2dup *BOX  (Z) SIDE @ 2/ dup *+DRAW  *BOX
   0 SIDE @ ZSIDE  (X) SIDE 2@ ZSIDE  (Y) SIDE @ 0 ZSIDE  AT ; 


\ >INSIDE  moves  CP  to a given offset 'inside' the previous
\    cube so that a new smaller cube may be produced.
: >INSIDE   (X) SIDE 2@  10 28 V*/  PenOff *+DRAW ; 


\ ******************************************************************************
\  692  3-D  Cube
\ ******************************************************************************

180 CONSTANT DIMENSION 


: SPOT   150 200 *AT ; 


: XYZ ( n - y x z)   dup dup 2/ ; 


\ (CUBES)  draws three cubes one inside the other.
: (CUBES)   SPOT  DIMENSION XYZ  CUBE  >INSIDE  DIMENSION 2/ XYZ
   CUBE  >INSIDE  DIMENSION 2/ 2/ XYZ  CUBE ; 


\ CUBES  draws three  cubes , then draws a small cube starting
\    in the lower left corner of the largest cubes that "grows
\    and consumes" the others.
: CUBES   
   5 SETUP DARK  +CLIP  
   DIMENSION 0 DO  
      SPOT  I XYZ
      WHITE CUBE  DARK-GREEN (CUBES)  SPOT  I XYZ  GRAY CUBE
   5 +LOOP  
   DIMENSION XYZ  WHITE CUBE  1SEC  RED 
   (CUBES)  
   2SEC ; 

\ ******************************************************************************
\  693  Demo-graphic
\ ******************************************************************************


: 1TITLE   5 SETUP DARK  450 60 *AT  BIG ACROSS  BLACK
   2 2 OVERSIZE ." GRAPH PLOTTING SUPPORT FEATURES"  STANDARD
   355 200 *AT  ." * POINT PLOTTIN."
   320 200 *AT  ." * LINE GRAPHS"
   285 200 *AT  ." * HISTOGRAMS"
   250 200 *AT  ." * BAR CHARTS"
   215 200 *AT  ." * FILLED AND TEXTURED GRAPHS"
   180 200 *AT  ." * 2-D WORLD COORDINATE SCALIN."
   145 200 *AT  ." * 2-D VECTOR ARITHMETIC"
   110 200 *AT  ." * HARD COPY PRINTER SUPPORT"
   4SEC 2SEC ; 


: 2TITLE   6 SETUP DARK  300 290 *AT  DARK-RED  8 LOGO  BIG
   ACROSS  51 230 *AT  2 2 OVERSIZE  ." FORTH, Inc."  STANDARD
   4SEC ; 

\ ******************************************************************************
\  695  Clock Demo
\ ******************************************************************************

\ PIVOT  is the center of the clock.
: PIVOT   250 320 *AT ; 


\ .HAND  draws a hand given its angle and length about  PIVOT .
: .HAND ( ang len -- )   
   >R  6 M* 270 M+  360 UM/MOD DROP
   360 SWAP - DEGREE  R> Radial  PIVOT *+DRAW 
; 

\ HAND  creates hands and defines when they should move.
: HAND ( n -- )   
   CREATE , , 0 ,  
   DOES> 
   ( n -- n' )   
   XOR'D
   dup @ >RGB SetDrawingColours   
   CELL+  dup CELL+ @  ?dup IF  OVER @ .HAND  THEN
   2dup @ .HAND  OVER SWAP CELL+ !  
   NORM 
; 

vDrawingForegoundColour 2@ 
   RED        100 HAND SECOND    
   DARK-GREEN 90  HAND MINUTE      
   DARK-GREEN 50  HAND HOUR      
SetDrawingColours   

: .HANDS ( n)   SECOND  60 / MINUTE  12 / HOUR  DROP ; 


: /HANDS   0 ['] SECOND 2 CELLS + !  0 ['] MINUTE 2 CELLS + !
   0 ['] HOUR 2 CELLS + ! ; 


\ ******************************************************************************
\  696  Analog clock
\ ******************************************************************************

VARIABLE 'ALARM 



\ TICKS  draws the hour markings on the clock.
: TICKS   SMALL  15 3 DO  I 30 * DEGREE RAD-ARC @ dup 20 / -
      Radial  vPixelsPerGlyphX 2@ NEGATE  1 2 V*/  V+  PIVOT  PenOff *+DRAW
      I' I - 0 <# #S #> WRITE  LOOP  BIG ; 


\ FACE  draws two concentric circles, then fills the region
\    between them.
: FACE   PIVOT  100 RADIUS  WHITE 0 360 DEGREE ARC
   CP 2@ DARK-GRAY PAINT  90 RADIUS  WHITE TICKS ; 


\ >TIME  sets the starting time of the clock.
: >TIME ( d - n)   DROP  100 /MOD  100 /MOD  60 * +  60 * + ; 


\ WAKEUP  sets what time the alarm will go off.
: WAKEUP ( hh:mm:ss)   >TIME 'ALARM ! ; 


\ CLICK  sounds the bell.
: CLICK   
\   50 MS  71 97 OUTPUT  50 MS  68 97 OUTPUT 
; 


\ .CLOCK  advances the hands (if needed) every second.
: .CLOCK ( n - n)   1000 MS  1+ dup .HANDS ; 


\ ******************************************************************************
\  697  Alarm
\ ******************************************************************************


: +QUAD ( v - v')   SWAP NEGATE ; 


\ QUAD-PTS  draws 4 points about  CP  with the given x and y
\    coordinates.
: QUAD-PTS ( v)   CP 2@ 2SWAP  4 0 DO  2OVER 2OVER V+ PEL  +QUAD
   LOOP  2DROP  AT ; 


\ (ALARM)  draws a 'radioactive' type of symbol used when the
\    'alarm' goes off.
: (ALARM)   360 0 DO  I DEGREE  I 45 + DEGREE ARC  90 +LOOP ; 


\ .ALARM  rings the alarm when the clock reaches  WAKEUP .
: .ALARM   PIVOT  5 0 DO  32 5 DO  I RADIUS  RED (ALARM)
      CLICK  WHITE (ALARM)  4 +LOOP  LOOP ; 


\ /CLOCK  "sets" the clock and draws it.
: /CLOCK ( - n)   /HANDS  FACE  3:59:50 >TIME  dup .HANDS ; 


\ /CLOCK  "sets" the clock and draws it.

\ CLOCK  starts the clock.
: CLOCK ( n)   BEGIN  dup 'ALARM @ < WHILE  .CLOCK  REPEAT
   DROP  .ALARM ; 




\ ******************************************************************************
\  698  Clock demo
\ ******************************************************************************


\ CLOCKS  is a combination of many of this package's capabilities
\    in a `real' application.  It includes the drawing of an
\    analog clock face with hour, minute, and second hands, and
\    an audible alarm.  Details of these functions are in the
\    preceding several blocks.
: CLOCKS   5 SETUP DARK  +CLIP  BLUE WASH  WHITE BIG ACROSS
   470 180 *AT  ." PRESENTING IBM-PC COLOR-GRAPHICS"
   425 305 *AT  ." from"
   375 280 *AT  ." FORTH,Inc."
    75 255 *AT  ." It's about time"
   /CLOCK  .CLOCK .CLOCK .CLOCK .CLOCK  YELLOW
   120 215 *AT  ."             saving"
   115 215 *AT  ."             ______"
    95 220 *AT  ."               V  "   CLOCK 2SEC ; 


\ ******************************************************************************
\  694  Scroll demo
\ ******************************************************************************


: SC-GRAPH   4 SETUP DARK  +CLIP  DARK-GREEN SMALL
   10 dup (X) GRIDS LABELS  10 dup (Y) GRIDS LABELS
   DATA RANDOMIZE  WHITE LINES  DATA GRAPH ; 


: SCROLLS 
\   GR
   4 #LINES  SC-GRAPH  alpha  BIG PAGE
   ." An additional feature of this package"  CR
   ." is a combined text and graphics mode,"  CR
   ." where text can be displayed at the"  CR
   ." same time as graphics."  4SEC 2SEC  PAGE
   ." Any number of lines may be `windowed'"  CR
   ." and will be scrolled without disturbing"  CR
   ." the graphics display."  4SEC 2SEC
   CR  ."  ok"  1SEC  CR  ."  ok"  1SEC  CR  ."  ok"  1SEC
\   CR  -SC  GR  
;

: DEMO   SC  CLOCKS  1TITLE  GALAXY  PATTERNS  MAZE  LOGOS
   SHAPES  LINE-DEMO  CUBES  1TITLE  GRAPHS  P&L  2TITLE
   alpha  -SC  20 0 TOP 2C! ; 

: LDEMO   BEGIN  DEMO  2SEC  AGAIN ; 

\\
: ShowGraph ( -- )

   X_OFFSET Y_OFFSET  1 1 V-  #768  #480 2 2 V+  RED Box
   
   100 0 do 

\   ShowGraphOuterBorder
ekey $18 = if  leave  then
   X_OFFSET Y_OFFSET          #768  #480         WHITE Rectangle
ekey $18 = if  leave  then
   X_OFFSET Y_OFFSET  1 1 V-  #768  #480 2 2 V+  BLACK Box
ekey $18 = if  leave  then
   X_OFFSET Y_OFFSET          #768  #480         BLACK Rectangle
ekey $18 = if  leave  then
   X_OFFSET Y_OFFSET  1 1 V-  #768  #480 2 2 V+  WHITE  Box
\   EKEY? IF  LEAVE  THEN
loop

   exit
   
   GREEN vDrawingForegoundColour !
   X_OFFSET Y_OFFSET BoxOrigin 2!
   #770     #482     BoxSize 2!
   BoxOrigin 2@                           BoxSizeX @ HorizontalLine \ cr >r swap . . r> . 
   BoxOrigin 2@ BoxSizeY @ +              BoxSizeX @ HorizontalLine \ cr >r swap . . r> . 
   BoxOrigin 2@                           BoxSizeY @ VerticalLine   \ cr >r swap . . r> . 
   BoxOriginX @ BoxSizeX @ + BoxOriginY @ BoxSizeY @ VerticalLine   \ cr >r swap . . r> . 
exit
   \ ________X____ ______Y________ length
   X_OFFSET        Y_OFFSET        #770 HorizontalLine
   X_OFFSET        Y_OFFSET #482 + #770 HorizontalLine
   X_OFFSET        Y_OFFSET        #482 VerticalLine  
   X_OFFSET #770 + Y_OFFSET        #482 VerticalLine  
;
