\ LogToFile.f  2016 Jun 10
\ *****************************************************************************
\ Log screen output to a file
\ *****************************************************************************

variable writeFileHandle
variable writeFileHandle2

variable readFileHandle  
variable readFileSize  
variable readFileBufferPtr  

0 writeFileHandle !
0 writeFileHandle2 !

: CloseWriteFile ( -- )
   writeFileHandle @ ?dup if  CLOSE-FILE drop  then
   0 writeFileHandle !
;

\ returns the name of the current folder, excluding path information
: GetFolderName2 ( -- c-addr n )   pad 256 over GetCurrentDirectory  -PATH ;


: AppendDate&TimeToSaveFilename ( -- )

\   InitSaveFilename     \ the basic name

   s" _"  SaveFilename append

\   GetFolderName2     \ the current folder name 
\   dup >r SaveFilename append   \ Note : this string already has '_' at each end
\   18 r> - 0 max  0 ?do  
\      s" _"  SaveFilename append
\   loop 
   @date (date)+ SaveFilename append  
   s" _"SaveFilename append
   @time (time)+ SaveFilename append
;

: SaveFile ( a n -- )

\   writeFileHandle @ ?dup if  CLOSE-FILE drop  then

\   cr ." SaveFile = " SaveFilename count type 

   SaveFilename count R/W CREATE-FILE if  
      drop
      cr  RESTORE. RED.
      ."  CREATE-FILE - cannot create file " SaveFilename count type  
      cr ." ( This is usually because Visual Studio or another application has made the file read-only - please edit the file and save it in VS..."
      RESTORE.
      key drop
      #2021 throw
   else
      writeFileHandle !
   then
   
   writeFileHandle @ WRITE-FILE if  
      RED. cr ." Cannot write to file! "  NORMAL.   #2022 throw  
   then

   writeFileHandle @ ?dup if  CLOSE-FILE drop  then
\   0 writeFileHandle !
;

MAX_PATH Buffer: SaveAsFilename[]

: SaveAsFilename ( a n $ -- )
   2dup SaveAsFilename[] place

\   cr ." SaveAsFilename = " SaveFilename count type 

     ( $ -- ) R/W CREATE-FILE if  
      drop
      cr  RESTORE. RED.
      ." SaveAsFilename :  CREATE-FILE - cannot create file " SaveAsFilename[] count type  
      RESTORE.
      #2022 throw
   else
      writeFileHandle2 !
   then
   
   writeFileHandle2 @ WRITE-FILE if  
      RED. cr ." Cannot write to file! "  NORMAL.  SaveAsFilename[] count type   #2022 throw  
   then

   writeFileHandle2 @ ?dup if  CLOSE-FILE drop  then
\   0 writeFileHandle2 !
;


{ 
\ **************************************
: CloseReadFile
   readFileHandle @ ?dup if  CLOSE-FILE drop  then
;

: AllocateReadFile ( $ -- )
   CloseReadFile
   2dup R/O OPEN-FILE if 
      drop
      cr  ."  AllocateReadFile OPEN-FILE - cannot open file "  type  ."  (Error 2031) "  CloseReadFile #2031 throw
   else
      readFileHandle !
      2drop
   then

   readFileHandle @ FILE-SIZE if  cr ." FILE-SIZE error! " .FileStats CloseReadFile #2032 throw  then  \ Note: FILE-SIZE returns a double
   ( File length double > single ) drop     
   dup 0= if  cr ." Zero FILE-SIZE! " .FileStats   CloseReadFile ( #2003 throw )  then 
   dup readFileSize !
   ALLOCATE if  cr ." ALLOCATE error! " .FileStats CloseReadFile #2034 throw  then  
   readFileBufferPtr !

   readFileBufferPtr @  readFileSize @  readFileHandle @ READ-FILE if  
      cr ." READ-FILE error!" .FileStats CloseReadFile #2035 throw  
   then
   readFileSize @ = not if  cr ." READ_SIZE returned size error! " .FileStats  #2036 throw  then

   CloseReadFile
;

: CompareFile ( c-addr n $ -- f )
   ['] AllocateReadFile catch if 
      2drop 
      cr ." Creating Backup file for the first time..." 
      0        \ the backup file doesn't exist yet, so return 0 meaning need an update, so we will create one
      exit 
   then
   readFileBufferPtr @
   readFileSize @
   compare 0=
   readFileBufferPtr @ free drop
;

: ttcf   s" hello" s" hello.txt" CompareFile . ;
} 

\ : ttsf   s" hello" s" hello.txt" SaveFile ;
\ : ttsf-  s" hellO" s" hello.txt" SaveFile ;


\ *****************************************************************************
\ Log to file
\ *****************************************************************************

variable #LogBuffer

24 2** constant |LogBuffer|	\ must be a power of 2
\ |LogBuffer| Buffer: LogBuffer
variable allocLogBuffer

: LogBuffer  allocLogBuffer @ ;

: InitLogBuffer ( -- ) 
   0 #LogBuffer ! 
   allocLogBuffer @ if  exit  then
   |LogBuffer| allocate abort" Could not allocate LogBuffer!"  allocLogBuffer !
;

: DeInitLogBuffer ( -- ) 
   allocLogBuffer @ 0= if  exit  then   
   allocLogBuffer @ free drop
;

: >LogBuffer ( - a)   LogBuffer  #LogBuffer @ |LogBuffer| 1- and  + ;

: LogEmit ( c)   >LogBuffer c!  #LogBuffer @ 1+  |LogBuffer| 1- and  #LogBuffer ! ;

: LogType ( a n)   over + swap ?do  i c@ LogEmit  loop ;

: LogToFile ( -- )   LogBuffer #LogBuffer @ SaveFile ;

\ *****************************************************************************
\ Vectored I/O to console and log file
\ *****************************************************************************

: EMIT(G)2 ( c)    dup EMIT(G) LogEmit ;

: TYPE(G)2 ( a n)    2dup TYPE(G) LogType ;

: ?TYPE(G)2 ( a n)    2dup ?TYPE(G) LogType ;

: CR(log)   #13 LogEmit  #10 LogEmit ; 

: CR(G)2   CR(G)  CR(log) ;

: PAGE(G)2   PAGE(G)   
   \ CR(log) CR(log)  80 0 do  [char] = LogEmit  loop   CR(log) 
\   s" \   ****** This file was created "  LogType  @DATE (DATE)+ LogType  #32 LogEmit  @TIME (TIME) LogType  CR(log)
;

CREATE LOG-TO-FILE-AND-CONSOLE
   4 CELLS ,             \ datasize
        19 ,             \ maxvector
   PHANDLE ,             \ PHANDLE
         0 ,             \ PREVIOUS
   ' NOOP ,              \ INVOKE    ( -- )
   ' NOOP ,              \ REVOKE    ( -- )
   ' /INPUT(G) ,         \ /INPUT    ( -- )
   ' EMIT(G)2 ,          \ EMIT      ( char -- )
   ' TYPE(G)2 ,          \ TYPE      ( addr len -- )
   ' ?TYPE(G)2 ,         \ ?TYPE     ( addr len -- )
   ' CR(G)2 ,            \ CR        ( -- )
   ' PAGE(G)2 ,          \ PAGE      ( -- )
   ' ATTRIBUTE(G) ,     \ ATTRIBUTE ( n -- )
   ' KEY(G) ,           \ KEY       ( -- char )
   ' KEY?(G) ,          \ KEY?      ( -- flag )
   ' EKEY(G) ,          \ EKEY      ( -- echar )
   ' EKEY?(G) ,         \ EKEY?     ( -- flag )
   ' AKEY(G) ,          \ AKEY      ( -- char )
   ' PUSHTEXT(G) ,      \ PUSHTEXT  ( addr len -- )
   ' AT-XY(G) ,         \ AT-XY     ( x y -- )
   ' GET-XY(G) ,        \ AT-XY?    ( -- x y )
   ' GET-SIZE(G) ,      \ GET-SIZE  ( -- x y )
   ' (ACCEPT) ,         \ ACCEPT    ( addr u1 -- u2 )

: +Log ( -- )   LOG-TO-FILE-AND-CONSOLE OPEN-PERSONALITY ;
: -Log ( -- )   CLOSE-PERSONALITY ;

\ *****************************************************************************
\ no output
\ *****************************************************************************

CREATE NULL-GUI
   4 CELLS ,            \ datasize
        19 ,            \ maxvector
   PHANDLE ,            \ PHANDLE
         0 ,            \ PREVIOUS
   ' NOOP ,             \ INVOKE    ( -- )
   ' NOOP ,             \ REVOKE    ( -- )
   ' /INPUT(G) ,        \ /INPUT    ( -- )
   ' drop ,             \ EMIT      ( char -- )
   ' 2drop ,            \ TYPE      ( addr len -- )
   ' 2drop ,            \ ?TYPE     ( addr len -- )
   ' NOOP ,             \ CR        ( -- )
   ' NOOP ,             \ PAGE      ( -- )
   ' drop ,             \ ATTRIBUTE ( n -- )
   ' KEY(G) ,           \ KEY       ( -- char )
   ' KEY?(G) ,          \ KEY?      ( -- flag )
   ' EKEY(G) ,          \ EKEY      ( -- echar )
   ' EKEY?(G) ,         \ EKEY?     ( -- flag )
   ' AKEY(G) ,          \ AKEY      ( -- char )
   ' PUSHTEXT(G) ,      \ PUSHTEXT  ( addr len -- )
   ' 2drop ,            \ AT-XY     ( x y -- )
   ' GET-XY(G) ,        \ AT-XY?    ( -- x y )
   ' GET-SIZE(G) ,      \ GET-SIZE  ( -- x y )
   ' (ACCEPT) ,         \ ACCEPT    ( addr u1 -- u2 )

: NullConsole ( -- )   NULL-GUI OPEN-PERSONALITY ;

\ *****************************************************************************
\ Vectored I/O to log file only
\ *****************************************************************************

: EMIT(G)0 ( c)    LogEmit ;

: TYPE(G)0 ( a n)    LogType ;

: ?TYPE(G)0 ( a n)    LogType ;

\ : CR(log)   #13 LogEmit  #10 LogEmit ; 

: CR(G)0   CR(log) ;

: PAGE(G)0      
   \ CR(log) CR(log)  80 0 do  [char] = LogEmit  loop   CR(log) 
\   s" \   ****** This file was created "  LogType  @DATE (DATE)+ LogType  #32 LogEmit  @TIME (TIME) LogType  CR(log)
;

CREATE LOG-TO-FILE-ONLY
   4 CELLS ,             \ datasize
        19 ,             \ maxvector
   PHANDLE ,             \ PHANDLE
         0 ,             \ PREVIOUS
   ' NOOP ,              \ INVOKE    ( -- )
   ' NOOP ,              \ REVOKE    ( -- )
   ' /INPUT(G) ,         \ /INPUT    ( -- )
   ' EMIT(G)0 ,          \ EMIT      ( char -- )
   ' TYPE(G)0 ,          \ TYPE      ( addr len -- )
   ' ?TYPE(G)0 ,         \ ?TYPE     ( addr len -- )
   ' CR(G)0 ,            \ CR        ( -- )
   ' PAGE(G)0 ,          \ PAGE      ( -- )
   ' ATTRIBUTE(G) ,     \ ATTRIBUTE ( n -- )
   ' KEY(G) ,           \ KEY       ( -- char )
   ' KEY?(G) ,          \ KEY?      ( -- flag )
   ' EKEY(G) ,          \ EKEY      ( -- echar )
   ' EKEY?(G) ,         \ EKEY?     ( -- flag )
   ' AKEY(G) ,          \ AKEY      ( -- char )
   ' PUSHTEXT(G) ,      \ PUSHTEXT  ( addr len -- )
   ' AT-XY(G) ,         \ AT-XY     ( x y -- )
   ' GET-XY(G) ,        \ AT-XY?    ( -- x y )
   ' GET-SIZE(G) ,      \ GET-SIZE  ( -- x y )
   ' (ACCEPT) ,         \ ACCEPT    ( addr u1 -- u2 )

: +Log0 ( -- )   LOG-TO-FILE-ONLY OPEN-PERSONALITY ;
\ : -Log ( -- )   CLOSE-PERSONALITY ;


\ *****************************************************************************

\ Usage example :

: ttLog
   InitLogBuffer
   cr ." This goes to the screen"
   +Log
   cr ." This goes to the file (and screen)"
   -Log
   cr ." This only goes to the screen again"
   
   cr ." The LogBuffer : "  LogBuffer 40 type 
   LogToFile   \ saves the logged data to the file, using the given filename, plus the date and time
;

\\
: ttlogtype  LogBuffer #LogBuffer @ type ;

: ttll  1000 0 do  s" Hello World     " LogType  loop  ttlog ;



\ *****************************************************************************

