\ ParseCertificate.f  2015 Mar 06

\ The special ASN.1 variable-length encoded length
: ASN.1GetLengthLimited ( -- u ) 
   PayloadDataGet8 dup $80 and if     
      case
      $80 of  -1  endof    \ indefinite length, terminated by and EOC with 0 length ( two 0's )
      $81 of  PayloadDataGet8  endof
      $82 of  PayloadDataGet8 8 lshift PayloadDataGet8 or  endof
      $83 of  PayloadDataGet8 8 lshift PayloadDataGet8 or  8 lshift PayloadDataGet8 or  endof
         cr  RED.  ." Illegal AXDR String length value = "  2 .hex   ."   ReadPayloadDataOffset = "  ReadPayloadDataOffset @ 4 .hex  RESTORE. 
         quit
      endcase
   then
; 

$100 Buffer: FetchBitStringValue[]

: FetchBitStringValue ( n -- )

   ReadPayloadDataOffset @ >PayloadData

   swap
   dup >r
   32 ToColumn dumpL

   CurrentTag @ $3F and ASN1_BIT_STRING = if 
      \   FetchBitStringValue[] c@ dup $04 =  swap $30 = or not if  RED. cr ." FetchBitStringValue[0] not $04 or $30 !!!"  RESTORE.  quit  then
      r@ $42 = if  1 BitStringLength$42 +!  then
      r@ $45 = if  1 BitStringLength$45 +!  then
      r@ $46 = if  1 BitStringLength$46 +!  then
      r@ $47 = if  1 BitStringLength$47 +!  then
      r@ $48 = if  1 BitStringLength$48 +!  then
      r@ $49 = if  1 BitStringLength$49 +!  then
      1 BitStringLengthTotal +! 
   then

   r> drop 

;

: (GetTagLength) ( -- )
   PayloadDataGet8 CurrentTag !    
   ASN.1GetLengthLimited CurrentTagLength ! 
;

: GetNextTLV ( -- )  
\   cr ." GetNextTLV CurrentTagLength = "  CurrentTagLength @ 4 .hex    ."   at ReadPayloadDataOffset = " ReadPayloadDataOffset @ 4 .hex
   CurrentTagLength @ ReadPayloadDataOffset +!  \ skip the curent Tag's value
   (GetTagLength) CurrentTagLength @ FetchBitStringValue  
;

variable NumberOf_ecdsa-with-SHA256

: GetNextTLVecdsa ( -- )
   NumberOf_ecdsa-with-SHA256 @ 0= if  
      32 ToColumn ."  Yes, we have found the first ecdsa-with-SHA256 "
      1 NumberOf_ecdsa-with-SHA256 !  
   else
      0 NumberOf_ecdsa-with-SHA256 !
   then
   GetNextTLV
;


: GetTagLengthValue ( -- )
   (GetTagLength)  
\   begin  ReadPayloadDataOffset @ SequenceEndTOS >= while  SequenceEndPop drop  repeat

   CurrentTag @ $20 and if    \ ConstructedPacket, no data, just the length of SEQUENCE or SET
\      ReadPayloadDataOffset @ CurrentTagLength @ +  SequenceEndPush 
   else  \ primitive packet, contains real data to check and skip 

      CurrentTag @ $3F and ASN1_OID = if 
 \        ReadPayloadDataOffset @ >PayloadData  CurrentTagLength @  id-ecPublicKey    count compare 0= if  cr ." id-ecPublicKey   " GetNextTLV  then
         ReadPayloadDataOffset @ >PayloadData  CurrentTagLength @  id-at-organizationUnitName(OU)  count compare 0= if   cr ." id-at-organizationUnitName(OU)"    GetNextTLV  then
         ReadPayloadDataOffset @ >PayloadData  CurrentTagLength @  id-at-commonName(CN)            count compare 0= if   cr ." id-at-commonName(CN)          "    GetNextTLV  then   
         ReadPayloadDataOffset @ >PayloadData  CurrentTagLength @  secp256r1/prime256v1(7)         count compare 0= if   cr ." secp256r1/prime256v1(7] "          GetNextTLV  then
         ReadPayloadDataOffset @ >PayloadData  CurrentTagLength @  ecdsa-with-SHA256               count compare 0= if   cr ." ecdsa-with-SHA256       "          GetNextTLVecdsa  then 
         ReadPayloadDataOffset @ >PayloadData  CurrentTagLength @  X509v3_CertificatePolicies      count compare 0= if   cr ." X509v3_CertificatePolicies"        GetNextTLV  then
         ReadPayloadDataOffset @ >PayloadData  CurrentTagLength @  X509v3_AuthorityKeyIdentifier   count compare 0= if   cr ." X509v3_AuthorityKeyIdentifier"     GetNextTLV  then
      then
      CurrentTagLength @ ReadPayloadDataOffset +!  \ skip the value
   then
;

: ShowBitStringFieldSizes ( -- )
   cr ." BitStringLength$42 = " 
                      BitStringLength$42 @ dup .  
      ."   ...$45 = " BitStringLength$45 @ dup . + 
      ."   ...$46 = " BitStringLength$46 @ dup . + 
      ."   ...$47 = " BitStringLength$47 @ dup . + 
      ."   ...$48 = " BitStringLength$48 @ dup . +  
      ."   ...$49 = " BitStringLength$49 @ dup . + 
      ."  Others = "  BitStringLengthTotal @ swap - dup . 
   0= not if RED. cr ." BitStringLength other != 0 !!!"  RESTORE.  quit  then
;

: ParseCertificate ( a n -- )     
\   cr 2dup DumpRelative 
   PayloadDataInit
   dup CertificateSize ! 
   over + swap ?do  i c@ PayloadDataPut8  loop

   cr ." ParseCertificate : "
   
   0 NumberOf_ecdsa-with-SHA256 !
   begin
      GetTagLengthValue 
   ReadPayloadDataOffset @ CertificateSize @ >= until

   cr

   ShowBitStringFieldSizes

   BitStringLength$42    @ TotalBitStringLength$42 +!  
   BitStringLength$45    @ TotalBitStringLength$45 +! 
   BitStringLength$46    @ TotalBitStringLength$46 +! 
   BitStringLength$47    @ TotalBitStringLength$47 +! 
   BitStringLength$48    @ TotalBitStringLength$48 +!  
   BitStringLength$49    @ TotalBitStringLength$49 +!  
   BitStringLengthTotal  @ TotalBitStringLengthTotal +!

   cr ." ParseCertificate done."
;

\\
