\ *****************************************************************************
\ Support.f Howerd Oakford  2021 Aug 11
\ Useful SwiftForth support words
\ *****************************************************************************

decimal 

: 2**   ( c -- u )   \ takes 2 to the power c
   1 swap lshift
;

8 2** constant |AttributeStack|  \ must be a power of 2
|AttributeStack| cells Buffer: AttributeStack[]  
variable AttributeStackPtr
: InitAttributeStack ( -- )
   0 AttributeStackPtr ! AttributeStack[] |AttributeStack| cells erase
;
InitAttributeStack

: >AttributeStackPtr ( n -- a )     |AttributeStack| 1-  and cells  AttributeStack[] + ; 

: AttributePush ( u -- )   1 AttributeStackPtr +!  AttributeStackPtr @ >AttributeStackPtr ! ;
: AttributePop  ( -- u )   AttributeStackPtr @ >AttributeStackPtr @  -1 AttributeStackPtr +! ;
: GetPreviousAttribute ( -- )   AttributeStackPtr @ 1+ >AttributeStackPtr @ ;
: CurrentAttribute ( -- a )   AttributeStackPtr @ >AttributeStackPtr ;

: ?NullConsole ( -- f )   'ATTRIBUTE @ ['] drop = ;

: ATTRIBUTE+ ( c -- ) 
\   ?NullConsole if  drop  exit  then 
   DUP AttributePush ATTRIBUTE
;

: GREEN.    ( -- )   7 ATTRIBUTE+ ;
: YELLOW.   ( -- )   6 ATTRIBUTE+ ;
: MAGENTA.  ( -- )   5 ATTRIBUTE+ ;
: CYAN.     ( -- )   4 ATTRIBUTE+ ;
: BLUE.     ( -- )   3 ATTRIBUTE+ ;
: RED.      ( -- )   2 ATTRIBUTE+ ;
: INVERSE.  ( -- )   1 ATTRIBUTE+ ;
: NORMAL.   ( -- )   0 ATTRIBUTE+ ;
: RESTORE.  ( -- )   AttributePop drop AttributePop ATTRIBUTE+ ;
: PREVIOUS. ( -- )   GetPreviousAttribute ATTRIBUTE+ ;

: -."  \ compiling: ( "ccc<">" -- )  executing: ( -- )
   POSTPONE S" POSTPONE MAGENTA. POSTPONE TYPE  POSTPONE RESTORE. ;  IMMEDIATE

\ *****************************************************************************
\ Number display
\ *****************************************************************************

\ display u as an unsigned hexadecimal number in a c character right justified field
: .hex ( u c -- )   base @ >r  hex  u.r  r> base ! ;
\ these include leading zeros 01  0123  01234567 
: 2.hex ( c -- )   base @ >r  hex  0 <#  # #  #> type  r> base ! ;
: 3.hex ( c -- )   base @ >r  hex  0 <#  # # #  #> type  r> base ! ;
: 4.hex ( w -- )   base @ >r  hex  0 <#  # # # #  #> type  r> base ! ;
: 6.hex ( w -- )   base @ >r  hex  0 <#  # # # # # #  #> type  r> base ! ;
: 8.hex ( u -- )   base @ >r  hex  0 <#  # # # # # # # #  #> type  r> base ! ;
\ these start with a '$' and include leading zeros  $01  $0123  $01234567 
: $2.hex ( c -- )   base @ >r  hex  0 <#  # #  [char] $ HOLD #> type  r> base ! ;
: $3.hex ( c -- )   base @ >r  hex  0 <#  # # # [char] $ HOLD #> type  r> base ! ;
: $4.hex ( w -- )   base @ >r  hex  0 <#  # # # #  [char] $ HOLD #> type  r> base ! ;
: $6.hex ( w -- )   base @ >r  hex  0 <#  # # # # # #  [char] $ HOLD #> type  r> base ! ;
: $8.hex ( u -- )   base @ >r  hex  0 <#  # # # # # # # #  [char] $ HOLD #> type  r> base ! ;
\ these start with a '$' and include leading zeros and end with a space and ','  $01 ,  $0123 ,  $01234567 , 
: $2.hex, ( c -- )   base @ >r  hex  0 <#  BL HOLD [char] , HOLD BL HOLD  # #  [char] $ HOLD #> type  r> base ! ;
: $3.hex, ( c -- )   base @ >r  hex  0 <#  BL HOLD [char] , HOLD BL HOLD  # # #  [char] $ HOLD #> type  r> base ! ;
: $4.hex, ( w -- )   base @ >r  hex  0 <#  BL HOLD [char] , HOLD BL HOLD  # # # #  [char] $ HOLD #> type  r> base ! ;
: $6.hex, ( w -- )   base @ >r  hex  0 <#  BL HOLD [char] , HOLD BL HOLD  # # # # # #  [char] $ HOLD #> type  r> base ! ;
: $8.hex, ( u -- )   base @ >r  hex  0 <#  BL HOLD [char] , HOLD BL HOLD  # # # # # # # #  [char] $ HOLD #> type  r> base ! ;

\ display u as an unsigned decimal number in a c character right justified field
: .dml ( u c -- )   base @ >r  decimal  u.r  r> base ! ;

: .L ( u n -- )   >r  s>d  (D.) dup >r TYPE  2r> - spaces ;    \ Left justifed signed
: U.L ( u n -- )   >r  s>d  (DU.) dup >r TYPE  2r> - spaces ;  \ Left justifed unsigned

: .P0 ( n -- )   S>D SWAP OVER DUP 0< if DNEGATE  then  <#                     #S  ROT SIGN  #> type ;
: (.P1) ( n -- $ )   S>D SWAP OVER DUP 0< if DNEGATE  then  <# #     [char] . HOLD #S  ROT SIGN  #> ;
: .P1 ( n -- )   (.P1) type ;
: .P2 ( n -- )   S>D SWAP OVER DUP 0< if DNEGATE  then  <# # #   [char] . HOLD #S  ROT SIGN  #> type ;
: .P3 ( n -- )   S>D SWAP OVER DUP 0< if DNEGATE  then  <# # # # [char] . HOLD #S  ROT SIGN  #> type ;
: (R.P0) ( n u -- )   >r  S>D SWAP OVER DUP 0< if DNEGATE  then  <#  #S  ROT SIGN  #>  r> over - 0 max spaces ;
: R.P0 ( n u -- )   (R.P0) type ;
: R.P1 ( n u -- )   >r  S>D SWAP OVER DUP 0< if DNEGATE  then  <# #     [char] . HOLD #S  ROT SIGN  #>  r> over - 0 max spaces type ;
: R.P2 ( n u -- )   >r  S>D SWAP OVER DUP 0< if DNEGATE  then  <# # #   [char] . HOLD #S  ROT SIGN  #>  r> over - 0 max spaces type ;
: R.P3 ( n u -- )   >r  S>D SWAP OVER DUP 0< if DNEGATE  then  <# # # # [char] . HOLD #S  ROT SIGN  #>  r> over - 0 max spaces type ;
: L.P1 ( u n -- )   >r  (.P1) dup >r TYPE  2r> - spaces ;    \ Left justifed signed, 1 decimal point

\ include enough but minimum leading zeros $01 , $0123 , $00012345
: .hexStretchy ( u -- )
   dup 0< if  8.hex  exit  then  \ 0< means top bit set in unsigned hex
   dup $100 < if  2.hex  exit  then
   dup $10000 < if  4.hex  exit  then
   8.hex
;

: DIGIT> ( c -- c )   [char] 0 - 0 max  dup 9 > if  #223 and 7 -  then  $0F and ;   \ convert the hex ASCII digit 0 to 9 or A to F or a to z to a 4 bit number
: HexC@ ( a -- c )   dup c@ DIGIT> $10 * swap 1+ c@ DIGIT> or ;                     \ convert the BigEndian hex value byte pair at a to an 8 bit number
: HexW@ ( a -- c )   cr ." please rename HexW@ as HexC@... " HexC@ ;

\ *****************************************************************************
\ Misc
\ *****************************************************************************
1     dup constant BIT0
   2* dup constant BIT1
   2* dup constant BIT2 
   2* dup constant BIT3 
   2* dup constant BIT4 
   2* dup constant BIT5 
   2* dup constant BIT6 
   2* dup constant BIT7 
   2* dup constant BIT8 
   2* dup constant BIT9
   2* dup constant BIT10 
   2* dup constant BIT11 
   2* dup constant BIT12
   2* dup constant BIT13
   2* dup constant BIT14
   2* dup constant BIT15
   2* dup constant BIT16 
   2* dup constant BIT17 
   2* dup constant BIT18 
   2* dup constant BIT19
   2* dup constant BIT20 
   2* dup constant BIT21 
   2* dup constant BIT22
   2* dup constant BIT23
   2* dup constant BIT24
   2* dup constant BIT25
   2* dup constant BIT26
   2* dup constant BIT27 
   2* dup constant BIT28 
   2* dup constant BIT29
   2* dup constant BIT30 
   2* dup constant BIT31 
drop

\ a cariage return and linefeed counted string
create $CRLF ( -- a )   $02 c,  $0D c, $0A c, 

\ convert 21/09/2015 to the number of days since 1st Jan 1900
\ Like M/D/Y but with days and months swapped
: D/M/Y ( ud -- u)   10000 UM/MOD  100 /MOD  swap ROT D-M-Y ;


: ttAttributes ( -- )   
   $100 0 do  
      i $0F and 0= if  cr  then 
      BLUE. i 3 .hex  1 spaces  i RESTORE. attribute+  1234 .  RESTORE.  2 spaces  
   loop 
;

: bell  1000 200 beep drop ;

\ move to the given column number
: ToColumn ( u -- )   get-xy drop - 1 max 200 min  spaces ;

#12 constant TAB_WIDTH 
: ToTab    ( u -- )   TAB_WIDTH * ToColumn ;


\ Move to the next column tab position
: AlignColumn ( u -- )   
   >r  get-xy drop r@ / 1+ r@ * get-xy drop - 1 max r@ min spaces 
   get-xy drop #80 r> - > if  cr  then
;

\ : ttac  40 0 do  9 AlignColumn i . loop ;

variable vTimeSync

: InitTimeSync ( -- )   counter vTimeSync ! ;

: TimeSync ( u -- ) 
   vTimeSync +!
   vTimeSync @  
   BEGIN
      PAUSE  1 Sleep DROP
      DUP EXPIRED
   UNTIL DROP
;

\ *****************************************************************************
\ Convert string to Double Number
\ *****************************************************************************

\ convert the ASCII string $ to a double number, using the current base
: Dnumber ( $ -- d )
\   over c@ [char] , = if  1 /string  then    \ remove a possible leading  ','
\   over c@ [char] , = if  1 /string  then    \ remove a possible second leading  ','
   2dup 2>r NUMBER? case
   0 of 
      cr RED.  ."  Invalid number string : " 2r@ Dump  NORMAL.  0.   ( key drop ) 
      #10023 throw
   endof
   1 of  s>d  endof
   2 of  endof
   cr ." what?"
   endcase
   2r> 2drop
;

\ convert the ASCII string $ to a double number, using base 16
: DnumberHex ( $ -- d )
   base @ >r  hex
   NUMBER? 
   r> base !
   case
   0 of 
      #10024 throw
      \ RED.  ."  Invalid number string : " 2r@ Dumpl  NORMAL.  0.   ( key drop ) 
   endof
   1 of  s>d  endof
   2 of  endof
   cr ." what?"
   endcase
;

\ *****************************************************************************
\ Unix timestamp
\ 
\ Unix0DAY is the Modified Julian Date for the start of Unix time.
\ 
\ Unix@NOW returns time of day in seconds since midnight and modified Julian
\ date (MJD) in days since 01/01/1970.
\ 
\ *****************************************************************************

#60 #60 * #24 * constant SECONDS_PER_DAY

#01/01/1970 M/D/Y CONSTANT Unix0DAY
#01/01/2000 M/D/Y CONSTANT Y2k0Day

Y2k0Day Unix0DAY - SECONDS_PER_DAY * constant 30years    \ number of seconds from 1970 Jan 01 to 2000 Jan 01 = 946684800

: Unix@NOW ( timestamp -- ud u )  SECONDS_PER_DAY /MOD 0 SWAP Unix0DAY + ;
: Unix@DATE ( timestamp -- n )    SECONDS_PER_DAY / Unix0DAY + ;
: Unix@TIME ( timestamp -- ud )   SECONDS_PER_DAY MOD 0 ;

: UnixTIME&DATE ( timestamp -- sec min hour day month year )
   Unix@NOW >R  #60 UM/MOD #60 /MOD  R> Y-DD DM ROT ;

\ 2015Sep04 format 
: (date)+ ( u1 -- c-addr u2)   BASE @ >R  DECIMAL  Y-DD  DM
   swap 
   0 <#  
\   [CHAR] _ HOLD
   # #
   2drop  
   3 * C" JanFebMarAprMayJunJulAugSepOctNovDec" +  3 0 DO  DUP C@ HOLD 1-  LOOP DROP  
   0 # # # #  
   #>  R> BASE ! 
;

\ 2015 Sep 04 format 
: (date_)+ ( u1 -- c-addr u2)   BASE @ >R  DECIMAL  Y-DD  DM
   swap 
   0 <#  
   # #
   BL HOLD
   2drop  
   3 * C" JanFebMarAprMayJunJulAugSepOctNovDec" +  3 0 DO  DUP C@ HOLD 1-  LOOP DROP  
   BL HOLD
   0 # # # #  
   #>  R> BASE ! 
;

: :00+ ( ud1 -- ud2)   DECIMAL  #  6 BASE !  # ( [CHAR] - HOLD ) ;

: (TIME)+ ( ud -- c-addr u)   BASE @ >R  <#  :00+ :00+
   DECIMAL # #  #>  R> BASE ! ;

: tttime  @date (date)+ type  ." _"  @time (time)+ type ;

: @UnixTimeStamp ( -- u )
   @TIME
   @DATE Unix0DAY - SECONDS_PER_DAY * 
   M+
   drop
;

: @Y2kTimeStamp ( -- u )      \ returns the number of seconds since 2000 Jan 01 00:00
   @TIME
   @DATE Y2k0Day - SECONDS_PER_DAY * 
   M+
   drop
;

: Y2kTimestampToTimeAndDate ( u -- d n )
   0 SECONDS_PER_DAY UM/MOD  Y2k0Day + >r 0 r>      
;

: .Time- ( ud -- )   (TIME) TYPE ;

: ShowY2kTimestamp ( u -- )
   Y2kTimestampToTimeAndDate (date_)+ type space  .Time-   
;

: UnixTimestampToTimeAndDate ( u -- d n )
   0 SECONDS_PER_DAY UM/MOD  Unix0DAY + >r 0 r>      
;

: ShowUnixTimestamp ( u -- )
   UnixTimestampToTimeAndDate (date_)+ type space  .Time-   
;

: ttY2k ( -- )   @Y2kTimeStamp ShowY2kTimestamp ;

\ show the number of seconds as  dd days hh:mm:ss  ( if dd is not 0 )
: ShowElapsedTime ( u -- )   
   SECONDS_PER_DAY /MOD 
   ?dup if  dup . ." day"  1 = not if  ." s"  then  space  then
   0 .TIME-
;

\ *****************************************************************************
\ Character analysis
\ *****************************************************************************

: Cspace= ( c -- f )
   >r
   r@ 32 =           \ space
   r@ [char] ( = or  \ ( will count as a space
   r@ [char] ) = or  \ ) will count as a space
   r@ [char] [ = or  \ [ will count as a space
   r@ [char] ] = or  \ ] will count as a space
   r> drop
;

: Cvalid= ( c -- f )  \ only these characters are valid in a C name
   >r  0
   r@ [char] _ = or \ an underscore
   r@ [char] 0 [char] 9 1+ within or
   r@ [char] a [char] z 1+ within or
   r@ [char] A [char] Z 1+ within or
   r> drop
;

\ *****************************************************************************
\ High-level version of -Trailing etc
\ *****************************************************************************
variable -trail#

: -TrailingPlusReverse ( a n - a n' )
   -trail# !
   dup

   dup -trail# @ + 1-     \ address of last character
   -trail# @ 0 ?do
      dup i - c@ Cspace= not if  leave  else  -1 -trail# +!  then
   loop
   drop
   -trail# @
;

: -TrailingPlus ( a n - a n' )
   0 -trail# !
   2dup
   over + swap ?do
      i c@ Cspace= if  leave  else  1 -trail# +!  then
   loop
   drop
   -trail# @
;

: -LeadingC ( a n - a' n' )
   2dup 
   over + swap ?do
      i c@ Cspace= not if  leave  then
      1 /string
   loop  
   0 max
;

: -Leading ( a n - a' n' )
   2dup 
   over + swap ?do
      i c@ $20 > if  leave  then
      1 /string
   loop  
   0 max
;

: -Comments ( $ -- $ )
   [char] / split  2swap 2drop
   [char] \ split  2swap 2drop
   [char] ( split  2swap 2drop
;

: tt-c ( -- )  
   s" code // and comments" 2dup cr type cr -Comments type
   s" code \ and comments" 2dup cr type cr -Comments type
   s" code ( and comments )" 2dup cr type cr -Comments type 
   s" ( and comments )" 2dup cr type cr -Comments ." >>>" type ." <<<"
   cr
;


: NumberOfLeading ( a n - c )    \ number of leading spaces in the given string
   0 -rot
   over + swap ?do
      i c@ Cspace= not if  leave  then
      1+
   loop  
;

: ttnl
   cr  s" 1234"         2dup type NumberOfLeading 20 ToColumn .
   cr  s"     1234"     2dup type NumberOfLeading 20 ToColumn .
   cr  s"     1234  "   2dup type NumberOfLeading 20 ToColumn .
   cr  s"     "         2dup type NumberOfLeading 20 ToColumn .
;

: tt-l
   cr
   s" hello   " 
   cr  ." Original string |" 2dup type ." |"
   cr  ."        -Leading |" 2dup -Leading  type ." |"
   cr
   s"    hello   " 
   cr  ." Original string |" 2dup type ." |"
   cr  ."        -Leading |" 2dup -Leading  type ." |"
   cr
   s"    hello" 
   cr  ." Original string |" 2dup type ." |"
   cr  ."        -Leading |" 2dup -Leading  type ." |"
;

: tt-t  
   s" hello(args        " 
   cr  ." Original string |" 2dup type ." |"
   cr  ."       -Trailing |" 2dup -Trailing  type ." |"
   cr  ."   -TrailingPlus |" -TrailingPlus  type ." |"
   cr
   s" helloargs     " 
   cr  ." Original string |"  2dup type ." |"
   cr  ."       -Trailing |"  2dup -Trailing  type ." |"
   cr  ."   -TrailingPlus |"   -TrailingPlus  type ." |"

   cr
   s" helloargs" 
   cr  ." Original string |"  2dup type ." |"
   cr  ."       -Trailing |"  2dup -Trailing  type ." |"
   cr  ."   -TrailingPlus |"   -TrailingPlus  type ." |"

;

\ *****************************************************************************
\ Parsing example
\ *****************************************************************************

variable DateTimeStamp

create MMMnames[] ," JanFebMarAprMayJunJulAugSepOctNovDec" 

: MMM>n ( $month -- n | 0 )   \ 0 means not found else n = month with January = 1
   0  -rot 
   #12 0 do 
      i 3 * MMMnames[] count drop + over 2over compare(NC) 0= if
         rot 1+ i + -rot
         leave
      then
   loop
   2drop
;

: ParseDateAndTime ( -- )
   [char] , word drop \ skip the day of the week and ',' 
   \ Date: Fri, 04 Sep 2015 11:29:20 GMT   
   bl word count Dnumber drop  \ dup 2 u.r space
   bl word count MMM>n         \ dup 2 u.r space
   bl word count Dnumber drop  \ dup 4 u.r space
   D-M-Y                       \ dup (date_)+ type space
   Unix0DAY -
   SECONDS_PER_DAY * DateTimeStamp !
 
   [char] : word count -Leading Dnumber drop  \ dup 2 u.r space
   #60 *
   [char] : word count -Leading Dnumber drop  \ dup 2 u.r space
   + #60 *
   bl word count Dnumber drop                 \ dup 2 u.r space
   +
   DateTimeStamp +!
  
   MAGENTA. 
   DateTimeStamp @ Unix@NOW (date_)+ type space  .Time
   space DateTimeStamp @ 10 u.r
   RESTORE.
;

\ *****************************************************************************
\ ASCII control character display
\ *****************************************************************************

$0A constant '/n' \ linefeed (newline)
$0D constant '/r' \ carriage return

$00 constant <NUL>	\ Null char
$01 constant <SOH>	\ Start of Heading
$02 constant <STX>	\ Start of Text
$03 constant <ETX>	\ End of Text
$04 constant <EOT>	\ End of Transmission
$05 constant <ENQ>	\ Enquiry
$06 constant <ACK>	\ Acknowledgment
$07 constant <BEL>	\ Bell
$08 constant <BS>    \ Back Space
$09 constant <HT>    \ Horizontal Tab
$0A constant <LF>    \ Line Feed
$0B constant <VT>    \ Vertical Tab
$0C constant <FF>    \ Form Feed
$0D constant <CR>    \ Carriage Return
$0E constant <SO>    \ Shift Out / X-On
$0F constant <SI>    \ Shift In / X-Off
$10 constant <DLE>	\ Data Line Escape
$11 constant <DC1>	\ Device Control 1 (oft. XON)
   $11 constant <XON>	\ Device Control 1 (oft. XON)
$12 constant <DC2>	\ Device Control 2
$13 constant <DC3>	\ Device Control 3 (oft. XOFF)
   $13 constant <XOFF>	\ Device Control 3 (oft. XOFF)
$14 constant <DC4>	\ Device Control 4
$15 constant <NAK>	\ Negative Acknowledgement
$16 constant <SYN>	\ Synchronous Idle
$17 constant <ETB>	\ End of Transmit Block
$18 constant <CAN>	\ Cancel
$19 constant <EM>	   \ End of Medium
$1A constant <SUB>	\ Substitute
$1B constant <ESC>	\ Escape
$1C constant <FS>    \ File Separator
$1D constant <GS>    \ Group Separator
$1E constant <RS>    \ Record Separator
$1F constant <US>    \ Unit Separator

variable SerialChecksum

: emit+ ( c -- ) 
   dup <CR> = not over 0 $20 within and  dup >r if  GREEN.  then 
   dup SerialChecksum @ xor SerialChecksum !
   dup case
   <NUL> of  ." <NUL>" drop  endof  \  Null char                     
   <SOH> of  ." <SOH>" drop  endof  \  Start of Heading              
   <STX> of  ." <STX>" drop  endof  \  Start of Text                 
   <ETX> of  ." <ETX>" drop  endof  \  End of Text                   
   <EOT> of  ." <EOT>" drop  endof  \  End of Transmission           
   <ENQ> of  ." <ENQ>" drop  endof  \  Enquiry                       
   <ACK> of  ." <ACK>" drop  endof  \  Acknowledgment                
   <BEL> of  ." <BEL>" drop  endof  \  Bell                          
   <BS>  of  ." <BS>"  drop  endof  \  Back Space                    
   <HT>  of  ." <HT>"  drop  endof  \  Horizontal Tab                
   <LF>  of  ." <LF>"  drop  endof  \  Line Feed                     
   <VT>  of  ." <VT>"  drop  endof  \  Vertical Tab                  
   <FF>  of  ." <FF>"  drop  endof  \  Form Feed                     
   <CR>  of  ( ." <CR>" ) drop  cr  endof  \  Carriage Return               
   <SO>  of  ." <SO>"  drop  endof  \  Shift Out / X-On              
   <SI>  of  ." <SI>"  drop  endof  \  Shift In / X-Off              
   <DLE> of  ." <DLE>" drop  endof  \  Data Line Escape              
   <DC1> of  ." <DC1-Xon>" drop  endof  \  Device Control 1 (oft. XON)   
   <DC2> of  ." <DC2>" drop  endof  \  Device Control 2              
   <DC3> of  ." <DC3-Xoff>" drop  endof  \  Device Control 3 (oft. XOFF)  
   <DC4> of  ." <DC4>" drop  endof  \  Device Control 4              
   <NAK> of  ." <NAK>" drop  endof  \  Negative Acknowledgement      
   <SYN> of  ." <SYN>" drop  endof  \  Synchronous Idle              
   <ETB> of  ." <ETB>" drop  endof  \  End of Transmit Block         
   <CAN> of  ." <CAN>" drop  endof  \  Cancel                        
   <EM>	of  ." <EM>"  drop  endof  \  End of Medium                 
   <SUB> of  ." <SUB>" drop  endof  \  Substitute                    
   <ESC> of  ." <ESC>" drop  endof  \  Escape                        
   <FS>  of  ." <FS>"  drop  endof  \  File Separator                
   <GS>  of  ." <GS>"  drop  endof  \  Group Separator               
   <RS>  of  ." <RS>"  drop  endof  \  Record Separator              
   <US>  of  ." <US>"  drop  endof  \  Unit Separator                
     emit
  endcase
  r> if  RESTORE. then
;

: type+ ( c-addr n -- )   over + swap ?do  i c@ emit+  loop ;

\ *****************************************************************************
\ ** Q Variables
\  Just like normal variables and buffers, but have more than one
\  instance, selected by the value in Qselect
\ *****************************************************************************

4 constant NUM_QSELECTS

variable Qselect     \ selects one set of variables and buffers. 
\ Note : must not be a USER variable!!! Windows "tasks" have ther own user space...

\ Four COM ports are defined : COM-A COM-B COM-C and COM-D , each port has its own "real" COM port setting and baudrate etc
\ All settings are saved in the registry and restored by GetRegValues
\ Type .COMS to see the current settings.
\ This allows a program to access more than one COM port
: COM-A ( -- )   0 Qselect ! ;
: COM-B ( -- )   1 Qselect ! ;
: COM-C ( -- )   2 Qselect ! ;
: COM-D ( -- )   3 Qselect ! ;

COM-A

{ 
: serialColour   
   Qselect @ case
   0 of  NORMAL.  endof
   1 of  BLUE.    endof
   2 of  MAGENTA. endof
   3 of  CYAN.    endof
   INVERSE.
   endcase
;

: .serialPort
   serialColour
   Qselect @ case
   0 of  ." COM port A "  endof
   1 of  ." COM port B "  endof
   0 of  ." COM port C "  endof
   1 of  ." COM port D "  endof
      RED. ." Unknown COM port =" dup 3 .hex
   endcase
   serialColour
;
}

: Qvariable     \ "<spaces>name" -- ; Exec: -- a-addr
  create
    NUM_QSELECTS cells allot
  does>
    Qselect @ NUM_QSELECTS 1- and cells +
;

: 2Qvariable     \ "name" -- ; [child] -- a-addr
\ *G Defines a 2variable, with a version for each Qselect value
   create
      NUM_QSELECTS 2* cells allot
   does>
      Qselect @  NUM_QSELECTS 1- and  2*  cells +
;

: Qbuffer:       \ n "name" -- ; [child] -- addr                 
( *G Create a multiple memory buffer called "name" which is 'n' bytes long.      )
( ** When "name" is executed the address of the buffer is returned.     )
    Create
    dup ,                  \ save the buffer size
    NUM_QSELECTS *         \ allot one for each comport
    Here Over Allot Swap Erase
  Does>
    dup @                  \ get the size
    Qselect @ NUM_QSELECTS 1- and * +  \ offset to the correct buffer
    cell +                 \ step over the size field
;

{ 
\ test Qvariables

Qvariable TestQvar

: TTQvar
   0 Qselect !  100 TestQvar !
   1 Qselect !  101 TestQvar !
   2 Qselect !  102 TestQvar !

   1 Qselect !  TestQvar @ cr ." 1 = " .
   0 Qselect !  TestQvar @  ."   0 = " .
   2 Qselect !  TestQvar @  ."   2 = " .
;

\ test Qbuffer:s
2 cells Qbuffer: TestQbuf

: TTqbuf
   0 Qselect !  100 0 TestQbuf 2!
   1 Qselect !  101 0 TestQbuf 2!
   2 Qselect !  102 0 TestQbuf 2!

   1 Qselect !  TestQbuf 2@ cr ." 1 = " d.
   0 Qselect !  TestQbuf 2@  ."   0 = " d.
   2 Qselect !  TestQbuf 2@  ."   2 = " d.
;
} 

{


\ *****************************************************************************
\ Non Volatile storage using the Registry
\ *****************************************************************************

FUNCTION: RegCreateKeyEx   ( a b c d e f g h i -- x )
FUNCTION: RegOpenKeyEx     ( a b c d e -- x )

$100 2 cells - constant KEYNAME_MAXPATH   \ maximum size of the subkey string

variable RegHandle 

: ConstructRegKeyName ( $ a -- z$ )
   >r
   s" InventioSoftware\" r@ place  
   0 max KEYNAME_MAXPATH 2 -     \ the 2 - allows space for the count byte and 0 terminator
   r@ c@ - 0 max  min            \ the size of the string already in the allocated buffer
   r@ append 
   s" \" r@ append
   0 r@ count + c!               \ null terminate the string
   r> count drop                 \ start at the data address
;

: OpenRegKey ( $ -- handle ior )  \ created is true if the key was created by this call, ior is 0 for success
   2 CELLS KEYNAME_MAXPATH + R-ALLOC >r \ allocate a buffer on the R-stack                                                              
   r@ ConstructRegKeyName
   \ call registry key open      \ LONG WINAPI RegOpenKeyEx(       LONG WINAPI RegCreateKeyEx(  
   HKEY_CURRENT_USER             \   _In_        HKEY hKey,          _In_        HKEY hKey,         \ HKEY_CURRENT_USER     
   ( $ ) swap                    \   _In_opt_    LPCTSTR lpSubKey,   _In_        LPCTSTR lpSubKey,  \ z" MySubKeyName"                                       
   0                             \   _Reserved_  DWORD ulOptions,    _Reserved_  DWORD Reserved,    \ 0                                                      
   KEY_ALL_ACCESS                \   _In_        REGSAM samDesired,                                                  
   R@ KEYNAME_MAXPATH +          \   _Out_       PHKEY phkResult     _In_        DWORD dwOptions,   \ 0 = REG_OPTION_NON_VOLATILE , 1 = REG_OPTION_VOLATILE , 2 = REG_OPTION_CREATE_LINK,  4 = REG_OPTION_BACKUP_RESTORE                              
   RegOpenKeyEx                \                                   );                                                                                        
   r> KEYNAME_MAXPATH + @ swap 1 =                                                
;                                                                   


: OpenOrCreateRegKey ( $ -- handle created ior )  \ created is true if the key was created by this call, ior is 0 for success
   2 CELLS KEYNAME_MAXPATH + R-ALLOC >r \ allocate a buffer on the R-stack. Note : the extra two cells are used for the phkResult and lpdwDisposition return values in the code below
   r@ ConstructRegKeyName
   \ call registry key open      \ LONG WINAPI RegCreateKeyEx(  
   HKEY_CURRENT_USER             \   _In_        HKEY hKey,         \ HKEY_CURRENT_USER     
   ( $ ) swap                    \   _In_        LPCTSTR lpSubKey,  \ z" MySubKeyName"                                       
   0                             \   _Reserved_  DWORD Reserved,    \ 0                                                      
   NULL                          \   _In_opt_    LPTSTR lpClass,    \ NULL                                                   
   REG_OPTION_NON_VOLATILE       \   _In_        DWORD dwOptions,   \ 0 = REG_OPTION_NON_VOLATILE , 1 = REG_OPTION_VOLATILE , 2 = REG_OPTION_CREATE_LINK,  4 = REG_OPTION_BACKUP_RESTORE    
   KEY_ALL_ACCESS                \   _In_        REGSAM samDesired, \ KEY_ALL_ACCESS (0xF003F)        
   NULL                          \   _In_opt_    LPSECURITY_ATTRIBUTES lpSecurityAttributes, \ NULL                          
   R@ KEYNAME_MAXPATH +          \   _Out_       PHKEY phkResult,   \ RegHandle                                              
   R@ KEYNAME_MAXPATH + cell+    \   _Out_opt_   LPDWORD lpdwDisposition                                                     
   RegCreateKeyEx                \ );                                                                                        
   r> KEYNAME_MAXPATH + 2@ swap 1 = rot                                               
;
}
: struct ( a c -- a' )   over  CONSTANT + ;
{
0
   MAX_PATH  struct oRegPortname
   cell      struct oRegPortNumber
   cell      struct oRegBaudrate
   cell      struct oRegPortSettings
   cell      struct oRegLogging
constant |SerialPort|

0 
   2 cells                     struct oRegWindowOrigin
   2 cells                     struct oRegWindowSize
   2 cells                     struct oRegShowHexDump
   1 cells                     struct oRegRemoteCurve_F2
   1 cells                     struct oRegRemoteXXX
   MAX_PATH 4 *                struct oRegFilenames   \ Four file names for application use
   |SerialPort| NUM_QSELECTS * struct SerialPorts[]
   cell struct oRegWindowDisplayScale
   cell struct RegNullTerminator
constant |RegConfig|

|RegConfig| 8192 >= [if]  cr .( Warning RegConfig[] >= 8192 bytes - too big!!! ) [then]

|RegConfig| Buffer: RegConfig[]

: >RegWindowOrigin       ( -- a )   oRegWindowOrigin RegConfig[] + ;
: >RegWindowSize         ( -- a )   oRegWindowSize RegConfig[] + ;
: >RegWindowDisplayScale ( -- a )   oRegWindowDisplayScale RegConfig[] + ;
: >RegShowHexDump        ( -- a )   oRegShowHexDump RegConfig[] + ;
: >RegRemoteCurve_F2     ( -- a )   oRegRemoteCurve_F2 RegConfig[] + ;
: >RegRemoteXXX          ( -- a )   oRegRemoteXXX RegConfig[] + ;

: >RegFilenames ( -- a )    oRegFilenames  RegConfig[] + ;
: RegFilename  ( n -- a )    MAX_PATH * >RegFilenames + ;   

: >SerialPortInfo ( Qselect# -- a )   NUM_QSELECTS 1- and  |SerialPort| * SerialPorts[] +  RegConfig[] + ;   

: >RegPortname     ( Qselect# -- a )   >SerialPortInfo oRegPortname +     ;  
: >RegPortNumber   ( Qselect# -- a )   >SerialPortInfo oRegPortNumber +   ;  
: >RegBaudrate     ( Qselect# -- a )   >SerialPortInfo oRegBaudrate +     ;
: >RegPortSettings ( Qselect# -- a )   >SerialPortInfo oRegPortSettings + ;
: >RegLogging      ( Qselect# -- a )   >SerialPortInfo oRegLogging +      ;   

: COMnumberToName ( c -- $ )
    s" \\.\COM" pad place 
   (.) 0 max  pad append
   pad count
;

: (GetRegValues) ( RegHandle -- )
\   cr ." (GetRegValues) "  dup 8 .hex space 
   cell R-ALLOC >r            \ LONG WINAPI RegQueryValueEx(                        
   ( RegHandle -- )           \   _In_         HKEY hKey,
   z" Config"                 \   _In_opt_     LPCTSTR lpValueName,
   NULL                       \   _Reserved_   LPDWORD lpReserved,
   0                          \   _Out_opt_    LPDWORD lpType,
   RegConfig[]                \   _Out_opt_    LPBYTE lpData,
   |RegConfig| r@ ! r>        \   _Inout_opt_  LPDWORD lpcbData
\   cr ." >>>> " .s  dup @ 8 .hex space 
   RegQueryValueEx            \ );  
   dup if  
      \ cr ." (GetRegValues) " dup .LastErrorText  
   then  throw
;      
       
: (SetRegValues) ( RegHandle -- )
\   cr ." (SetRegValues) "  dup 8 .hex space 
                        \ LONG WINAPI RegSetValueEx(                        
   ( RegHandle -- )     \   _In_        HKEY hKey,
   z" Config"           \   _In_opt_    LPCTSTR lpValueName,
   NULL                 \   _Reserved_  DWORD Reserved,
   REG_BINARY           \   _In_        DWORD dwType,
   RegConfig[]          \   _In_        const BYTE *lpData,
   |RegConfig|          \   _In_        DWORD cbData
   RegSetValueEx        \ );
   dup if  
      \ cr ." (SetRegValues) " dup .LastErrorText  
   then throw
;    

: CloseRegKey ( RegHandle -- )
   RegCloseKey drop 
;

: SetRegValues ( -- )
   Acronym OpenOrCreateRegKey ERROR_SUCCESS = not if  #9000 throw  then  
   dup if  
\      cr ." SetRegValues "  dup .LastErrorText  
   then  throw
   >r
   r@ (SetRegValues)
   r> CloseRegKey
;

0 
ENUM 8N1
ENUM 8E1
ENUM 7E1
drop

: (InitRegValues) ( -- )
\   cr ."    (InitRegValues)"
   RegConfig[] |RegConfig| erase \ erase the filename etc
   \ filename has already been erased
   #50 #50 >RegWindowOrigin 2!
   #624 #400 >RegWindowSize 2!
   2 >RegWindowDisplayScale !
   0. >RegShowHexDump 2!
   1 >RegRemoteCurve_F2 !
   0 >RegRemoteXXX !
   NUM_QSELECTS 0 do  
      i 1+ dup i >RegPortNumber !
      COMnumberToName  i >RegPortname place  
      #9600 i >RegBaudrate ! 
      8N1   i >RegPortSettings !  \ default to 8N1 
      0     i >RegLogging ! 
  loop
;

: GetRegValues ( -- ) 
   Acronym OpenOrCreateRegKey 2drop \ ERROR_SUCCESS = not if  #9000 throw  then  
   >r 
   r@ ['] (GetRegValues) catch if   \ if there is no registry entry yet, we must create one
      drop
 \     cr ." New registry entry being made - Init and SetRegValues " 
      (InitRegValues)
      r@ (SetRegValues)  
   then
   r> CloseRegKey
;

: ShowRegValues ( -- )
   cr 
   GetRegValues
   cr ." WindowOrigin = "  >RegWindowOrigin 2@ swap . ." x "  . 
   cr ." WindowSize = "  >RegWindowSize 2@ swap . ." x "  . 
   cr ." Filename   = "  >RegFilenames count type
   cr ." COM# Name      Baudrate  settings logging "
   NUM_QSELECTS 0 do  
      cr i 2 u.r  3 spaces  
      i >RegPortname count type 
      15 ToColumn i >RegBaudrate @ 8 u.r  1 spaces  i >RegPortSettings @ 2 .hex  8 spaces   i >RegLogging @ . 
   loop
   cr ." WindowDisplayScale = "  >RegWindowDisplayScale @ .
;

: InitRegValues ( -- )
   (InitRegValues)
   SetRegValues
;

{  
ONG WINAPI RegDeleteKeyEx(
  _In_        HKEY hKey,
  _In_        LPCTSTR lpSubKey,
  _In_        REGSAM samDesired,
  _Reserved_  DWORD Reserved
);
]


2 IMPORT: RegDeleteKey  ( hKey z$ -- ior ) 
\ 4 IMPORT: RegDeleteKeyEx  ( hKey lpSubKey samDesired 0 -- ior ) \ Note : WinXP Pro 64 bitand above only !!! 
\ 2 IMPORT: RegDeleteTree ( hKey z$ -- ior )

$200 constant KEY_WOW64_32KEY
$100 constant KEY_WOW64_64KEY

: (RegDelete) ( -- )   \ delete the sub key "Acronym" (defined by the application) from the InventioSoftware root in the registry
   KEYNAME_MAXPATH cell + R-ALLOC    \ allocate a buffer on the r-stack, with an extra cell for the null terminator
   dup KEYNAME_MAXPATH cell + erase  
   dup  Acronym  rot swap cmove  \ create a z$ null terminated string

   0 0 OpenRegKey drop  \ open the "InventioSoftware" root 
\   dup 0= if  2drop  cr ." already deleted!"  exit  then  \ already deleted!
   ( handle ) >r
 \  r@  swap  KEY_WOW64_32KEY KEY_WOW64_64KEY or  0 RegDeleteKeyEx dup if  
   r@  swap  RegDeleteKey dup if  
      \ cr ." RegDeleteKey failed with error " dup .  ." : "  .LastErrorText  
      drop  #9001  throw  
   else  
      drop  
   then 
   r> CloseRegKey
;

: RegDelete ( -- )   \ delete the sub key from the InventioSoftware root in the registry
   ['] (RegDelete) catch if  
\      cr ." already deleted!"  
   then  
;

}
\ *****************************************************************************
\ Alternative to S" and ."  using  %  as a delimiter
\ *****************************************************************************

: ,% ( -- )   [CHAR] % STRING ;

: S% ( "ccc"<"> -- )
   STATE @ IF  POSTPONE (S") ,%  EXIT  THEN
   [CHAR] % PARSE  >QPAD COUNT ;  IMMEDIATE

: .%  \ compiling: ( "ccc<">" -- )  executing: ( -- )
   POSTPONE S% POSTPONE TYPE ;  IMMEDIATE

\ : ttalternate% ( -- )   .% hello "world"!% ;

: "FOX" ( -- c-addr n )  
   s" THE QUICK BROWN FOX JUMPS OVER THE LAZY DOG" 
;

: .Len ( u -- )   dup  .  1023 + 1024 /  ."   = "  .  ." Ki bytes"  ;

: BE! ( x addr -- )     \ BigEndian !  for LittleEndian processors
   over $0000000FF and             over 3 + c!
   over $00000FF00 and $08 Rshift  over 2 + c!
   over $000FF0000 and $10 Rshift  over 1 + c!
   swap $0FF000000 and $18 Rshift  swap     c! 
;

\ : BE@ ( addr -- x )     \ BigEndian @ for LittleEndian processors
\    dup  3 + c@
\    over 2 + c@ $08 Lshift or
\    over 1 + c@ $10 Lshift or
\    swap     c@ $18 Lshift or
\ ;

\ like +! , but OR's the bits in u into the variable
: +or ( u a -- )   >r  r@ @ or r> ! ;

\ like +! , but AND's out the bits in u from the variable
: -or ( u a -- )   >r -1 xor  r@ @ and r> ! ;

: vM+ ( n1 n2 n -- n1+n n2+n )   >r  r@ +  swap r> +  swap ;

variable DumpOffset

: DUMPLINE+ ( addr u -- )
   1 ?SCROLL  BASE @ >R  HEX
   DumpOffset @ 3 U.R 2 spaces  $10 DumpOffset +!  2DUP DUMPHEX  DUMPTEXT  ."   \ "  
   R> BASE ! ;

: DUMP+ ( addr u -- )   0 DumpOffset !  /SCROLL
   BEGIN ( a n)  2DUP 16 MIN DUMPLINE+
      16 /STRING  DUP 0 <= UNTIL 2DROP ;

: DUMPLINEcapture ( addr u -- )
   1 ?SCROLL  BASE @ >R  HEX
   ."    16c,  "  2DUP DUMPHEX  DUMPTEXT  ."   \ "  DumpOffset @ 3 U.R 2 spaces  $10 DumpOffset +!
   R> BASE ! ;

: DUMPcapture ( addr u -- )   0 DumpOffset !  /SCROLL
   BEGIN ( a n)  2DUP 16 MIN DUMPLINEcapture
      16 /STRING  DUP 0 <= UNTIL 2DROP ;


$48 constant BYTESPERLINE

: DUMPHEXN ( addr n -- )
   BYTESPERLINE 0 DO  DUP 0 <= IF  3 SPACES
      ELSE  OVER C@ 
      3 lshift  
      0 <# # # #> TYPE SPACE  THEN
   1 /STRING  LOOP 2DROP ;

: DUMPLINEcaptureN ( addr u -- )
   1 ?SCROLL  BASE @ >R  HEX
   ."    $48 nc,  "  2DUP DUMPHEXN  DUMPTEXT  ."   \ "  DumpOffset @ dup $2.hex  2 spaces  emit+   1 DumpOffset +!
   R> BASE ! ;

: DUMPcaptureN ( addr u -- )   0 DumpOffset !  /SCROLL
   BEGIN ( a n)  2DUP BYTESPERLINE MIN DUMPLINEcaptureN
      BYTESPERLINE /STRING  DUP 0 <= UNTIL 2DROP ;

\ Add 16 bytes to the structure, using the 16 hex ASCII words following, then ignore the rest of the line
\ This allows captured packet data to be copied and pasted...
: 16c,   base @ >r  hex 
   #16 0 do   bl word  count number  c,  loop 
   r> base ! 
   postpone \ 
;  

\ Add n bytes to the structure, using the n hex ASCII words following, then ignore the rest of the line
\ This allows captured packet data to be copied and pasted...
: nc, ( -- )  base @ >r  hex 
   ( n -- ) 0 do   bl word  count number  c,  loop 
   r> base ! 
   postpone \ 
;  

\ Add n bytes to the structure, using the n hex ASCII words following, then ignore the rest of the line
\ This allows captured packet data to be copied and pasted...
: c," ( -- )  base @ >r  hex 
   [char] " word count dup 2/ c, 
   over + swap ?do  i HexC@  c,  2 +loop 
   r> base ! 
   postpone \ 
;  
: c,"- ( -- )  base @ >r  hex 
   [char] " word count \ dup 2/ c, 
   over + swap ?do  i HexC@  c,  2 +loop 
   r> base ! 
   postpone \ 
; 

: ekeyQuit?
   ekey #06 ( ctrlF) = if  cr ." Done. "  cr  ."  ok "  quit  then 
;

\ Limited length type, replace control characters by '.'
: Ltype ( a n c -- )   -rot dup >r  over + swap ?do  i c@ dup $20 < if  drop  [char] .  then  emit  loop   r> - 0 max  spaces ;

: ttlt  s" Fred" 10 Ltype ;

\ dump on the same line
: dumpL ( a n -- )   0 max  $100 min  space  dup 2.hex ."  |  " over + swap ?do  i c@  2.hex space  loop ; 

variable LineXpos
\ dump on the same line, up to 64 bytes, then move to the next line
: dumpL+ ( a n -- )   
   Get-XY drop LineXpos !
   0 max  $100 min  space  dup 2.hex ."  |  "  
   ( n -- ) 0 ?do  i $1F and 0=  i 0= not and if  cr  LineXpos @ 7 + ToColumn  then dup i + c@  2.hex space  loop ; 

\ dump on the same line, up to 64 bytes, then move to the next line
: dumpL++ ( a n -- )   
   Get-XY drop LineXpos !
   0 max  $1000 min  space  dup 2.hex ."  |  "  
   ( n -- ) 0 ?do  i $1F and 0=  i 0= not and if  cr  LineXpos @ 7 + ToColumn  then dup i + c@  
   
   dup [char] : = if  INVERSE. 2.hex  RESTORE.  else  2.hex  then  space  loop ; 


: Y/N-EXIT ( -- )   cr  ." 'Y/N-EXIT'  is deprecated... please use  'PressY' "  key UPPER [char] Y = not if   r> drop  exit  then ;

: PressY ( -- f )   key UPPER [char] Y = ;

\ *****************************************************************************
\ Pseudo Random Number Generator
\ *****************************************************************************

variable vRandom

: InitRandom   0 vRandom ! ;
InitRandom

\ simple multiplicative - shows vertial line patterns
\ : Random32 ( -- u )   vRandom @  3141592621 *  1+  DUP vRandom !  ; 

\ Marsaglia, "Xorshift RNGs".  http://www.jstatsoft.org/v08/i14/paper
: Random32 ( -- u )
    vRandom @
    dup 0= or
    dup 6 lshift xor
    dup 21 rshift xor
    dup 7 lshift xor
    dup vRandom ! ; 

: Random ( n -- u )   1+ Random32 UM* NIP ;

\ *****************************************************************************
\ *****************************************************************************

: 2Tone ( -- )   1760 400 beep drop  2640 400 beep drop ;

: AudibleWait
   0 
   5 0 do  2Tone  key? if  1- leave  then  loop 
   0= if  key drop  then 
;

variable StackDepthValue   \ used to saved stack depth changes deep in code...

: .StackDepths
   depth if   cr ." *** depth = "  depth .   ."     TOS = "  dup . depth 1 > if  over .  then   then
   StackDepthValue @ if  ." *** StackDepthValue change = "  StackDepthValue @ . then
;

\ *****************************************************************************
\ Windows questions
\ *****************************************************************************

: Monologue ( z$ z$ -- )
   HWND  -rot  
   MB_TASKMODAL
\   MB_ICONWARNING or      \ add a big "!"
   MB_ICONINFORMATION or  \ add an "i"
\   MB_OK or                   \ "OK" button
   MessageBox drop ;

: ttmono   z" Message text" z" Monlogue title bar text "  Monologue ;

: Dialogue ( z$ z$ -- n )  \ returns 1 for "OK" 
   HWND  -rot  

   MB_TASKMODAL
   MB_OKCANCEL or       \ Ok and cancel buttons
   MB_ICONQUESTION or   \ add a big "?"

   MessageBox ;

: ttdia   z" Message text" z" Dialogue title bar text "  Dialogue cr ." Button = " . ;


\ *****************************************************************************
\ stack depth tests
\ *****************************************************************************
variable StartDepth

: Init?Depth   depth StartDepth ! ;

: ?Depth ( $ -- )   depth dup  StartDepth @ 2+ = not if  cr ." ?Depth = " .   2 spaces type   key 27 = if  quit  then  else  2drop  drop  then ;

\ *****************************************************************************
\ ForkLike
\ Adapted from the SwiftForth file fork.f - thanks!
\ Not quite a real Unix  fork() , as it does not clone the callers state
\ but useful nonetheless
\ *****************************************************************************

{ --------------------------------------------------------------------
_PARAM_0 has a pointer to the startup data, which needs to be
released as soon as it is used. The callback will initialize
a dictionary, a user area, and the stacks. The data passed
is simply a pointer to a structure which contains

   +0  'this
   +4  'self
   +8  xt to run
   +12 user parameter

Local variables and objects may not be passed across this interface

ForkLike() passes param into the new thread as top-of-stack.
-------------------------------------------------------------------- }
variable #forks

: InitForks  0 #forks ! ;

:NONAME ( -- )
   _PARAM_0 @+ 'THIS !  @+ 'SELF !  @+  SWAP @  _PARAM_0 FREE DROP
   CATCH ExitThread ;  1 CB: cb_FORKED

: ForkLike() ( param xt -- handle )
   1 #forks +!  #forks @ 1500 > if  2drop  INVALID_HANDLE_VALUE  exit  then
   SWAP  0 >R
   4 CELLS ALLOCATE IF DROP INVALID_HANDLE_VALUE EXIT THEN
   DUP >R  THIS !+  SELF !+  SWAP !+ !
   0 65536 cb_FORKED R> 0 RP@ CreateThread   R> DROP ;

: ForkLike ( xt -- )   0 SWAP ForkLike() drop ;


\ *****************************************************************************
\ Titles
\ *****************************************************************************

: CharLine ( n c -- )   swap 0 ?do  dup EMIT  loop drop ;
: stars ( n -- )   [char] * CharLine ;

variable LengthSoFar
#96 constant TITLE_LINE_WIDTH
#96 constant TITLE_WIDTH

: (ShowTitleLine) ( $ c -- )
   >r
\   cr TITLE_LINE_WIDTH r@ CharLine
   cr  -Trailing  dup 4 + LengthSoFar !      \ the text plus 2 spaces before and two after
   TITLE_LINE_WIDTH LengthSoFar @ - 1 max    \ room left for 'stars'
   2/ dup r@ CharLine                        \ half of them before
   LengthSoFar +!
   2 spaces 
   NORMAL.  type  RESTORE.
   2 spaces 
   TITLE_LINE_WIDTH LengthSoFar @ - 1 max r@ CharLine   \ room left for 'stars' at the end
\   cr TITLE_LINE_WIDTH r@ CharLine
   r> drop
;

: ShowTitleLine= ( $ -- )   [char] = (ShowTitleLine) ;
: ShowTitleLine- ( $ -- )   [char] - (ShowTitleLine) ;

: ShowTitle ( $ -- ) 
   TITLE_WIDTH stars
   cr  -Trailing  dup 4 + LengthSoFar !   \ the text plus 2 spaces before and two after
   TITLE_WIDTH LengthSoFar @ -            \ room left for stars
   2/ dup stars                           \ half of them before
   LengthSoFar +!
   2 spaces 
   NORMAL.  type  RESTORE.
   2 spaces 
   TITLE_WIDTH LengthSoFar @ - stars      \ room left for stars at the end
   cr TITLE_WIDTH stars
;

\ *****************************************************************************
\ Show the following lines in the include file
\ Useful when a new file is being created by +Log
\ *****************************************************************************

\ text between Show{ and End} is typed or compiled to be typed when the compiled word is executed
: Show{ ( -- ) \ } 
   begin 
      
      REFILL 0= if #1235 throw  then   \ get the next line

      SOURCE 2dup s" End}" SEARCH(NC) >r 2drop r> 
      if   \ we have found the end marker
         2drop -1
      else
         STATE @ if  \ compiling
            POSTPONE CR  POSTPONE (S")  ( a n -- ) STRING,  POSTPONE TYPE   
         else        \ interpreting
            cr ( a n -- ) type
         then
         0
      then 
   until
   REFILL 0= if #1235 throw  then            \ discard the line containing End}
; IMMEDIATE

\ example code : text between Show{ and End} is typed
{ 
Show{ 
hello world!!!
1 2 3 4 5 

22            33           44        


13 14 15 6 
44
End}

: ttSNL
   Show{ 
   hello world!!!
   1 2 3 4 5 

   22            33           44        
   
   
   13 14 15 6 
   44
   End}
;

{
variable indent
variable StructLevel   \ where we are in the elements in the structure

3 constant INDENT_SPACES

: cr+   cr  indent @ StructLevel @ if  1  else  0  then  + INDENT_SPACES * spaces ;
}

\ *****************************************************************************
\ MD5 display
\ *****************************************************************************

\ dispay the MD5 hash of n bytes at a, plus the NumName string (see NumName.f for details)
: MD5show ( -- )   
   CYAN.  
   .MD5hash 
   RESTORE.
   BLUE.  2 spaces  MD5hash @ .NumNames  space  RESTORE.   
;

\ *****************************************************************************
\ File read using a buffer and GetFileX 
\ *****************************************************************************

#20 2** constant |MyFileBuffer| 
|MyFileBuffer| Buffer: MyFileBuffer[] 
variable MyFileHandle
variable MyFileSize
\ 2variable MyFileNameSaved
MAX_PATH Buffer: MyFileNameSaved[]

: ReadMyFile ( $filename -- )   \ read a file into my buffer
   2dup MyFileNameSaved[] place
   R/O OPEN-FILE if  cr RED. ." Could not open file "  RESTORE.  BLUE.  MyFileNameSaved[] count type  RESTORE. space  #10001 throw  then  MyFileHandle !
   MyFileBuffer[] |MyFileBuffer| MyFileHandle @  READ-FILE  if  
      MyFileHandle @ CLOSE-FILE drop  cr RED. ." Could not read file "  RESTORE.  #10002 throw  
   then
   MyFileSize ! 
   MyFileHandle @ CLOSE-FILE drop
   MyFileSize @ |MyFileBuffer| >= if  cr RED. ." File is too big for the buffer!!!  " RESTORE.  #10003 throw  then
;

: WriteMyFile ( $filename -- )   \ write the buffer into the given filename
   2dup MyFileNameSaved[] place
   R/W CREATE-FILE if  cr RED. ." WriteMyFile could not open file for writing "  BLUE.  MyFileNameSaved[] count type  RESTORE.  #10004 throw  then  MyFileHandle !
   MyFileBuffer[] MyFileSize @ MyFileHandle @ WRITE-FILE  if  
      MyFileHandle @ CLOSE-FILE drop  cr RED. ." Could not write file "  RESTORE.  #10005 throw  then
   MyFileHandle @ CLOSE-FILE drop
;

: WriteBackMyFile ( -- )
   cr ." Writing the buffer to file "  BLUE. MyFileNameSaved[] count type RESTORE.
   MyFileNameSaved[] count WriteMyFile
;

\ *****************************************************************************
\ Big 32 bit count versions of Place , Append and Count
\ The first 32 bit cell contains a 32 bit byte count
\ *****************************************************************************

: BigPlace ( a n a -- )
   2dup !  cell+  swap  cells  move 
;

: BigCount ( a -- a n )
   dup @ swap cell+ swap
;

: BigAppend ( a n a -- )
   2dup 2>r  BigCount + swap move  2r> +! ;
;


\ *****************************************************************************
\ *****************************************************************************

create SomeSpaces  $20 c, $20 c, $20 c, $20 c, $20 c, $20 c, $20 c, $20 c, 

: fdump ( $filename -- )   \ dump a binary file   
   ReadMyFile
   MyFileBuffer[] MyFileSize @ MD5  MD5Show
   MyFileBuffer[] MyFileSize @ Dump
;

: 100fdump ( $filename -- )   \ dump a binary file   
   ReadMyFile
   MyFileBuffer[] MyFileSize @ MD5  MD5Show
   MyFileBuffer[] MyFileSize @ $100 min Dump
;

: fdumpCapture ( $filename -- )   \ dump a binary file   
   ReadMyFile
   MyFileBuffer[] MyFileSize @ MD5  MD5Show
   MyFileBuffer[] MyFileSize @ DumpCapture
;


: fdumpCaptureN ( $filename -- )   \ dump a binary file   
   ReadMyFile
\   MyFileBuffer[] MyFileSize @ MD5  MD5Show
   MyFileBuffer[] MyFileSize @ DumpCaptureN
;

variable #crs
variable #lfs
variable #nulls
variable lastCR

: fdumpLineLengths ( $filename -- )
   ReadMyFile
   0 #crs !  0 #lfs !  0 #nulls ! 0 lastCR !
   MyFileBuffer[] MyFileSize @ over + swap ?do
      i c@ $0D = if  
         cr #crs @ 4 u.r 2 spaces i MyFileBuffer[] - dup >r 4 u.r space  r@ lastCR @ - 4 u.r  
         r> lastCR !
         1 #crs +! 
      then 
      i c@ $0A = if  1 #lfs +! then
      i c@   0 = if  1 #nulls +! then
   loop
   cr ." #crs = " #crs @ .  ."   #lfs = " #lfs @ .  ."   #nulls = " #nulls @ .
   #crs @ #lfs @ = not if  cr ." #crs and #lfs not equal!!!" then 
;

variable MyFileKeyPtr 

: GetFileInit ( -- )   0 MyFileKeyPtr ! ;

: GetFileLimited ( n -- n )   0 max  MyFileSize @ 1- 0 max min ;  \ safe access into the buffer

: >GetFile ( n -- a )   GetFileLimited  MyFileBuffer[] + ;        \ safe access to the Nth byte in the buffer

: >GetFileAddress ( -- a )   MyFileKeyPtr @ >GetFile ;

\ Get the next octet value from the file buffer
: GetFile8 ( -- c ) 
   >GetFileAddress c@
   MyFileKeyPtr @ 1+ dup MyFileSize @ > if  #10002 throw  then  MyFileKeyPtr !   \ throw at End Of File
\   MyFileKeyPtr @ 1+ GetFileLimited MyFileKeyPtr !                              \ carry on reading the last octet
;

\ Get the next octet value from the file buffer and leave it in the buffer
: PeekFile8 ( -- c ) 
   >GetFileAddress c@
;

\ Get the previous octet value from the file buffer ( and leave it in the buffer )
: PreviousFile8 ( -- c ) 
   MyFileKeyPtr @ 1- 0 max >GetFile c@
;



\ Get the next LittleBigEndian 2 octet value from the file buffer
: GetFile16LE ( -- w )   GetFile8  GetFile8 #08 lshift  or ;

\ Get the next LittleEndian 4 octet value from the file buffer
: GetFile32LE ( -- u )   GetFile16LE  GetFile16LE #16 lshift  or ;

\ Get the next BigEndian 2 octet value from the file buffer
: GetFile16 ( -- w )   GetFile8 #08 lshift  GetFile8 or ;

\ Get the next BigEndian 4 octet value from the file buffer
: GetFile32 ( -- u )   GetFile16  GetFile16 #16 lshift  or ;

: FileTypeL ( n -- )   
   MyFileKeyPtr @ >r   
   0 ?do  GetFile8 emit  loop 
   r> MyFileKeyPtr !
; 

: FileType ( offset n -- )   
   MyFileKeyPtr @ >r  
   swap MyFileKeyPtr ! 
   0 ?do  GetFile8 emit  loop 
   r> MyFileKeyPtr !
;

: FileDump ( offset n -- )   
   MyFileKeyPtr @ >r   
   swap MyFileKeyPtr !
   0 ?do  
      i $0F and 0= if  cr+ MyFileKeyPtr @ i + 4.hex 2 spaces  then 
      GetFile8 2.hex space  
   loop 
   r> MyFileKeyPtr !
; 

: FileDumpL ( n -- )   \ show n octets friom the current location in the file, restore the ptr
   MyFileKeyPtr @ >r   
   0 ?do  
      GetFile8 2.hex space  
   loop 
   r> MyFileKeyPtr !
;

: SkipFile ( n -- )   MyFileKeyPtr @ +  GetFileLimited  MyFileKeyPtr ! ; 

: fScan ( -- )
   MyFileBuffer[] MyFileSize @ over + swap ?do
\      i c@ $00 $20 within 
\      i c@ $0A =  i c@ $0D = or  not and 
      i $1F and 0= if  cr  i 8 .hex 2 spaces  then     
      i w@ $0500 = 
      if
         INVERSE.
\         cr i 8 .hex  space   ." = "  i c@ 3 .hex 
      else
         BLUE.
      then
      i w@ 4.hex space 
      RESTORE.
   2 +loop
;

: ttfscan  
   s" WST_Engl.ttf" fDump 
   fScan
;

: DumpCompare ( a a' n -- )
   0 ?do 
      i $0F and 0= if  cr i 6 u.r  ."  |  "  then   
      2dup 
      i + c@ 
      swap i + c@ 
      dup >r = if  NORMAL.  else  MAGENTA.  then   r>  2 .hex  space  RESTORE.
   loop
   2drop
;

: Wcount ( a -- a' n )
   >r  r@ 2+  r> w@
;

: Lcount ( a -- a' n )
   >r  r@ cell+  r> @
;

\ *****************************************************************************
\ Line buffer
\ *****************************************************************************

\ process the accumulated line buffer
: ProcessLine ( c-addr u -- )
   cr -Trailing 2dup type
    -Leading  ['] Dnumber catch if  2drop  else  D/M/Y space 10 u.r  then
;

9 2** constant |MySend|       \ the size of our send buffer - must be a power of 2
|MySend| Buffer: MySend[]     \ our send buffer
variable #MySend              \ how many bytes to send so far
variable #LastMySend          \ how many bytes to send so far

: bufferEmit ( c -- )
   MySend[] #MySend @ |MySend| 1- and + c!    \ store the character in the send buffer
   1 #MySend +!             \ increment the count
;

: InitNemit ( -- )   0 #MySend ! ;

\ accumulate characters, act on Backspace, ctrlC, Escape and Enter
\ Note the  ProcessLine  which does all of the work when you press Enter
: LineEmit ( c -- )
   dup <LF> = if  drop  exit  then  \ ignore LineFeeds
   dup <CR> = if
      drop
      #MySend @ #LastMySend !
      #MySend @ if
         MySend[] #MySend @ ProcessLine 
         InitNemit
      then
   else  
\   dup $08 = if  08 emit  bl emit  08 emit  #MySend @ 1- 0 max  #MySend !  drop  exit  then
\   dup <ESC> = if  cr ." Done."  #1006 throw  then 
\   dup $03 = if  cr ." ctrlC"  #1007 throw  then 
\   dup emit
   bufferEmit
   then
;


: FileLineDump ( $filename -- )  
   InitNemit
   ReadMyFile
   MyFileKeyPtr @ >r   
   begin

   MyFileKeyPtr @ MyFileSize @ < while
      GetFile8 LineEmit
   repeat

\   0 ?do  
\      i $0F and 0= if  cr+ MyFileKeyPtr @ i + 4.hex 2 spaces  then 
\      GetFile8 2.hex space  
\   loop 
   r> MyFileKeyPtr !
; 

: ttfld ( -- )  s" MyTextFile.txt" FileLineDump ;

\ *****************************************************************************
\ Misc
\ *****************************************************************************

: #define ( -<name expression>- )
        create
        bl word count Dnumber drop ,  does>  @
;

: TestBit ( c a -- f )   @ swap 2** and ;


: SetBit ( c a -- )   >r  2** r@ @ or r> ! ;
: ClearBit ( c a -- )   >r  2** invert r@ @ and r> ! ;

: ToggleBit ( c a -- )   2dup TestBit if  ClearBit  else  SetBit  then ;

: .On/Off ( c a -- )   TestBit if  INVERSE.  ." +"  else  NORMAL. ." -"  then  BLUE. ;


: CalcChecksum ( a n -- c )   0  -rot over + swap ?do  i c@ xor  loop ;    \ the checksum used in the protocol

: Toggle ( a -- )   dup @ 0= swap ! ;  \ toggle the state of the variable at a

: cToggle ( a -- )   dup c@ 0= swap c! ;  \ toggle the state of the cvariable at a

\ *****************************************************************************
\ *****************************************************************************
\ *****************************************************************************

