\ ******************************************************************************
\ vt100host.f 2016 Nov 15 ANSI VT100 terminal SwiftForth version
\ ******************************************************************************

variable vANSIstate \ 0 = not escaped, 1 = esc received, 2 = '[' received
variable vNumArgs
variable vTheNumber
2variable ANSI_X_Y   \ last x,y location set
ANSI_X_Y constant ANSI_Y     \ last y location set
ANSI_X_Y cell+ constant ANSI_X     \ last x location set

$10 constant |Vstack| 
|Vstack| cells Buffer: Vstack[]

variable Vptr

: InitVT100  
   0 Vptr ! Vstack[]  |Vstack| cells erase 
   0 vNumArgs !  0 vTheNumber !  0 vANSIstate !  
;
InitVT100

: >Vstack ( -- a )
   Vptr @ |Vstack| 1- and cells Vstack[] + 
;

: Vpush ( -- n )
   >Vstack !  1 Vptr +!
\   red. ."  vpush " restore.
;

: Vpop ( -- n )
   -1 Vptr +!  >Vstack @  
\   red. ."  vpop " restore.
;

variable #chars

: VTNUM. ( u -- )
   0 #chars !
   begin
      #10 /mod
   over while
   1 #chars +!
   repeat
   2drop
   #chars @ 
   0 ?do  [char] 0 + SerialEmit  loop
;

defer MyAction

$20 constant |CPU%string| 
|CPU%string| Buffer: CPU%string[]
|CPU%string| Buffer: CPU%stringLast[]
variable vNumberToCapture
variable vCaptureStartOffset

\ *****************************************************************************
\ Virtual VT100 terminal screen
\ *****************************************************************************
$80 constant cXnumberChars  $50 constant cXnumberCharsShown
$40 constant cYnumberLines  $10 constant cYnumberLinesShown

cXnumberChars cYnumberLines * 2* constant |VirtualScreen| 
|VirtualScreen| Buffer: VirtualScreen[]

variable vVirtualScreenPtr

: InitVirtualScreen ( -- )
   VirtualScreen[] |VirtualScreen| erase
   0 vVirtualScreenPtr !
;
InitVirtualScreen

: >VirtualScreen ( n -- a )
   2* |VirtualScreen| 1- and  VirtualScreen[] +
;

: VSemit ( c -- )  
   CurrentAttribute @ >< or  vVirtualScreenPtr @ >VirtualScreen w!
   1 vVirtualScreenPtr +!
;

: ShowVirtualScreenLine ( line -- )
   cXnumberChars * >VirtualScreen
   cXnumberCharsShown 0 do
      dup w@ dup >< $FF and attribute  $FF and emit  
      2+
   loop  
   drop
;

: ShowVirtualScreen ( -- )
   NORMAL. 
   0 #23 At-XY cXnumberCharsShown 0 do [char] = emit  loop
   
   cYnumberLinesShown 0 do
      0  i #24 + At-XY   
      i ShowVirtualScreenLine
   loop
   
   RESTORE. NORMAL. 
   0 #23 cYnumberLinesShown + At-XY cXnumberCharsShown 0 do [char] = emit  loop
   RESTORE.
;

: VT_At-XY ( x y -- )   
   2dup  At-XY  
   2dup cXnumberChars 2* * + vVirtualScreenPtr !
   ANSI_X_Y 2! 
{ 
   ANSI_Y @ $01 = if
      ANSI_X @ $0E $22 1+ within if
         \ 15 spaces  get-xy swap 12 - swap at-xy  2dup MAGENTA. swap 2.hex space 2.hex space RESTORE.  
         ANSI_X @ $0E - vCaptureStartOffset !
         $23 ANSI_X @ - vNumberToCapture !
         vNumberToCapture @ 0 max |CPU%string| 1- min 0 ?do
            serialKey dup [char] 0 [char] 9 1+ within over [char] . = or if
               CPU%string[] vCaptureStartOffset @ + i + c!  
            else 
               drop
               \ leave
            then
         loop
         CPU%string[] CPU%stringLast[] |CPU%string| move
      then
      ANSI_X_Y 2@ drop $3B < if
         get-xy BLUE.  0 #26 at-xy CPU%stringLast[] |CPU%string| dumpL RESTORE.  
         cr ." X,Y = "   ANSI_X_Y 2@ swap 2 .hex space 2 .hex space 
         cr ." vNumberToCapture = " vNumberToCapture @ 2 .hex
         cr ." vCaptureStartOffset = " vCaptureStartOffset @ 2 .hex
         at-xy
      then
   then
   2drop
\   MyAction
}
\   cr ." Depth = " depth .
;  

: InitCPU%string[] ( -- )
   CPU%string[] |CPU%string| erase  CPU%stringLast[] |CPU%string| erase
;

\ respond to the 'n' requests
: ANSIresponse ( n)   
\   cr ." ANSIresponse : "
   case
   0 of    endof
   3 of    endof
   5 of    endof
   6 of  
      #27 SerialEmit  S" [" SerialType Get-XY 
      VTNUM. [CHAR] ; SerialEmit  VTNUM. #13 SerialEmit
   endof
   drop
   endcase
;

: SetAttribute ( n -- ) 
   case  \ foreground colours	
      #30 of  0  ATTRIBUTE  endof \ BLACK 
      #31 of  2  ATTRIBUTE  endof \ RED		
      #32 of  7  ATTRIBUTE  endof \ GREEN	
      #33 of  6  ATTRIBUTE  endof \ YELLOW	
      #34 of  3  ATTRIBUTE  endof \ BLUE		
      #35 of  5  ATTRIBUTE  endof \ MAGENTA	
      #36 of  4  ATTRIBUTE  endof \ CYAN		
      #37 of  1  ATTRIBUTE  endof \ WHITE	
      0  ATTRIBUTE
   endcase  
;                 

\ when a non-digit ends a number, put the value on the Vstack
: StackNumber ( -- )
   vTheNumber @ Vpush   \ stack the number we have just got
   0 vTheNumber !       \ reset for the next number
   1 vNumArgs +!        \ up the count of arguments received
;

\ add digit c to the number being accumulated
: ANSInumber ( c -- )
   $30 - 0 max 
   vTheNumber @ #10 *  +  vTheNumber !
;

\ position values have 1 added to them before being sent, so we subtract 1 here
: GetTheNumber-1 ( -- n )   vTheNumber @ 1- 0 max ;

: ANSIaction ( c -- )
   dup >r
   case 
   [char] A of  Get-XY  GetTheNumber-1  - 0 max  At-XY  endof
   [char] B of  Get-XY  GetTheNumber-1  +  At-XY  endof
   [char] C of  Get-XY  swap GetTheNumber-1  +  swap  At-XY  endof
   [char] D of  Get-XY  swap GetTheNumber-1  -  0 max  swap  At-XY  endof
   [char] H of  
       vNumArgs @ 0= if  0 0  else  
         Vpop 1- 0 max  GetTheNumber-1 0 max  then  
\      cr ." H = " 2dup swap . .   ."   Vnum = "  vNumArgs @ .
      swap VT_At-XY  
   endof
   [char] f of  
      vNumArgs @ 0= if  0 0  else  Vpop 1- 0 max GetTheNumber-1   then  
      swap VT_At-XY  
   endof
   [char] J of  PAGE  endof
   [char] K of  Get-XY swap drop 0 swap 2dup At-XY  80 spaces  At-XY  endof
   [char] M of  CR  endof
   [char] X of  ( ??? )  endof
   [char] ; of  StackNumber  endof 
   [char] m of  vTheNumber @ SetAttribute  endof
   [char] n of  vTheNumber @ ANSIresponse  endof
   [char] c of  PAGE  0 ATTRIBUTE  endof
   [char] l of  ( ??? ) endof
   [char] r of  ( ??? ) endof
   [char] ? of  ( ??? ) endof
   
   cr ." Unknown ANSIaction = " . 

   endcase
   r> [char] ; = not
   if  
      0 vANSIstate !     \ return to normal state
      InitVT100
   then
;

: ANSIchar91 ( c) \ process a '[' escape sequence
   dup [char] 0 [char] 9 1+ within if 
      ANSInumber
   else
      ANSIaction
   then
;

: ANSIswitch ( c)
\   cr ." ANSIswitch = " dup 3 u.r space  
\   WaitKey
   case
   [char] [     of  2 vANSIstate !  endof
   [char] ( ( ) of  1 attribute  0 vANSIstate !     
      serialKey 
\      dup red. ." >>>>>>>> " 2 .hex space  restore.
      drop
   endof \ the ( ) is for the text editor
   [char] )     of  0 attribute  0 vANSIstate !  endof
   0 vANSIstate !
   endcase
;

\ like  EMIT , but decodes and actions VT100 escape sequences
: ANSIemit ( c -- )
   vANSIstate @ case
   0 of  
      dup #27 = if  
         drop 
         \ start procesing escape sequence
         0 vNumArgs !  0 vTheNumber !  1 vANSIstate !  
      else  
         VSemit  
      then  
   endof
   1 of  
\      cr ." State = 1"
      ANSIswitch  
   endof \ according to the character after <esc>
   2 of  
\      cr ." State = 2"
\      DUP emit
      ANSIchar91  
   endof
   endcase
;

{ 
\ Example polyForth vectored VT100 I/O

( ANSI terminal)   EMPTY   DECIMAL

: MSG   CREATE  DOES>  COUNT TYPE ;

MSG esc[   2 C, 27 C, 91 C,

: (CR)   L# TALLY  TOP 1+ C@ ?DUP IF  L# @ OVER -
      7 MOD + 0 TAB  CLEAN  ELSE  10 ( LF) EMIT
   THEN  13 ( cr) EMIT  -1 L# +! ;

: (PAGE)   0. !TOP  esc[ ." H"  esc[ ." J" ;

: (TAB) ( l c)   BASE @ >R  DECIMAL  esc[  SWAP
   1+ (.) TYPE  ." ;"  1+ (.) TYPE  ." H"  R> BASE ! ;

: (MARK) ( a n)   >R  PAD I SMOVE  esc[ ." 7m"  PAD R> TYPE
   esc[ ." m" ;

MSG (CLEAN)   3 C, 27 C, 91 C, 75 C,

' (CR) 'CR !  ' (PAGE) 'PAGE !  ' (TAB) 'TAB !
' (MARK) 'MARK !  ' (CLEAN) 'CLEAN !

HERE H 2+ !  EMPTY
}

